!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief This module defines the grid data type and some basic operations on it
!> \note
!>      pw_grid_create : set the defaults
!>      pw_grid_release : release all memory connected to type
!>      pw_grid_setup  : main routine to set up a grid
!>           input: cell (the box for the grid)
!>                  pw_grid (the grid; pw_grid%grid_span has to be set)
!>                  cutoff (optional, if not given pw_grid%bounds has to be set)
!>                  pe_group (optional, if not given we have a local grid)
!>
!>                  if no cutoff or a negative cutoff is given, all g-vectors
!>                  in the box are included (no spherical cutoff)
!>
!>                  for a distributed setup the array in para rs_dims has to
!>                  be initialized
!>           output: pw_grid
!>
!>      pw_grid_change : updates g-vectors after a change of the box
!> \par History
!>      JGH (20-12-2000) : Adapted for parallel use
!>      JGH (07-02-2001) : Added constructor and destructor routines
!>      JGH (21-02-2003) : Generalized reference grid concept
!>      JGH (19-11-2007) : Refactoring and modularization
!>      JGH (21-12-2007) : pw_grid_setup refactoring
!> \author apsi
!>      CJM
! **************************************************************************************************
MODULE pw_grids
   USE ISO_C_BINDING,                   ONLY: C_F_POINTER,&
                                              C_INT,&
                                              C_LOC,&
                                              C_PTR,&
                                              C_SIZE_T
   USE kinds,                           ONLY: dp,&
                                              int_8,&
                                              int_size
   USE mathconstants,                   ONLY: twopi
   USE mathlib,                         ONLY: det_3x3,&
                                              inv_3x3
   USE message_passing,                 ONLY: &
        MPI_COMM_SELF, mp_allgather, mp_cart_coords, mp_cart_create, mp_cart_rank, &
        mp_comm_compare, mp_comm_dup, mp_comm_free, mp_dims_create, mp_environ, mp_max, mp_min, &
        mp_sum
   USE pw_grid_info,                    ONLY: pw_find_cutoff,&
                                              pw_grid_bounds_from_n,&
                                              pw_grid_init_setup
   USE pw_grid_types,                   ONLY: FULLSPACE,&
                                              HALFSPACE,&
                                              PW_MODE_DISTRIBUTED,&
                                              PW_MODE_LOCAL,&
                                              map_pn,&
                                              pw_grid_type
   USE util,                            ONLY: get_limit,&
                                              sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: pw_grid_create, pw_grid_retain, pw_grid_release
   PUBLIC :: get_pw_grid_info, pw_grid_compare
   PUBLIC :: pw_grid_setup
   PUBLIC :: pw_grid_change

   INTEGER :: grid_tag = 0
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pw_grids'


   ! Distribution in g-space can be
   INTEGER, PARAMETER, PUBLIC               :: do_pw_grid_blocked_false = 0, &
                                               do_pw_grid_blocked_true = 1, &
                                               do_pw_grid_blocked_free = 2
CONTAINS

! **************************************************************************************************
!> \brief Initialize a PW grid with all defaults
!> \param pw_grid ...
!> \param pe_group ...
!> \param local ...
!> \par History
!>      JGH (21-Feb-2003) : initialize pw_grid%reference
!> \author JGH (7-Feb-2001) & fawzi
! **************************************************************************************************
   SUBROUTINE pw_grid_create(pw_grid, pe_group, local)

      TYPE(pw_grid_type), POINTER                        :: pw_grid
      INTEGER, INTENT(in)                                :: pe_group
      LOGICAL, INTENT(IN), OPTIONAL                      :: local

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_create', routineP = moduleN//':'//routineN

      LOGICAL                                            :: my_local

      my_local = .FALSE.
      IF (PRESENT(local)) my_local = local
      IF(.NOT.(.NOT. ASSOCIATED(pw_grid)))CALL cp__a("pw/pw_grids.F",122)
      ALLOCATE (pw_grid)
      pw_grid%bounds = 0
      pw_grid%cutoff = 0.0_dp
      pw_grid%grid_span = FULLSPACE
      pw_grid%para%mode = PW_MODE_LOCAL
      pw_grid%para%rs_dims = 0
      pw_grid%reference = 0
      pw_grid%ref_count = 1
      NULLIFY (pw_grid%g)
      NULLIFY (pw_grid%gsq)
      NULLIFY (pw_grid%g_hat)
      NULLIFY (pw_grid%g_hatmap)
      NULLIFY (pw_grid%gidx)
      NULLIFY (pw_grid%grays)
      NULLIFY (pw_grid%mapl%pos)
      NULLIFY (pw_grid%mapl%neg)
      NULLIFY (pw_grid%mapm%pos)
      NULLIFY (pw_grid%mapm%neg)
      NULLIFY (pw_grid%mapn%pos)
      NULLIFY (pw_grid%mapn%neg)
      NULLIFY (pw_grid%para%yzp)
      NULLIFY (pw_grid%para%yzq)
      NULLIFY (pw_grid%para%nyzray)
      NULLIFY (pw_grid%para%bo)
      NULLIFY (pw_grid%para%pos_of_x)

      ! assign a unique tag to this grid
      grid_tag = grid_tag+1
      pw_grid%id_nr = grid_tag

      ! parallel info
      CALL mp_comm_dup(pe_group, pw_grid%para%group)
      CALL mp_environ(pw_grid%para%group_size, &
                      pw_grid%para%my_pos, &
                      pw_grid%para%group)
      pw_grid%para%group_head_id = 0
      pw_grid%para%group_head = &
         (pw_grid%para%group_head_id == pw_grid%para%my_pos)
      IF (pw_grid%para%group_size > 1 .AND. (.NOT. my_local)) THEN
         pw_grid%para%mode = PW_MODE_DISTRIBUTED
      ELSE
         pw_grid%para%mode = PW_MODE_LOCAL
      END IF

   END SUBROUTINE pw_grid_create

! **************************************************************************************************
!> \brief Check if two pw_grids are equal
!> \param grida ...
!> \param gridb ...
!> \retval equal ...
!> \par History
!>      none
!> \author JGH (14-Feb-2001)
! **************************************************************************************************
   FUNCTION pw_grid_compare(grida, gridb) RESULT(equal)

      TYPE(pw_grid_type), INTENT(IN)                     :: grida, gridb
      LOGICAL                                            :: equal

!------------------------------------------------------------------------------

      IF (grida%id_nr == gridb%id_nr) THEN
         equal = .TRUE.
      ELSE
         ! for the moment all grids with different identifiers are considered as not equal
         ! later we can get this more relaxed
         equal = .FALSE.
      END IF

   END FUNCTION pw_grid_compare

! **************************************************************************************************
!> \brief Access to information stored in the pw_grid_type
!> \param pw_grid ...
!> \param id_nr ...
!> \param mode ...
!> \param vol ...
!> \param dvol ...
!> \param npts ...
!> \param ngpts ...
!> \param ngpts_cut ...
!> \param dr ...
!> \param cutoff ...
!> \param orthorhombic ...
!> \param gvectors ...
!> \param gsquare ...
!> \par History
!>      none
!> \author JGH (17-Nov-2007)
! **************************************************************************************************
   SUBROUTINE get_pw_grid_info(pw_grid, id_nr, mode, vol, dvol, npts, ngpts, &
                               ngpts_cut, dr, cutoff, orthorhombic, gvectors, gsquare)

      TYPE(pw_grid_type), POINTER                        :: pw_grid
      INTEGER, INTENT(OUT), OPTIONAL                     :: id_nr, mode
      REAL(dp), INTENT(OUT), OPTIONAL                    :: vol, dvol
      INTEGER, DIMENSION(3), INTENT(OUT), OPTIONAL       :: npts
      INTEGER(int_8), INTENT(OUT), OPTIONAL              :: ngpts, ngpts_cut
      REAL(dp), DIMENSION(3), INTENT(OUT), OPTIONAL      :: dr
      REAL(dp), INTENT(OUT), OPTIONAL                    :: cutoff
      LOGICAL, INTENT(OUT), OPTIONAL                     :: orthorhombic
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: gvectors
      REAL(dp), DIMENSION(:), OPTIONAL, POINTER          :: gsquare

      CHARACTER(len=*), PARAMETER :: routineN = 'get_pw_grid_info', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(pw_grid%ref_count > 0))CALL cp__a("pw/pw_grids.F",231)

      IF (PRESENT(id_nr)) id_nr = pw_grid%id_nr
      IF (PRESENT(mode)) mode = pw_grid%para%mode
      IF (PRESENT(vol)) vol = pw_grid%vol
      IF (PRESENT(dvol)) dvol = pw_grid%dvol
      IF (PRESENT(npts)) npts(1:3) = pw_grid%npts(1:3)
      IF (PRESENT(ngpts)) ngpts = pw_grid%ngpts
      IF (PRESENT(ngpts_cut)) ngpts_cut = pw_grid%ngpts_cut
      IF (PRESENT(dr)) dr = pw_grid%dr
      IF (PRESENT(cutoff)) cutoff = pw_grid%cutoff
      IF (PRESENT(orthorhombic)) orthorhombic = pw_grid%orthorhombic
      IF (PRESENT(gvectors)) gvectors => pw_grid%g
      IF (PRESENT(gsquare)) gsquare => pw_grid%gsq

   END SUBROUTINE get_pw_grid_info

! **************************************************************************************************
!> \brief Set some information stored in the pw_grid_type
!> \param pw_grid ...
!> \param grid_span ...
!> \param npts ...
!> \param bounds ...
!> \param cutoff ...
!> \param spherical ...
!> \par History
!>      none
!> \author JGH (19-Nov-2007)
! **************************************************************************************************
   SUBROUTINE set_pw_grid_info(pw_grid, grid_span, npts, bounds, cutoff, spherical)

      TYPE(pw_grid_type), POINTER                        :: pw_grid
      INTEGER, INTENT(in), OPTIONAL                      :: grid_span
      INTEGER, DIMENSION(3), INTENT(IN), OPTIONAL        :: npts
      INTEGER, DIMENSION(2, 3), INTENT(IN), OPTIONAL     :: bounds
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: cutoff
      LOGICAL, INTENT(IN), OPTIONAL                      :: spherical

      CHARACTER(len=*), PARAMETER :: routineN = 'set_pw_grid_info', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(pw_grid%ref_count > 0))CALL cp__a("pw/pw_grids.F",272)

      IF (PRESENT(grid_span)) THEN
         pw_grid%grid_span = grid_span
      END IF
      IF (PRESENT(bounds) .AND. PRESENT(npts)) THEN
         pw_grid%bounds = bounds
         pw_grid%npts = npts
         IF(.NOT.(ALL(npts == bounds(2, :)-bounds(1, :)+1)))CALL cp__a("pw/pw_grids.F",280)
      ELSE IF (PRESENT(bounds)) THEN
         pw_grid%bounds = bounds
         pw_grid%npts = bounds(2, :)-bounds(1, :)+1
      ELSE IF (PRESENT(npts)) THEN
         pw_grid%npts = npts
         pw_grid%bounds = pw_grid_bounds_from_n(npts)
      END IF
      IF (PRESENT(cutoff)) THEN
         pw_grid%cutoff = cutoff
         IF (PRESENT(spherical)) THEN
            pw_grid%spherical = spherical
         ELSE
            pw_grid%spherical = .FALSE.
         END IF
      END IF

   END SUBROUTINE set_pw_grid_info

! **************************************************************************************************
!> \brief sets up a pw_grid
!> \param cell_hmat ...
!> \param pw_grid ...
!> \param grid_span ...
!> \param cutoff ...
!> \param bounds ...
!> \param bounds_local ...
!> \param npts ...
!> \param spherical ...
!> \param odd ...
!> \param fft_usage ...
!> \param ncommensurate ...
!> \param icommensurate ...
!> \param blocked ...
!> \param ref_grid ...
!> \param rs_dims ...
!> \param iounit ...
!> \author JGH (21-Dec-2007)
!> \note
!>      this is the function that should be used in the future
! **************************************************************************************************
   SUBROUTINE pw_grid_setup(cell_hmat, pw_grid, grid_span, cutoff, bounds, bounds_local, npts, &
                            spherical, odd, fft_usage, ncommensurate, icommensurate, blocked, ref_grid, &
                            rs_dims, iounit)

      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: cell_hmat
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      INTEGER, INTENT(in), OPTIONAL                      :: grid_span
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: cutoff
      INTEGER, DIMENSION(2, 3), INTENT(IN), OPTIONAL     :: bounds, bounds_local
      INTEGER, DIMENSION(3), INTENT(IN), OPTIONAL        :: npts
      LOGICAL, INTENT(in), OPTIONAL                      :: spherical, odd, fft_usage
      INTEGER, INTENT(in), OPTIONAL                      :: ncommensurate, icommensurate, blocked
      TYPE(pw_grid_type), INTENT(IN), OPTIONAL           :: ref_grid
      INTEGER, DIMENSION(2), INTENT(in), OPTIONAL        :: rs_dims
      INTEGER, INTENT(in), OPTIONAL                      :: iounit

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_setup', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, my_icommensurate, &
                                                            my_ncommensurate
      INTEGER, DIMENSION(3)                              :: n
      LOGICAL                                            :: my_fft_usage, my_odd, my_spherical
      REAL(KIND=dp)                                      :: cell_deth, my_cutoff
      REAL(KIND=dp), DIMENSION(3, 3)                     :: cell_h_inv

      CALL timeset(routineN, handle)

      cell_deth = ABS(det_3x3(cell_hmat))
      IF (cell_deth < 1.0E-10_dp) THEN
         CALL cp_abort(cp__l("pw/pw_grids.F",350), &
                       "An invalid set of cell vectors was specified. "// &
                       "The determinant det(h) is too small")
      END IF
      cell_h_inv = inv_3x3(cell_hmat)

      IF (PRESENT(grid_span)) THEN
         CALL set_pw_grid_info(pw_grid, grid_span=grid_span)
      END IF

      IF (PRESENT(spherical)) THEN
         my_spherical = spherical
      ELSE
         my_spherical = .FALSE.
      END IF

      IF (PRESENT(odd)) THEN
         my_odd = odd
      ELSE
         my_odd = .FALSE.
      END IF

      IF (PRESENT(fft_usage)) THEN
         my_fft_usage = fft_usage
      ELSE
         my_fft_usage = .FALSE.
      END IF

      IF (PRESENT(ncommensurate)) THEN
         my_ncommensurate = ncommensurate
         IF (PRESENT(icommensurate)) THEN
            my_icommensurate = icommensurate
         ELSE
            my_icommensurate = MIN(1, ncommensurate)
         END IF
      ELSE
         my_ncommensurate = 0
         my_icommensurate = 1
      END IF

      IF (PRESENT(bounds)) THEN
         IF (PRESENT(cutoff)) THEN
            CALL set_pw_grid_info(pw_grid, bounds=bounds, cutoff=cutoff, &
                                  spherical=my_spherical)
         ELSE
            n = bounds(2, :)-bounds(1, :)+1
            my_cutoff = pw_find_cutoff(n, cell_h_inv)
            my_cutoff = 0.5_dp*my_cutoff*my_cutoff
            CALL set_pw_grid_info(pw_grid, bounds=bounds, cutoff=my_cutoff, &
                                  spherical=my_spherical)
         END IF
      ELSE IF (PRESENT(npts)) THEN
         n = npts
         IF (PRESENT(cutoff)) THEN
            my_cutoff = cutoff
         ELSE
            my_cutoff = pw_find_cutoff(npts, cell_h_inv)
            my_cutoff = 0.5_dp*my_cutoff*my_cutoff
         END IF
         IF (my_fft_usage) THEN
            n = pw_grid_init_setup(cell_hmat, cutoff=my_cutoff, &
                                   spherical=my_spherical, odd=my_odd, fft_usage=my_fft_usage, &
                                   ncommensurate=my_ncommensurate, icommensurate=my_icommensurate, &
                                   ref_grid=ref_grid, n_orig=n)
         END IF
         CALL set_pw_grid_info(pw_grid, npts=n, cutoff=my_cutoff, &
                               spherical=my_spherical)
      ELSE IF (PRESENT(cutoff)) THEN
         n = pw_grid_init_setup(cell_hmat, cutoff=cutoff, &
                                spherical=my_spherical, odd=my_odd, fft_usage=my_fft_usage, &
                                ncommensurate=my_ncommensurate, icommensurate=my_icommensurate, &
                                ref_grid=ref_grid)
         CALL set_pw_grid_info(pw_grid, npts=n, cutoff=cutoff, &
                               spherical=my_spherical)
      ELSE
         CALL cp__b("pw/pw_grids.F",425,"BOUNDS, NPTS or CUTOFF have to be specified")
      END IF

      CALL pw_grid_setup_internal(cell_hmat, cell_h_inv, cell_deth, pw_grid, bounds_local=bounds_local, &
                                  blocked=blocked, ref_grid=ref_grid, rs_dims=rs_dims, iounit=iounit)


      CALL timestop(handle)

   END SUBROUTINE pw_grid_setup


! **************************************************************************************************
!> \brief sets up a pw_grid, needs valid bounds as input, it is up to you to
!>      make sure of it using pw_grid_bounds_from_n
!> \param cell_hmat ...
!> \param cell_h_inv ...
!> \param cell_deth ...
!> \param pw_grid ...
!> \param bounds_local ...
!> \param blocked ...
!> \param ref_grid ...
!> \param rs_dims ...
!> \param iounit ...
!> \par History
!>      JGH (20-Dec-2000) : Adapted for parallel use
!>      JGH (28-Feb-2001) : New optional argument fft_usage
!>      JGH (21-Mar-2001) : Reference grid code
!>      JGH (21-Mar-2001) : New optional argument symm_usage
!>      JGH (22-Mar-2001) : Simplify group assignment (mp_comm_dup)
!>      JGH (21-May-2002) : Remove orthorhombic keyword (code is fast enough)
!>      JGH (19-Feb-2003) : Negative cutoff can be used for non-spheric grids
!>      Joost VandeVondele (Feb-2004) : optionally generate pw grids that are commensurate in rs
!>      JGH (18-Dec-2007) : Refactoring
!> \author fawzi
!> \note
!>      this is the function that should be used in the future
! **************************************************************************************************
   SUBROUTINE pw_grid_setup_internal(cell_hmat, cell_h_inv, cell_deth, pw_grid, bounds_local, &
                                     blocked, ref_grid, rs_dims, iounit)
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: cell_hmat, cell_h_inv
      REAL(KIND=dp), INTENT(IN)                          :: cell_deth
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      INTEGER, DIMENSION(2, 3), INTENT(IN), OPTIONAL     :: bounds_local
      INTEGER, INTENT(in), OPTIONAL                      :: blocked
      TYPE(pw_grid_type), INTENT(in), OPTIONAL           :: ref_grid
      INTEGER, DIMENSION(2), INTENT(in), OPTIONAL        :: rs_dims
      INTEGER, INTENT(in), OPTIONAL                      :: iounit

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_setup_internal', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ires, n(3)
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: yz_mask
      REAL(KIND=dp)                                      :: ecut

!------------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(pw_grid)))CALL cp__a("pw/pw_grids.F",575)
      IF(.NOT.(pw_grid%ref_count > 0))CALL cp__a("pw/pw_grids.F",576)

      ! set pointer to possible reference grid
      IF (PRESENT(ref_grid)) THEN
         pw_grid%reference = ref_grid%id_nr
      END IF

      IF (pw_grid%spherical) THEN
         ecut = pw_grid%cutoff
      ELSE
         ecut = 1.e10_dp
      END IF

      n(:) = pw_grid%npts(:)

      ! Find the number of grid points
      ! yz_mask counts the number of g-vectors orthogonal to the yz plane
      ! the indices in yz_mask are from -n/2 .. n/2 shifted by n/2 + 1
      ! these are not mapped indices !
      ALLOCATE (yz_mask(n(2), n(3)))
      CALL pw_grid_count(cell_h_inv, pw_grid, ecut, yz_mask)

      ! Check if reference grid is compatible
      IF (PRESENT(ref_grid)) THEN
         IF(.NOT.(pw_grid%para%mode == ref_grid%para%mode))CALL cp__a("pw/pw_grids.F",600)
         IF (pw_grid%para%mode == PW_MODE_DISTRIBUTED) THEN
            CALL mp_comm_compare(pw_grid%para%group, ref_grid%para%group, ires)
            IF(.NOT.(ires <= 3))CALL cp__a("pw/pw_grids.F",603)
         END IF
         IF(.NOT.(pw_grid%grid_span == ref_grid%grid_span))CALL cp__a("pw/pw_grids.F",605)
         IF(.NOT.(pw_grid%spherical .EQV. ref_grid%spherical))CALL cp__a("pw/pw_grids.F",606)
      END IF

      ! Distribute grid
      CALL pw_grid_distribute(pw_grid, yz_mask, bounds_local=bounds_local, ref_grid=ref_grid, blocked=blocked, &
                              rs_dims=rs_dims)

      ! Allocate the grid fields
      CALL pw_grid_allocate(pw_grid, pw_grid%ngpts_cut_local, &
                            pw_grid%bounds)

      ! Fill in the grid structure
      CALL pw_grid_assign(cell_h_inv, pw_grid, ecut)

      ! Sort g vector wrt length (only local for each processor)
      CALL pw_grid_sort(pw_grid, ref_grid)

      CALL pw_grid_remap(pw_grid, yz_mask)

      DEALLOCATE (yz_mask)

      CALL cell2grid(cell_hmat, cell_h_inv, cell_deth, pw_grid)
      !
      ! Output: All the information of this grid type
      !

      IF (PRESENT(iounit)) THEN
         CALL pw_grid_print(pw_grid, iounit)
      END IF

      CALL timestop(handle)

   END SUBROUTINE pw_grid_setup_internal

! **************************************************************************************************
!> \brief Helper routine used by pw_grid_setup_internal and pw_grid_change
!> \param cell_hmat ...
!> \param cell_h_inv ...
!> \param cell_deth ...
!> \param pw_grid ...
!> \par History moved common code into new subroutine
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE cell2grid(cell_hmat, cell_h_inv, cell_deth, pw_grid)
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: cell_hmat, cell_h_inv
      REAL(KIND=dp), INTENT(IN)                          :: cell_deth
      TYPE(pw_grid_type), POINTER                        :: pw_grid

      pw_grid%vol = ABS(cell_deth)
      pw_grid%dvol = pw_grid%vol/REAL(pw_grid%ngpts, KIND=dp)
      pw_grid%dr(1) = SQRT(SUM(cell_hmat(:, 1)**2)) &
                      /REAL(pw_grid%npts(1), KIND=dp)
      pw_grid%dr(2) = SQRT(SUM(cell_hmat(:, 2)**2)) &
                      /REAL(pw_grid%npts(2), KIND=dp)
      pw_grid%dr(3) = SQRT(SUM(cell_hmat(:, 3)**2)) &
                      /REAL(pw_grid%npts(3), KIND=dp)
      pw_grid%dh(:, 1) = cell_hmat(:, 1)/REAL(pw_grid%npts(1), KIND=dp)
      pw_grid%dh(:, 2) = cell_hmat(:, 2)/REAL(pw_grid%npts(2), KIND=dp)
      pw_grid%dh(:, 3) = cell_hmat(:, 3)/REAL(pw_grid%npts(3), KIND=dp)
      pw_grid%dh_inv(1, :) = cell_h_inv(1, :)*REAL(pw_grid%npts(1), KIND=dp)
      pw_grid%dh_inv(2, :) = cell_h_inv(2, :)*REAL(pw_grid%npts(2), KIND=dp)
      pw_grid%dh_inv(3, :) = cell_h_inv(3, :)*REAL(pw_grid%npts(3), KIND=dp)

      IF ((cell_hmat(1, 2) == 0.0_dp) .AND. (cell_hmat(1, 3) == 0.0_dp) .AND. &
          (cell_hmat(2, 1) == 0.0_dp) .AND. (cell_hmat(2, 3) == 0.0_dp) .AND. &
          (cell_hmat(3, 1) == 0.0_dp) .AND. (cell_hmat(3, 2) == 0.0_dp)) THEN
         pw_grid%orthorhombic = .TRUE.
      ELSE
         pw_grid%orthorhombic = .FALSE.
      END IF
   END SUBROUTINE cell2grid

! **************************************************************************************************
!> \brief Output of information on pw_grid
!> \param pw_grid ...
!> \param info ...
!> \author JGH[18-05-2007] from earlier versions
! **************************************************************************************************
   SUBROUTINE pw_grid_print(pw_grid, info)

      TYPE(pw_grid_type), POINTER                        :: pw_grid
      INTEGER, INTENT(IN)                                :: info

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_print', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      INTEGER(KIND=int_8)                                :: n(3)
      REAL(KIND=dp)                                      :: rv(3, 3)

!------------------------------------------------------------------------------
!
! Output: All the information of this grid type
!

      IF (pw_grid%para%mode == PW_MODE_LOCAL) THEN
         IF (info >= 0) THEN
            WRITE (info, '(/,A,T71,I10)') &
               " PW_GRID| Information for grid number ", pw_grid%id_nr
            WRITE (info, '(" PW_GRID| Cutoff [a.u.]",T71,f10.1)') pw_grid%cutoff
            IF (pw_grid%spherical) THEN
               WRITE (info, '(A,T78,A)') " PW_GRID| spherical cutoff: ", "YES"
               WRITE (info, '(A,T71,I10)') " PW_GRID| Grid points within cutoff", &
                  pw_grid%ngpts_cut
            ELSE
               WRITE (info, '(A,T78,A)') " PW_GRID| spherical cutoff: ", " NO"
            END IF
            DO i = 1, 3
               WRITE (info, '(A,I3,T30,2I8,T62,A,T71,I10)') " PW_GRID|   Bounds ", &
                  i, pw_grid%bounds(1, I), pw_grid%bounds(2, I), &
                  "Points:", pw_grid%npts(I)
            END DO
            WRITE (info, '(A,G12.4,T50,A,T67,F14.4)') &
               " PW_GRID| Volume element (a.u.^3)", &
               pw_grid%dvol, " Volume (a.u.^3) :", pw_grid%vol
            IF (pw_grid%grid_span == HALFSPACE) THEN
               WRITE (info, '(A,T72,A)') " PW_GRID| Grid span", "HALFSPACE"
            ELSE
               WRITE (info, '(A,T72,A)') " PW_GRID| Grid span", "FULLSPACE"
            END IF
         END IF
      ELSE

         n(1) = pw_grid%ngpts_cut_local
         n(2) = pw_grid%ngpts_local
         CALL mp_sum(n(1:2), pw_grid%para%group)
         n(3) = SUM(pw_grid%para%nyzray)
         rv(:, 1) = REAL(n, KIND=dp)/REAL(pw_grid%para%group_size, KIND=dp)
         n(1) = pw_grid%ngpts_cut_local
         n(2) = pw_grid%ngpts_local
         CALL mp_max(n(1:2), pw_grid%para%group)
         n(3) = MAXVAL(pw_grid%para%nyzray)
         rv(:, 2) = REAL(n, KIND=dp)
         n(1) = pw_grid%ngpts_cut_local
         n(2) = pw_grid%ngpts_local
         CALL mp_min(n(1:2), pw_grid%para%group)
         n(3) = MINVAL(pw_grid%para%nyzray)
         rv(:, 3) = REAL(n, KIND=dp)

         IF (pw_grid%para%group_head .AND. info >= 0) THEN
            WRITE (info, '(/,A,T71,I10)') &
               " PW_GRID| Information for grid number ", pw_grid%id_nr
            WRITE (info, '(A,T60,I10,A)') &
               " PW_GRID| Grid distributed over ", pw_grid%para%group_size, &
               " processors"
            WRITE (info, '(A,T71,2I5)') &
               " PW_GRID| Real space group dimensions ", pw_grid%para%rs_dims
            IF (pw_grid%para%blocked) THEN
               WRITE (info, '(A,T78,A)') " PW_GRID| the grid is blocked: ", "YES"
            ELSE
               WRITE (info, '(A,T78,A)') " PW_GRID| the grid is blocked: ", " NO"
            END IF
            WRITE (info, '(" PW_GRID| Cutoff [a.u.]",T71,f10.1)') pw_grid%cutoff
            IF (pw_grid%spherical) THEN
               WRITE (info, '(A,T78,A)') " PW_GRID| spherical cutoff: ", "YES"
               WRITE (info, '(A,T71,I10)') " PW_GRID| Grid points within cutoff", &
                  pw_grid%ngpts_cut
            ELSE
               WRITE (info, '(A,T78,A)') " PW_GRID| spherical cutoff: ", " NO"
            END IF
            DO i = 1, 3
               WRITE (info, '(A,I3,T30,2I8,T62,A,T71,I10)') " PW_GRID|   Bounds ", &
                  i, pw_grid%bounds(1, I), pw_grid%bounds(2, I), &
                  "Points:", pw_grid%npts(I)
            END DO
            WRITE (info, '(A,G12.4,T50,A,T67,F14.4)') &
               " PW_GRID| Volume element (a.u.^3)", &
               pw_grid%dvol, " Volume (a.u.^3) :", pw_grid%vol
            IF (pw_grid%grid_span == HALFSPACE) THEN
               WRITE (info, '(A,T72,A)') " PW_GRID| Grid span", "HALFSPACE"
            ELSE
               WRITE (info, '(A,T72,A)') " PW_GRID| Grid span", "FULLSPACE"
            END IF
            WRITE (info, '(A,T48,A)') " PW_GRID|   Distribution", &
               "  Average         Max         Min"
            WRITE (info, '(A,T45,F12.1,2I12)') " PW_GRID|   G-Vectors", &
               rv(1, 1), NINT(rv(1, 2)), NINT(rv(1, 3))
            WRITE (info, '(A,T45,F12.1,2I12)') " PW_GRID|   G-Rays   ", &
               rv(3, 1), NINT(rv(3, 2)), NINT(rv(3, 3))
            WRITE (info, '(A,T45,F12.1,2I12)') " PW_GRID|   Real Space Points", &
               rv(2, 1), NINT(rv(2, 2)), NINT(rv(2, 3))
         END IF ! group head
      END IF ! local

   END SUBROUTINE pw_grid_print

! **************************************************************************************************
!> \brief Distribute grids in real and Fourier Space to the processors in group
!> \param pw_grid ...
!> \param yz_mask ...
!> \param bounds_local ...
!> \param ref_grid ...
!> \param blocked ...
!> \param rs_dims ...
!> \par History
!>      JGH (01-Mar-2001) optional reference grid
!>      JGH (22-May-2002) bug fix for pre_tag and HALFSPACE grids
!>      JGH (09-Sep-2003) reduce scaling for distribution
!> \author JGH (22-12-2000)
! **************************************************************************************************
   SUBROUTINE pw_grid_distribute(pw_grid, yz_mask, bounds_local, ref_grid, blocked, rs_dims)

      TYPE(pw_grid_type), POINTER                        :: pw_grid
      INTEGER, DIMENSION(:, :), INTENT(INOUT)            :: yz_mask
      INTEGER, DIMENSION(2, 3), INTENT(IN), OPTIONAL     :: bounds_local
      TYPE(pw_grid_type), INTENT(IN), OPTIONAL           :: ref_grid
      INTEGER, INTENT(IN), OPTIONAL                      :: blocked
      INTEGER, DIMENSION(2), INTENT(in), OPTIONAL        :: rs_dims

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_distribute', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blocked_local, coor(2), gmax, handle, i, &
                                                            i1, i2, ip, ipl, ipp, itmp, j, l, lby, &
                                                            lbz, lo(2), m, n, np, ns, nx, ny, nz, &
                                                            rsd(2)
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: pemap
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: yz_index
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: axis_dist_all
      INTEGER, DIMENSION(2, 3)                           :: axis_dist
      LOGICAL                                            :: blocking

!------------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      lby = pw_grid%bounds(1, 2)
      lbz = pw_grid%bounds(1, 3)

      pw_grid%ngpts = PRODUCT(INT(pw_grid%npts, KIND=int_8))
      IF(.NOT.(ALL(pw_grid%para%rs_dims == 0)))CALL cp__a("pw/pw_grids.F",835)
      IF (PRESENT(rs_dims)) THEN
         pw_grid%para%rs_dims = rs_dims
      END IF

      IF (PRESENT(blocked)) THEN
         blocked_local = blocked
      ELSE
         blocked_local = do_pw_grid_blocked_free
      ENDIF

      pw_grid%para%blocked = .FALSE.

      IF (pw_grid%para%mode == PW_MODE_LOCAL) THEN

         pw_grid%para%ray_distribution = .FALSE.

         pw_grid%bounds_local = pw_grid%bounds
         pw_grid%npts_local = pw_grid%npts
         IF(.NOT.(pw_grid%ngpts_cut < HUGE(pw_grid%ngpts_cut_local)))CALL cp__a("pw/pw_grids.F",854)
         pw_grid%ngpts_cut_local = INT(pw_grid%ngpts_cut)
         IF(.NOT.(pw_grid%ngpts < HUGE(pw_grid%ngpts_local)))CALL cp__a("pw/pw_grids.F",856)
         pw_grid%ngpts_local = INT(pw_grid%ngpts)
         pw_grid%para%rs_dims = 1
         CALL mp_cart_create(MPI_COMM_SELF, 2, &
                             pw_grid%para%rs_dims, &
                             pw_grid%para%rs_pos, &
                             pw_grid%para%rs_group)
         CALL mp_cart_rank(pw_grid%para%rs_group, &
                           pw_grid%para%rs_pos, &
                           pw_grid%para%rs_mpo)
         pw_grid%para%group_size = 1
         ALLOCATE (pw_grid%para%bo(2, 3, 0:0, 3))
         DO i = 1, 3
            pw_grid%para%bo(1, 1:3, 0, i) = 1
            pw_grid%para%bo(2, 1:3, 0, i) = pw_grid%npts(1:3)
         END DO

      ELSE

         !..find the real space distribution
         nx = pw_grid%npts(1)
         ny = pw_grid%npts(2)
         nz = pw_grid%npts(3)
         np = pw_grid%para%group_size

         ! The user can specify 2 strictly positive indices => specific layout
         !                      1 strictly positive index   => the other is fixed by the number of CPUs
         !                      0 strictly positive indices => fully free distribution
         ! if fully free or the user request can not be fulfilled, we optimize heuristically ourselves by:
         !                      1) nx>np -> taking a plane distribution (/np,1/)
         !                      2) nx<np -> taking the most square distribution
         ! if blocking is free:
         !                      1) blocked=.FALSE. for plane distributions
         !                      2) blocked=.TRUE.  for non-plane distributions
         IF (ANY(pw_grid%para%rs_dims <= 0)) THEN
            IF (ALL(pw_grid%para%rs_dims <= 0)) THEN
               pw_grid%para%rs_dims = (/0, 0/)
            ELSE
               IF (pw_grid%para%rs_dims(1) > 0) THEN
                  pw_grid%para%rs_dims(2) = np/pw_grid%para%rs_dims(1)
               ELSE
                  pw_grid%para%rs_dims(1) = np/pw_grid%para%rs_dims(2)
               ENDIF
            ENDIF
         ENDIF
         ! reset if the distribution can not be fulfilled
         IF (PRODUCT(pw_grid%para%rs_dims) .NE. np) pw_grid%para%rs_dims = (/0, 0/)
         ! reset if the distribution can not be dealt with (/1,np/)
         IF (ALL(pw_grid%para%rs_dims == (/1, np/))) pw_grid%para%rs_dims = (/0, 0/)

         ! if (/0,0/) now, we can optimize it ourselves
         IF (ALL(pw_grid%para%rs_dims == (/0, 0/))) THEN
            ! only small grids have a chance to be 2d distributed
            IF (nx < np) THEN
               ! gives the most square looking distribution
               CALL mp_dims_create(np, pw_grid%para%rs_dims)
               ! we tend to like the first index being smaller than the second
               IF (pw_grid%para%rs_dims(1) > pw_grid%para%rs_dims(2)) THEN
                  itmp = pw_grid%para%rs_dims(1)
                  pw_grid%para%rs_dims(1) = pw_grid%para%rs_dims(2)
                  pw_grid%para%rs_dims(2) = itmp
               ENDIF
               ! but should avoid having the first index 1 in all cases
               IF (pw_grid%para%rs_dims(1) == 1) THEN
                  itmp = pw_grid%para%rs_dims(1)
                  pw_grid%para%rs_dims(1) = pw_grid%para%rs_dims(2)
                  pw_grid%para%rs_dims(2) = itmp
               ENDIF
            ELSE
               pw_grid%para%rs_dims = (/np, 1/)
            ENDIF
         ENDIF

         ! now fix the blocking if we have a choice
         SELECT CASE (blocked_local)
         CASE (do_pw_grid_blocked_false)
            blocking = .FALSE.
         CASE (do_pw_grid_blocked_true)
            blocking = .TRUE.
         CASE (do_pw_grid_blocked_free)
            IF (ALL(pw_grid%para%rs_dims == (/np, 1/))) THEN
               blocking = .FALSE.
            ELSE
               blocking = .TRUE.
            ENDIF
         CASE DEFAULT
            CALL cp__b("pw/pw_grids.F",942,"")
         END SELECT

         !..create group for real space distribution
         CALL mp_cart_create(pw_grid%para%group, 2, &
                             pw_grid%para%rs_dims, &
                             pw_grid%para%rs_pos, &
                             pw_grid%para%rs_group)
         CALL mp_cart_rank(pw_grid%para%rs_group, &
                           pw_grid%para%rs_pos, &
                           pw_grid%para%rs_mpo)

         IF (PRESENT(bounds_local)) THEN
            pw_grid%bounds_local = bounds_local
         ELSE
            lo = get_limit(nx, pw_grid%para%rs_dims(1), &
                           pw_grid%para%rs_pos(1))
            pw_grid%bounds_local(:, 1) = lo+pw_grid%bounds(1, 1)-1
            lo = get_limit(ny, pw_grid%para%rs_dims(2), &
                           pw_grid%para%rs_pos(2))
            pw_grid%bounds_local(:, 2) = lo+pw_grid%bounds(1, 2)-1
            pw_grid%bounds_local(:, 3) = pw_grid%bounds(:, 3)
         END IF

         pw_grid%npts_local(:) = pw_grid%bounds_local(2, :) &
                                 -pw_grid%bounds_local(1, :)+1

         !..the third distribution is needed for the second step in the FFT
         ALLOCATE (pw_grid%para%bo(2, 3, 0:np-1, 3))
         rsd = pw_grid%para%rs_dims

         IF (PRESENT(bounds_local)) THEN
            ! axis_dist tells what portion of 1 .. nx , 1 .. ny , 1 .. nz are in the current process
            DO i = 1, 3
               axis_dist(:, i) = bounds_local(:, i)-pw_grid%bounds(1, i)+1
            END DO
            ALLOCATE (axis_dist_all(2, 3, np))
            CALL mp_allgather(axis_dist, axis_dist_all, pw_grid%para%rs_group)
            DO ip = 0, np-1
               CALL mp_cart_coords(pw_grid%para%rs_group, ip, coor)
               ! distribution xyZ
               pw_grid%para%bo(1:2, 1, ip, 1) = axis_dist_all(1:2, 1, ip+1)
               pw_grid%para%bo(1:2, 2, ip, 1) = axis_dist_all(1:2, 2, ip+1)
               pw_grid%para%bo(1, 3, ip, 1) = 1
               pw_grid%para%bo(2, 3, ip, 1) = nz
               ! distribution xYz
               pw_grid%para%bo(1:2, 1, ip, 2) = axis_dist_all(1:2, 1, ip+1)
               pw_grid%para%bo(1, 2, ip, 2) = 1
               pw_grid%para%bo(2, 2, ip, 2) = ny
               pw_grid%para%bo(1:2, 3, ip, 2) = get_limit(nz, rsd(2), coor(2))
               ! distribution Xyz
               pw_grid%para%bo(1, 1, ip, 3) = 1
               pw_grid%para%bo(2, 1, ip, 3) = nx
               pw_grid%para%bo(1:2, 2, ip, 3) = get_limit(ny, rsd(1), coor(1))
               pw_grid%para%bo(1:2, 3, ip, 3) = get_limit(nz, rsd(2), coor(2))
            END DO
            DEALLOCATE (axis_dist_all)
         ELSE
            DO ip = 0, np-1
               CALL mp_cart_coords(pw_grid%para%rs_group, ip, coor)
               ! distribution xyZ
               pw_grid%para%bo(1:2, 1, ip, 1) = get_limit(nx, rsd(1), coor(1))
               pw_grid%para%bo(1:2, 2, ip, 1) = get_limit(ny, rsd(2), coor(2))
               pw_grid%para%bo(1, 3, ip, 1) = 1
               pw_grid%para%bo(2, 3, ip, 1) = nz
               ! distribution xYz
               pw_grid%para%bo(1:2, 1, ip, 2) = get_limit(nx, rsd(1), coor(1))
               pw_grid%para%bo(1, 2, ip, 2) = 1
               pw_grid%para%bo(2, 2, ip, 2) = ny
               pw_grid%para%bo(1:2, 3, ip, 2) = get_limit(nz, rsd(2), coor(2))
               ! distribution Xyz
               pw_grid%para%bo(1, 1, ip, 3) = 1
               pw_grid%para%bo(2, 1, ip, 3) = nx
               pw_grid%para%bo(1:2, 2, ip, 3) = get_limit(ny, rsd(1), coor(1))
               pw_grid%para%bo(1:2, 3, ip, 3) = get_limit(nz, rsd(2), coor(2))
            END DO
         END IF
         !..find the g space distribution
         pw_grid%ngpts_cut_local = 0

         ALLOCATE (pw_grid%para%nyzray(0:np-1))

         ALLOCATE (pw_grid%para%yzq(ny, nz))

         IF (pw_grid%spherical .OR. pw_grid%grid_span == HALFSPACE &
             .OR. .NOT. blocking) THEN

            pw_grid%para%ray_distribution = .TRUE.

            pw_grid%para%yzq = -1
            IF (PRESENT(ref_grid)) THEN
               ! tag all vectors from the reference grid
               CALL pre_tag(pw_grid, yz_mask, ref_grid)
            END IF

            ! Round Robin distribution
            ! Processors 0 .. NP-1, NP-1 .. 0  get the largest remaining batch
            ! of g vectors in turn

            i1 = SIZE(yz_mask, 1)
            i2 = SIZE(yz_mask, 2)
            ALLOCATE (yz_index(2, i1*i2))
            CALL order_mask(yz_mask, yz_index)
            DO i = 1, i1*i2
               lo(1) = yz_index(1, i)
               lo(2) = yz_index(2, i)
               IF (lo(1)*lo(2) == 0) CYCLE
               gmax = yz_mask(lo(1), lo(2))
               IF (gmax == 0) CYCLE
               yz_mask(lo(1), lo(2)) = 0
               ip = MOD(i-1, 2*np)
               IF (ip > np-1) ip = 2*np-ip-1
               IF (ip == pw_grid%para%my_pos) THEN
                  pw_grid%ngpts_cut_local = pw_grid%ngpts_cut_local+gmax
               END IF
               pw_grid%para%yzq(lo(1), lo(2)) = ip
               IF (pw_grid%grid_span == HALFSPACE) THEN
                  m = -lo(1)-2*lby+2
                  n = -lo(2)-2*lbz+2
                  pw_grid%para%yzq(m, n) = ip
                  yz_mask(m, n) = 0
               END IF
            END DO

            DEALLOCATE (yz_index)

            ! Count the total number of rays on each processor
            pw_grid%para%nyzray = 0
            DO i = 1, nz
               DO j = 1, ny
                  ip = pw_grid%para%yzq(j, i)
                  IF (ip >= 0) pw_grid%para%nyzray(ip) = &
                     pw_grid%para%nyzray(ip)+1
               END DO
            END DO

            ! Allocate mapping array (y:z, nray, nproc)
            ns = MAXVAL(pw_grid%para%nyzray(0:np-1))
            ALLOCATE (pw_grid%para%yzp(2, ns, 0:np-1))

            ! Fill mapping array, recalculate nyzray for convenience
            pw_grid%para%nyzray = 0
            DO i = 1, nz
               DO j = 1, ny
                  ip = pw_grid%para%yzq(j, i)
                  IF (ip >= 0) THEN
                     pw_grid%para%nyzray(ip) = &
                        pw_grid%para%nyzray(ip)+1
                     ns = pw_grid%para%nyzray(ip)
                     pw_grid%para%yzp(1, ns, ip) = j
                     pw_grid%para%yzp(2, ns, ip) = i
                     IF (ip == pw_grid%para%my_pos) THEN
                        pw_grid%para%yzq(j, i) = ns
                     ELSE
                        pw_grid%para%yzq(j, i) = -1
                     END IF
                  ELSE
                     pw_grid%para%yzq(j, i) = -2
                  END IF
               END DO
            END DO

            pw_grid%ngpts_local = PRODUCT(pw_grid%npts_local)

         ELSE
            !
            !  block distribution of g vectors, we do not have a spherical cutoff
            !

            pw_grid%para%blocked = .TRUE.
            pw_grid%para%ray_distribution = .FALSE.

            DO ip = 0, np-1
               m = pw_grid%para%bo(2, 2, ip, 3)- &
                   pw_grid%para%bo(1, 2, ip, 3)+1
               n = pw_grid%para%bo(2, 3, ip, 3)- &
                   pw_grid%para%bo(1, 3, ip, 3)+1
               pw_grid%para%nyzray(ip) = n*m
            END DO

            ipl = pw_grid%para%rs_mpo
            l = pw_grid%para%bo(2, 1, ipl, 3)- &
                pw_grid%para%bo(1, 1, ipl, 3)+1
            m = pw_grid%para%bo(2, 2, ipl, 3)- &
                pw_grid%para%bo(1, 2, ipl, 3)+1
            n = pw_grid%para%bo(2, 3, ipl, 3)- &
                pw_grid%para%bo(1, 3, ipl, 3)+1
            pw_grid%ngpts_cut_local = l*m*n
            pw_grid%ngpts_local = pw_grid%ngpts_cut_local

            pw_grid%para%yzq = 0
            ny = pw_grid%para%bo(2, 2, ipl, 3)- &
                 pw_grid%para%bo(1, 2, ipl, 3)+1
            DO n = pw_grid%para%bo(1, 3, ipl, 3), &
                   pw_grid%para%bo(2, 3, ipl, 3)
               i = n-pw_grid%para%bo(1, 3, ipl, 3)
               DO m = pw_grid%para%bo(1, 2, ipl, 3), &
                      pw_grid%para%bo(2, 2, ipl, 3)
                  j = m-pw_grid%para%bo(1, 2, ipl, 3)+1
                  pw_grid%para%yzq(m, n) = j+i*ny
               END DO
            END DO

            ! Allocate mapping array (y:z, nray, nproc)
            ns = MAXVAL(pw_grid%para%nyzray(0:np-1))
            ALLOCATE (pw_grid%para%yzp(2, ns, 0:np-1))
            pw_grid%para%yzp = 0

            ALLOCATE (pemap(0:np-1))
            pemap = 0
            pemap(pw_grid%para%my_pos) = pw_grid%para%rs_mpo
            CALL mp_sum(pemap, pw_grid%para%group)

            DO ip = 0, np-1
               ipp = pemap(ip)
               ns = 0
               DO n = pw_grid%para%bo(1, 3, ipp, 3), &
                      pw_grid%para%bo(2, 3, ipp, 3)
                  i = n-pw_grid%bounds(1, 3)+1
                  DO m = pw_grid%para%bo(1, 2, ipp, 3), &
                         pw_grid%para%bo(2, 2, ipp, 3)
                     j = m-pw_grid%bounds(1, 2)+1
                     ns = ns+1
                     pw_grid%para%yzp(1, ns, ip) = j
                     pw_grid%para%yzp(2, ns, ip) = i
                  END DO
               END DO
               IF(.NOT.(ns == pw_grid%para%nyzray(ip)))CALL cp__a("pw/pw_grids.F",1169)
            END DO

            DEALLOCATE (pemap)

         END IF

      END IF

      ! pos_of_x(i) tells on which cpu pw%cr3d(i,:,:) is located
      ! should be computable in principle, without the need for communication
      IF (pw_grid%para%mode .EQ. PW_MODE_DISTRIBUTED) THEN
         ALLOCATE (pw_grid%para%pos_of_x(pw_grid%bounds(1, 1):pw_grid%bounds(2, 1)))
         pw_grid%para%pos_of_x = 0
         pw_grid%para%pos_of_x(pw_grid%bounds_local(1, 1):pw_grid%bounds_local(2, 1)) = pw_grid%para%my_pos
         CALL mp_sum(pw_grid%para%pos_of_x, pw_grid%para%group)
      ELSE
         ! this should not be needed
         ALLOCATE (pw_grid%para%pos_of_x(pw_grid%bounds(1, 1):pw_grid%bounds(2, 1)))
         pw_grid%para%pos_of_x = 0
      END IF

      CALL timestop(handle)

   END SUBROUTINE pw_grid_distribute

! **************************************************************************************************
!> \brief ...
!> \param pw_grid ...
!> \param yz_mask ...
!> \param ref_grid ...
! **************************************************************************************************
   SUBROUTINE pre_tag(pw_grid, yz_mask, ref_grid)

      TYPE(pw_grid_type), POINTER                        :: pw_grid
      INTEGER, DIMENSION(:, :), INTENT(INOUT)            :: yz_mask
      TYPE(pw_grid_type), INTENT(IN)                     :: ref_grid

      INTEGER                                            :: gmax, ig, ip, lby, lbz, my, mz, ny, nz, &
                                                            uby, ubz, y, yp, z, zp

!------------------------------------------------------------------------------

      ny = ref_grid%npts(2)
      nz = ref_grid%npts(3)
      lby = pw_grid%bounds(1, 2)
      lbz = pw_grid%bounds(1, 3)
      uby = pw_grid%bounds(2, 2)
      ubz = pw_grid%bounds(2, 3)
      my = SIZE(yz_mask, 1)
      mz = SIZE(yz_mask, 2)

      ! loop over all processors and all g vectors yz lines on this processor
      DO ip = 0, ref_grid%para%group_size-1
         DO ig = 1, ref_grid%para%nyzray(ip)
            ! go from mapped coordinates to original coordinates
            ! 0 .. N-1 -> -n/2 .. (n+1)/2
            y = ref_grid%para%yzp(1, ig, ip)-1
            IF (y > ny/2) y = y-ny
            z = ref_grid%para%yzp(2, ig, ip)-1
            IF (z > nz/2) z = z-nz
            ! check if this is inside the realm of the new grid
            IF (y < lby .OR. y > uby .OR. z < lbz .OR. z > ubz) CYCLE
            ! go to shifted coordinates
            y = y-lby+1
            z = z-lbz+1
            ! this tag is outside the cutoff range of the new grid
            IF (pw_grid%grid_span == HALFSPACE) THEN
               yp = -y-2*lby+2
               zp = -z-2*lbz+2
               ! if the referenz grid is larger than the mirror point may be
               ! outside the new grid even if the original point is inside
               IF (yp < 1 .OR. yp > my .OR. zp < 1 .OR. zp > mz) CYCLE
               gmax = MAX(yz_mask(y, z), yz_mask(yp, zp))
               IF (gmax == 0) CYCLE
               yz_mask(y, z) = 0
               yz_mask(yp, zp) = 0
               pw_grid%para%yzq(y, z) = ip
               pw_grid%para%yzq(yp, zp) = ip
            ELSE
               gmax = yz_mask(y, z)
               IF (gmax == 0) CYCLE
               yz_mask(y, z) = 0
               pw_grid%para%yzq(y, z) = ip
            END IF
            IF (ip == pw_grid%para%my_pos) THEN
               pw_grid%ngpts_cut_local = pw_grid%ngpts_cut_local+gmax
            END IF
         END DO
      END DO

   END SUBROUTINE pre_tag

   !------------------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param yz_mask ...
!> \param yz_index ...
! **************************************************************************************************
   SUBROUTINE order_mask(yz_mask, yz_index)

      INTEGER, DIMENSION(:, :), INTENT(IN)               :: yz_mask
      INTEGER, DIMENSION(:, :), INTENT(OUT)              :: yz_index

      CHARACTER(len=*), PARAMETER :: routineN = 'order_mask', routineP = moduleN//':'//routineN

      INTEGER                                            :: i1, i2, ic, icount, ii, im, jc, jj

!NB load balance
!------------------------------------------------------------------------------
!NB spiral out from origin, so that even if overall grid is full and
!NB block distributed, spherical cutoff still leads to good load
!NB balance in cp_ddapc_apply_CD

      i1 = SIZE(yz_mask, 1)
      i2 = SIZE(yz_mask, 2)
      yz_index = 0

      icount = 1
      ic = i1/2
      jc = i2/2
      ii = ic
      jj = jc
      IF (ii > 0 .AND. ii <= i1 .AND. jj > 0 .AND. jj <= i2) THEN
         IF (yz_mask(ii, jj) /= 0) THEN
            yz_index(1, icount) = ii
            yz_index(2, icount) = jj
            icount = icount+1
         ENDIF
      ENDIF
      DO im = 1, MAX(ic+1, jc+1)
         ii = ic-im
         DO jj = jc-im, jc+im
            IF (ii > 0 .AND. ii <= i1 .AND. jj > 0 .AND. jj <= i2) THEN
               IF (yz_mask(ii, jj) /= 0) THEN
                  yz_index(1, icount) = ii
                  yz_index(2, icount) = jj
                  icount = icount+1
               ENDIF
            ENDIF
         END DO
         ii = ic+im
         DO jj = jc-im, jc+im
            IF (ii > 0 .AND. ii <= i1 .AND. jj > 0 .AND. jj <= i2) THEN
               IF (yz_mask(ii, jj) /= 0) THEN
                  yz_index(1, icount) = ii
                  yz_index(2, icount) = jj
                  icount = icount+1
               ENDIF
            ENDIF
         END DO
         jj = jc-im
         DO ii = ic-im+1, ic+im-1
            IF (ii > 0 .AND. ii <= i1 .AND. jj > 0 .AND. jj <= i2) THEN
               IF (yz_mask(ii, jj) /= 0) THEN
                  yz_index(1, icount) = ii
                  yz_index(2, icount) = jj
                  icount = icount+1
               ENDIF
            ENDIF
         END DO
         jj = jc+im
         DO ii = ic-im+1, ic+im-1
            IF (ii > 0 .AND. ii <= i1 .AND. jj > 0 .AND. jj <= i2) THEN
               IF (yz_mask(ii, jj) /= 0) THEN
                  yz_index(1, icount) = ii
                  yz_index(2, icount) = jj
                  icount = icount+1
               ENDIF
            ENDIF
         END DO
      END DO

   END SUBROUTINE order_mask
! **************************************************************************************************
!> \brief compute the length of g vectors
!> \param h_inv ...
!> \param length_x ...
!> \param length_y ...
!> \param length_z ...
!> \param length ...
!> \param l ...
!> \param m ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE pw_vec_length(h_inv, length_x, length_y, length_z, length, l, m, n)

      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: h_inv
      REAL(KIND=dp), INTENT(OUT)                         :: length_x, length_y, length_z, length
      INTEGER, INTENT(IN)                                :: l, m, n

      length_x &
         = REAL(l, dp)*h_inv(1, 1) &
           +REAL(m, dp)*h_inv(2, 1) &
           +REAL(n, dp)*h_inv(3, 1)
      length_y &
         = REAL(l, dp)*h_inv(1, 2) &
           +REAL(m, dp)*h_inv(2, 2) &
           +REAL(n, dp)*h_inv(3, 2)
      length_z &
         = REAL(l, dp)*h_inv(1, 3) &
           +REAL(m, dp)*h_inv(2, 3) &
           +REAL(n, dp)*h_inv(3, 3)

      ! enforce strict zero-ness in this case (compiler optimization)
      IF (l == 0 .AND. m == 0 .AND. n == 0) THEN
         length_x = 0
         length_y = 0
         length_z = 0
      ENDIF

      length_x = length_x*twopi
      length_y = length_y*twopi
      length_z = length_z*twopi

      length = length_x**2+length_y**2+length_z**2

   END SUBROUTINE

! **************************************************************************************************
!> \brief Count total number of g vectors
!> \param h_inv ...
!> \param pw_grid ...
!> \param cutoff ...
!> \param yz_mask ...
!> \par History
!>      JGH (22-12-2000) : Adapted for parallel use
!> \author apsi
!>      Christopher Mundy
! **************************************************************************************************
   SUBROUTINE pw_grid_count(h_inv, pw_grid, cutoff, yz_mask)

      REAL(KIND=dp), DIMENSION(3, 3)                     :: h_inv
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      REAL(KIND=dp), INTENT(IN)                          :: cutoff
      INTEGER, DIMENSION(:, :), INTENT(OUT)              :: yz_mask

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_count', routineP = moduleN//':'//routineN

      INTEGER                                            :: l, m, mm, n, n_upperlimit, nlim(2), nn
      INTEGER(KIND=int_8)                                :: gpt
      INTEGER, DIMENSION(:, :), POINTER                  :: bounds
      REAL(KIND=dp)                                      :: length, length_x, length_y, length_z

      bounds => pw_grid%bounds

      IF (pw_grid%grid_span == HALFSPACE) THEN
         n_upperlimit = 0
      ELSE IF (pw_grid%grid_span == FULLSPACE) THEN
         n_upperlimit = bounds(2, 3)
      ELSE
         CALL cp__b("pw/pw_grids.F",1421,"No type set for the grid")
      END IF

      ! finds valid g-points within grid
      gpt = 0
      IF (pw_grid%para%mode == PW_MODE_LOCAL) THEN
         nlim(1) = bounds(1, 3)
         nlim(2) = n_upperlimit
      ELSE IF (pw_grid%para%mode == PW_MODE_DISTRIBUTED) THEN
         n = n_upperlimit-bounds(1, 3)+1
         nlim = get_limit(n, pw_grid%para%group_size, pw_grid%para%my_pos)
         nlim = nlim+bounds(1, 3)-1
      ELSE
         CALL cp__b("pw/pw_grids.F",1434,"para % mode not specified")
      END IF

      yz_mask = 0
      DO n = nlim(1), nlim(2)
         nn = n-bounds(1, 3)+1
         DO m = bounds(1, 2), bounds(2, 2)
            mm = m-bounds(1, 2)+1
            DO l = bounds(1, 1), bounds(2, 1)
               IF (pw_grid%grid_span == HALFSPACE .AND. n == 0) THEN
                  IF ((m == 0 .AND. l > 0) .OR. (m > 0)) CYCLE
               END IF

               CALL pw_vec_length(h_inv, length_x, length_y, length_z, length, l, m, n)

               IF (0.5_dp*length <= cutoff) THEN
                  gpt = gpt+1
                  yz_mask(mm, nn) = yz_mask(mm, nn)+1
               END IF

            END DO
         END DO
      END DO

      ! number of g-vectors for grid
      IF (pw_grid%para%mode == PW_MODE_DISTRIBUTED) THEN
         CALL mp_sum(gpt, pw_grid%para%group)
         CALL mp_sum(yz_mask, pw_grid%para%group)
      ENDIF
      pw_grid%ngpts_cut = gpt

   END SUBROUTINE pw_grid_count

! **************************************************************************************************
!> \brief Setup maps from 1d to 3d space
!> \param h_inv ...
!> \param pw_grid ...
!> \param cutoff ...
!> \par History
!>      JGH (29-12-2000) : Adapted for parallel use
!> \author apsi
!>      Christopher Mundy
! **************************************************************************************************
   SUBROUTINE pw_grid_assign(h_inv, pw_grid, cutoff)

      REAL(KIND=dp), DIMENSION(3, 3)                     :: h_inv
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      REAL(KIND=dp), INTENT(IN)                          :: cutoff

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_assign', routineP = moduleN//':'//routineN

      INTEGER                                            :: gpt, handle, i, ip, l, lby, lbz, ll, m, &
                                                            mm, n, n_upperlimit, nn
      INTEGER(KIND=int_8)                                :: gpt_global
      INTEGER, DIMENSION(2, 3)                           :: bol, bounds
      REAL(KIND=dp)                                      :: length, length_x, length_y, length_z

      CALL timeset(routineN, handle)

      bounds = pw_grid%bounds
      lby = pw_grid%bounds(1, 2)
      lbz = pw_grid%bounds(1, 3)

      IF (pw_grid%grid_span == HALFSPACE) THEN
         n_upperlimit = 0
      ELSE IF (pw_grid%grid_span == FULLSPACE) THEN
         n_upperlimit = bounds(2, 3)
      ELSE
         CALL cp__b("pw/pw_grids.F",1502,"No type set for the grid")
      END IF

      ! finds valid g-points within grid
      IF (pw_grid%para%mode == PW_MODE_LOCAL) THEN
         gpt = 0
         DO n = bounds(1, 3), n_upperlimit
            DO m = bounds(1, 2), bounds(2, 2)
               DO l = bounds(1, 1), bounds(2, 1)
                  IF (pw_grid%grid_span == HALFSPACE .AND. n == 0) THEN
                     IF ((m == 0 .AND. l > 0) .OR. (m > 0)) CYCLE
                  END IF

                  CALL pw_vec_length(h_inv, length_x, length_y, length_z, length, l, m, n)

                  IF (0.5_dp*length <= cutoff) THEN
                     gpt = gpt+1
                     pw_grid%g(1, gpt) = length_x
                     pw_grid%g(2, gpt) = length_y
                     pw_grid%g(3, gpt) = length_z
                     pw_grid%gsq(gpt) = length
                     pw_grid%g_hat(1, gpt) = l
                     pw_grid%g_hat(2, gpt) = m
                     pw_grid%g_hat(3, gpt) = n
                  END IF

               END DO
            END DO
         END DO

      ELSE

         IF (pw_grid%para%ray_distribution) THEN

            gpt = 0
            ip = pw_grid%para%my_pos
            DO i = 1, pw_grid%para%nyzray(ip)
               n = pw_grid%para%yzp(2, i, ip)+lbz-1
               m = pw_grid%para%yzp(1, i, ip)+lby-1
               IF (n > n_upperlimit) CYCLE
               DO l = bounds(1, 1), bounds(2, 1)
                  IF (pw_grid%grid_span == HALFSPACE .AND. n == 0) THEN
                     IF ((m == 0 .AND. l > 0) .OR. (m > 0)) CYCLE
                  END IF

                  CALL pw_vec_length(h_inv, length_x, length_y, length_z, length, l, m, n)

                  IF (0.5_dp*length <= cutoff) THEN
                     gpt = gpt+1
                     pw_grid%g(1, gpt) = length_x
                     pw_grid%g(2, gpt) = length_y
                     pw_grid%g(3, gpt) = length_z
                     pw_grid%gsq(gpt) = length
                     pw_grid%g_hat(1, gpt) = l
                     pw_grid%g_hat(2, gpt) = m
                     pw_grid%g_hat(3, gpt) = n
                  END IF

               END DO
            END DO

         ELSE

            bol = pw_grid%para%bo(:, :, pw_grid%para%rs_mpo, 3)
            gpt = 0
            DO n = bounds(1, 3), bounds(2, 3)
               IF (n < 0) THEN
                  nn = n+pw_grid%npts(3)+1
               ELSE
                  nn = n+1
               END IF
               IF (nn < bol(1, 3) .OR. nn > bol(2, 3)) CYCLE
               DO m = bounds(1, 2), bounds(2, 2)
                  IF (m < 0) THEN
                     mm = m+pw_grid%npts(2)+1
                  ELSE
                     mm = m+1
                  END IF
                  IF (mm < bol(1, 2) .OR. mm > bol(2, 2)) CYCLE
                  DO l = bounds(1, 1), bounds(2, 1)
                     IF (l < 0) THEN
                        ll = l+pw_grid%npts(1)+1
                     ELSE
                        ll = l+1
                     END IF
                     IF (ll < bol(1, 1) .OR. ll > bol(2, 1)) CYCLE

                     CALL pw_vec_length(h_inv, length_x, length_y, length_z, length, l, m, n)

                     gpt = gpt+1
                     pw_grid%g(1, gpt) = length_x
                     pw_grid%g(2, gpt) = length_y
                     pw_grid%g(3, gpt) = length_z
                     pw_grid%gsq(gpt) = length
                     pw_grid%g_hat(1, gpt) = l
                     pw_grid%g_hat(2, gpt) = m
                     pw_grid%g_hat(3, gpt) = n

                  END DO
               END DO
            END DO

         END IF

      END IF

      ! Check the number of g-vectors for grid
      IF(.NOT.(pw_grid%ngpts_cut_local == gpt))CALL cp__a("pw/pw_grids.F",1609)
      IF (pw_grid%para%mode == PW_MODE_DISTRIBUTED) THEN
         gpt_global = gpt
         CALL mp_sum(gpt_global, pw_grid%para%group)
         IF(.NOT.(pw_grid%ngpts_cut == gpt_global))CALL cp__a("pw/pw_grids.F",1613)
      ENDIF

      pw_grid%have_g0 = .FALSE.
      pw_grid%first_gne0 = 1
      DO gpt = 1, pw_grid%ngpts_cut_local
         IF (ALL(pw_grid%g_hat(:, gpt) == 0)) THEN
            pw_grid%have_g0 = .TRUE.
            pw_grid%first_gne0 = 2
            EXIT
         END IF
      END DO

      CALL pw_grid_set_maps(pw_grid%grid_span, pw_grid%g_hat, &
                            pw_grid%mapl, pw_grid%mapm, pw_grid%mapn, pw_grid%npts)

      CALL timestop(handle)

   END SUBROUTINE pw_grid_assign

! **************************************************************************************************
!> \brief Setup maps from 1d to 3d space
!> \param grid_span ...
!> \param g_hat ...
!> \param mapl ...
!> \param mapm ...
!> \param mapn ...
!> \param npts ...
!> \par History
!>      JGH (21-12-2000) : Size of g_hat locally determined
!> \author apsi
!>      Christopher Mundy
!> \note
!>      Maps are to full 3D space (not distributed)
! **************************************************************************************************
   SUBROUTINE pw_grid_set_maps(grid_span, g_hat, mapl, mapm, mapn, npts)

      INTEGER, INTENT(IN)                                :: grid_span
      INTEGER, DIMENSION(:, :), INTENT(IN)               :: g_hat
      TYPE(map_pn), INTENT(INOUT)                        :: mapl, mapm, mapn
      INTEGER, DIMENSION(:), INTENT(IN)                  :: npts

      INTEGER                                            :: gpt, l, m, n, ng

      ng = SIZE(g_hat, 2)

      DO gpt = 1, ng
         l = g_hat(1, gpt)
         m = g_hat(2, gpt)
         n = g_hat(3, gpt)
         IF (l < 0) THEN
            mapl%pos(l) = l+npts(1)
         ELSE
            mapl%pos(l) = l
         END IF
         IF (m < 0) THEN
            mapm%pos(m) = m+npts(2)
         ELSE
            mapm%pos(m) = m
         END IF
         IF (n < 0) THEN
            mapn%pos(n) = n+npts(3)
         ELSE
            mapn%pos(n) = n
         END IF

         ! Generating the maps to the full 3-d space

         IF (grid_span == HALFSPACE) THEN

            IF (l <= 0) THEN
               mapl%neg(l) = -l
            ELSE
               mapl%neg(l) = npts(1)-l
            END IF
            IF (m <= 0) THEN
               mapm%neg(m) = -m
            ELSE
               mapm%neg(m) = npts(2)-m
            END IF
            IF (n <= 0) THEN
               mapn%neg(n) = -n
            ELSE
               mapn%neg(n) = npts(3)-n
            END IF

         END IF

      END DO

   END SUBROUTINE pw_grid_set_maps

! **************************************************************************************************
!> \brief Allocate all (Pointer) Arrays in pw_grid
!> \param pw_grid ...
!> \param ng ...
!> \param bounds ...
!> \par History
!>      JGH (20-12-2000) : Added status variable
!>                         Bounds of arrays now from calling routine, this
!>                         makes it independent from parallel setup
!> \author apsi
!>      Christopher Mundy
! **************************************************************************************************
   SUBROUTINE pw_grid_allocate(pw_grid, ng, bounds)

      ! Argument
      TYPE(pw_grid_type), INTENT(INOUT)        :: pw_grid
      INTEGER, INTENT(IN)                      :: ng
      INTEGER, DIMENSION(:, :), INTENT(IN)     :: bounds

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_allocate', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                    :: nmaps

      INTEGER                                  :: handle

      CALL timeset(routineN, handle)

      ALLOCATE (pw_grid%g(3, ng))
      ALLOCATE (pw_grid%gsq(ng))
      ALLOCATE (pw_grid%g_hat(3, ng))

      nmaps = 1
      IF (pw_grid%grid_span == HALFSPACE) nmaps = 2
      ALLOCATE (pw_grid%g_hatmap(1, 1))

      IF (pw_grid%para%mode == PW_MODE_DISTRIBUTED) THEN
         ALLOCATE (pw_grid%grays(pw_grid%npts(1), &
                                 pw_grid%para%nyzray(pw_grid%para%my_pos)))
      END IF

      ALLOCATE (pw_grid%mapl%pos(bounds(1, 1):bounds(2, 1)))
      ALLOCATE (pw_grid%mapl%neg(bounds(1, 1):bounds(2, 1)))
      ALLOCATE (pw_grid%mapm%pos(bounds(1, 2):bounds(2, 2)))
      ALLOCATE (pw_grid%mapm%neg(bounds(1, 2):bounds(2, 2)))
      ALLOCATE (pw_grid%mapn%pos(bounds(1, 3):bounds(2, 3)))
      ALLOCATE (pw_grid%mapn%neg(bounds(1, 3):bounds(2, 3)))

      CALL timestop(handle)

   END SUBROUTINE pw_grid_allocate

! **************************************************************************************************
!> \brief Sort g-vectors according to length
!> \param pw_grid ...
!> \param ref_grid ...
!> \par History
!>      JGH (20-12-2000) : allocate idx, ng = SIZE ( pw_grid % gsq ) the
!>                         sorting is local and independent from parallelisation
!>                         WARNING: Global ordering depends now on the number
!>                                  of cpus.
!>      JGH (28-02-2001) : check for ordering against reference grid
!>      JGH (01-05-2001) : sort spherical cutoff grids also within shells
!>                         reference grids for non-spherical cutoffs
!>      JGH (20-06-2001) : do not sort non-spherical grids
!>      JGH (19-02-2003) : Order all grids, this makes subgrids also for
!>                         non-spherical cutoffs possible
!>      JGH (21-02-2003) : Introduce gather array for general reference grids
!> \author apsi
!>      Christopher Mundy
! **************************************************************************************************
   SUBROUTINE pw_grid_sort(pw_grid, ref_grid)

      ! Argument
      TYPE(pw_grid_type), INTENT(INOUT)                  :: pw_grid
      TYPE(pw_grid_type), INTENT(IN), OPTIONAL           :: ref_grid

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_sort', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ig, ih, ip, is, it, ng, ngr
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: idx
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: int_tmp
      LOGICAL                                            :: g_found
      REAL(KIND=dp)                                      :: gig, gigr
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: real_tmp

      CALL timeset(routineN, handle)

      ng = SIZE(pw_grid%gsq)
      ALLOCATE (idx(ng))

      ! grids are (locally) ordered by length of G-vectors
      CALL sort(pw_grid%gsq, ng, idx)
      ! within shells order wrt x,y,z
      CALL sort_shells(pw_grid%gsq, pw_grid%g_hat, idx)

      ALLOCATE (real_tmp(3, ng))
      DO i = 1, ng
         real_tmp(1, i) = pw_grid%g(1, idx(i))
         real_tmp(2, i) = pw_grid%g(2, idx(i))
         real_tmp(3, i) = pw_grid%g(3, idx(i))
      ENDDO
      DO i = 1, ng
         pw_grid%g(1, i) = real_tmp(1, i)
         pw_grid%g(2, i) = real_tmp(2, i)
         pw_grid%g(3, i) = real_tmp(3, i)
      ENDDO
      DEALLOCATE (real_tmp)

      ALLOCATE (int_tmp(3, ng))
      DO i = 1, ng
         int_tmp(1, i) = pw_grid%g_hat(1, idx(i))
         int_tmp(2, i) = pw_grid%g_hat(2, idx(i))
         int_tmp(3, i) = pw_grid%g_hat(3, idx(i))
      ENDDO
      DO i = 1, ng
         pw_grid%g_hat(1, i) = int_tmp(1, i)
         pw_grid%g_hat(2, i) = int_tmp(2, i)
         pw_grid%g_hat(3, i) = int_tmp(3, i)
      ENDDO
      DEALLOCATE (int_tmp)

      DEALLOCATE (idx)

      ! check if ordering is compatible to reference grid
      IF (PRESENT(ref_grid)) THEN
         ngr = SIZE(ref_grid%gsq)
         ngr = MIN(ng, ngr)
         IF (pw_grid%spherical) THEN
            IF (.NOT. ALL(pw_grid%g_hat(1:3, 1:ngr) &
                          == ref_grid%g_hat(1:3, 1:ngr))) THEN
               CALL cp__b("pw/pw_grids.F",1851,"G space sorting not compatible")
            END IF
         ELSE
            ALLOCATE (pw_grid%gidx(1:ngr))
            pw_grid%gidx = 0
            ! first try as many trivial associations as possible
            it = 0
            DO ig = 1, ngr
               IF (.NOT. ALL(pw_grid%g_hat(1:3, ig) &
                             == ref_grid%g_hat(1:3, ig))) EXIT
               pw_grid%gidx(ig) = ig
               it = ig
            END DO
            ! now for the ones that could not be done
            IF (ng == ngr) THEN
               ! for the case pw_grid <= ref_grid
               is = it
               DO ig = it+1, ngr
                  gig = pw_grid%gsq(ig)
                  gigr = MAX(1._dp, gig)
                  g_found = .FALSE.
                  DO ih = is+1, SIZE(ref_grid%gsq)
                     IF (ABS(gig-ref_grid%gsq(ih))/gigr > 1.e-12_dp) CYCLE
                     g_found = .TRUE.
                     EXIT
                  END DO
                  IF (.NOT. g_found) THEN
                     WRITE (*, "(A,I10,F20.10)") "G-vector", ig, pw_grid%gsq(ig)
                     CALL cp__b("pw/pw_grids.F",1879,"G vector not found")
                  END IF
                  ip = ih-1
                  DO ih = ip+1, SIZE(ref_grid%gsq)
                     IF (ABS(gig-ref_grid%gsq(ih))/gigr > 1.e-12_dp) CYCLE
                     IF (pw_grid%g_hat(1, ig) /= ref_grid%g_hat(1, ih)) CYCLE
                     IF (pw_grid%g_hat(2, ig) /= ref_grid%g_hat(2, ih)) CYCLE
                     IF (pw_grid%g_hat(3, ig) /= ref_grid%g_hat(3, ih)) CYCLE
                     pw_grid%gidx(ig) = ih
                     EXIT
                  END DO
                  IF (pw_grid%gidx(ig) == 0) THEN
                     WRITE (*, "(A,2I10)") " G-Shell ", is+1, ip+1
                     WRITE (*, "(A,I10,3I6,F20.10)") &
                        " G-vector", ig, pw_grid%g_hat(1:3, ig), pw_grid%gsq(ig)
                     DO ih = 1, SIZE(ref_grid%gsq)
                        IF (pw_grid%g_hat(1, ig) /= ref_grid%g_hat(1, ih)) CYCLE
                        IF (pw_grid%g_hat(2, ig) /= ref_grid%g_hat(2, ih)) CYCLE
                        IF (pw_grid%g_hat(3, ig) /= ref_grid%g_hat(3, ih)) CYCLE
                        WRITE (*, "(A,I10,3I6,F20.10)") &
                           " G-vector", ih, ref_grid%g_hat(1:3, ih), ref_grid%gsq(ih)
                     END DO
                     CALL cp__b("pw/pw_grids.F",1901,"G vector not found")
                  END IF
                  is = ip
               END DO
            ELSE
               ! for the case pw_grid > ref_grid
               is = it
               DO ig = it+1, ngr
                  gig = ref_grid%gsq(ig)
                  gigr = MAX(1._dp, gig)
                  g_found = .FALSE.
                  DO ih = is+1, ng
                     IF (ABS(pw_grid%gsq(ih)-gig)/gigr > 1.e-12_dp) CYCLE
                     g_found = .TRUE.
                     EXIT
                  END DO
                  IF (.NOT. g_found) THEN
                     WRITE (*, "(A,I10,F20.10)") "G-vector", ig, ref_grid%gsq(ig)
                     CALL cp__b("pw/pw_grids.F",1919,"G vector not found")
                  END IF
                  ip = ih-1
                  DO ih = ip+1, ng
                     IF (ABS(pw_grid%gsq(ih)-gig)/gigr > 1.e-12_dp) CYCLE
                     IF (pw_grid%g_hat(1, ih) /= ref_grid%g_hat(1, ig)) CYCLE
                     IF (pw_grid%g_hat(2, ih) /= ref_grid%g_hat(2, ig)) CYCLE
                     IF (pw_grid%g_hat(3, ih) /= ref_grid%g_hat(3, ig)) CYCLE
                     pw_grid%gidx(ig) = ih
                     EXIT
                  END DO
                  IF (pw_grid%gidx(ig) == 0) THEN
                     WRITE (*, "(A,2I10)") " G-Shell ", is+1, ip+1
                     WRITE (*, "(A,I10,3I6,F20.10)") &
                        " G-vector", ig, ref_grid%g_hat(1:3, ig), ref_grid%gsq(ig)
                     DO ih = 1, ng
                        IF (pw_grid%g_hat(1, ih) /= ref_grid%g_hat(1, ig)) CYCLE
                        IF (pw_grid%g_hat(2, ih) /= ref_grid%g_hat(2, ig)) CYCLE
                        IF (pw_grid%g_hat(3, ih) /= ref_grid%g_hat(3, ig)) CYCLE
                        WRITE (*, "(A,I10,3I6,F20.10)") &
                           " G-vector", ih, pw_grid%g_hat(1:3, ih), pw_grid%gsq(ih)
                     END DO
                     CALL cp__b("pw/pw_grids.F",1941,"G vector not found")
                  END IF
                  is = ip
               END DO
            END IF
            ! test if all g-vectors are associated
            IF (ANY(pw_grid%gidx == 0)) THEN
               CALL cp__b("pw/pw_grids.F",1948,"G space sorting not compatible")
            END IF
         END IF
      END IF

      !check if G=0 is at first position
      IF (pw_grid%have_g0) THEN
         IF (pw_grid%gsq(1) /= 0.0_dp) THEN
            CALL cp__b("pw/pw_grids.F",1956,"G = 0 not in first position")
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE pw_grid_sort

! **************************************************************************************************
!> \brief ...
!> \param gsq ...
!> \param g_hat ...
!> \param idx ...
! **************************************************************************************************
   SUBROUTINE sort_shells(gsq, g_hat, idx)

      ! Argument
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: gsq
      INTEGER, DIMENSION(:, :), INTENT(IN)               :: g_hat
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: idx

      CHARACTER(len=*), PARAMETER :: routineN = 'sort_shells', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: small = 5.e-16_dp

      INTEGER                                            :: handle, ig, ng, s1, s2
      REAL(KIND=dp)                                      :: s_begin

      CALL timeset(routineN, handle)

! Juergs temporary hack to get the grids sorted for large (4000Ry) cutoffs.
! might need to call lapack for machine precision.

      ng = SIZE(gsq)
      s_begin = -1.0_dp
      s1 = 0
      s2 = 0
      ig = 1
      DO ig = 1, ng
         IF (ABS(gsq(ig)-s_begin) < small) THEN
            s2 = ig
         ELSE
            CALL redist(g_hat, idx, s1, s2)
            s_begin = gsq(ig)
            s1 = ig
            s2 = ig
         END IF
      END DO
      CALL redist(g_hat, idx, s1, s2)

      CALL timestop(handle)

   END SUBROUTINE sort_shells

! **************************************************************************************************
!> \brief ...
!> \param g_hat ...
!> \param idx ...
!> \param s1 ...
!> \param s2 ...
! **************************************************************************************************
   SUBROUTINE redist(g_hat, idx, s1, s2)

      ! Argument
      INTEGER, DIMENSION(:, :), INTENT(IN)               :: g_hat
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: idx
      INTEGER, INTENT(IN)                                :: s1, s2

      CHARACTER(len=*), PARAMETER :: routineN = 'redist', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ii, n1, n2, n3, ns
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: indl
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: slen

      IF (s2 <= s1) RETURN
      ns = s2-s1+1
      ALLOCATE (indl(ns))
      ALLOCATE (slen(ns))

      DO i = s1, s2
         ii = idx(i)
         n1 = g_hat(1, ii)
         n2 = g_hat(2, ii)
         n3 = g_hat(3, ii)
         slen(i-s1+1) = 1000.0_dp*REAL(n1, dp)+ &
                        REAL(n2, dp)+0.001_dp*REAL(n3, dp)
      END DO
      CALL sort(slen, ns, indl)
      DO i = 1, ns
         ii = indl(i)+s1-1
         indl(i) = idx(ii)
      END DO
      idx(s1:s2) = indl(1:ns)

      DEALLOCATE (indl)
      DEALLOCATE (slen)

   END SUBROUTINE redist

! **************************************************************************************************
!> \brief Reorder yzq and yzp arrays for parallel FFT according to FFT mapping
!> \param pw_grid ...
!> \param yz ...
!> \par History
!>      none
!> \author JGH (17-Jan-2001)
! **************************************************************************************************
   SUBROUTINE pw_grid_remap(pw_grid, yz)

      ! Argument
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      INTEGER, DIMENSION(:, :), INTENT(OUT)              :: yz

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_remap', routineP = moduleN//':'//routineN

      INTEGER                                            :: gpt, handle, i, ip, is, j, m, n, ny, nz
      INTEGER, DIMENSION(:), POINTER                     :: mapm, mapn

      IF (pw_grid%para%mode == PW_MODE_LOCAL) RETURN

      CALL timeset(routineN, handle)

      ny = pw_grid%npts(2)
      nz = pw_grid%npts(3)

      yz = 0
      pw_grid%para%yzp = 0
      pw_grid%para%yzq = 0

      mapm => pw_grid%mapm%pos
      mapn => pw_grid%mapn%pos

      DO gpt = 1, SIZE(pw_grid%gsq)
         m = mapm(pw_grid%g_hat(2, gpt))+1
         n = mapn(pw_grid%g_hat(3, gpt))+1
         yz(m, n) = yz(m, n)+1
      END DO
      IF (pw_grid%grid_span == HALFSPACE) THEN
         mapm => pw_grid%mapm%neg
         mapn => pw_grid%mapn%neg
         DO gpt = 1, SIZE(pw_grid%gsq)
            m = mapm(pw_grid%g_hat(2, gpt))+1
            n = mapn(pw_grid%g_hat(3, gpt))+1
            yz(m, n) = yz(m, n)+1
         END DO
      END IF

      ip = pw_grid%para%my_pos
      is = 0
      DO i = 1, nz
         DO j = 1, ny
            IF (yz(j, i) > 0) THEN
               is = is+1
               pw_grid%para%yzp(1, is, ip) = j
               pw_grid%para%yzp(2, is, ip) = i
               pw_grid%para%yzq(j, i) = is
            END IF
         END DO
      END DO

      IF(.NOT.(is == pw_grid%para%nyzray(ip)))CALL cp__a("pw/pw_grids.F",2115)
      CALL mp_sum(pw_grid%para%yzp, pw_grid%para%group)

      CALL timestop(handle)

   END SUBROUTINE pw_grid_remap

! **************************************************************************************************
!> \brief Recalculate the g-vectors after a change of the box
!> \param cell_hmat ...
!> \param pw_grid ...
!> \par History
!>      JGH (20-12-2000) : get local grid size from definition of g.
!>                         Assume that gsq is allocated.
!>                         Local routine, no information on distribution of
!>                         PW required.
!>      JGH (8-Mar-2001) : also update information on volume and grid spaceing
!> \author apsi
!>      Christopher Mundy
! **************************************************************************************************
   SUBROUTINE pw_grid_change(cell_hmat, pw_grid)
      ! Argument
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: cell_hmat
      TYPE(pw_grid_type), POINTER                        :: pw_grid

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_change', routineP = moduleN//':'//routineN

      INTEGER                                            :: gpt
      REAL(KIND=dp)                                      :: cell_deth, l, m, n
      REAL(KIND=dp), DIMENSION(3, 3)                     :: cell_h_inv
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: g

      cell_deth = ABS(det_3x3(cell_hmat))
      IF (cell_deth < 1.0E-10_dp) THEN
         CALL cp_abort(cp__l("pw/pw_grids.F",2149), &
                       "An invalid set of cell vectors was specified. "// &
                       "The determinant det(h) is too small")
      END IF
      cell_h_inv = inv_3x3(cell_hmat)

      g => pw_grid%g

      CALL cell2grid(cell_hmat, cell_h_inv, cell_deth, pw_grid)

      DO gpt = 1, SIZE(g, 2)

         l = twopi*REAL(pw_grid%g_hat(1, gpt), KIND=dp)
         m = twopi*REAL(pw_grid%g_hat(2, gpt), KIND=dp)
         n = twopi*REAL(pw_grid%g_hat(3, gpt), KIND=dp)

         g(1, gpt) = l*cell_h_inv(1, 1)+m*cell_h_inv(2, 1)+n*cell_h_inv(3, 1)
         g(2, gpt) = l*cell_h_inv(1, 2)+m*cell_h_inv(2, 2)+n*cell_h_inv(3, 2)
         g(3, gpt) = l*cell_h_inv(1, 3)+m*cell_h_inv(2, 3)+n*cell_h_inv(3, 3)

         pw_grid%gsq(gpt) = g(1, gpt)*g(1, gpt) &
                            +g(2, gpt)*g(2, gpt) &
                            +g(3, gpt)*g(3, gpt)

      END DO

   END SUBROUTINE pw_grid_change

! **************************************************************************************************
!> \brief retains the given pw grid
!> \param pw_grid the pw grid to retain
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
!> \note
!>      see doc/ReferenceCounting.html
! **************************************************************************************************
   SUBROUTINE pw_grid_retain(pw_grid)
      TYPE(pw_grid_type), POINTER                        :: pw_grid

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(pw_grid)))CALL cp__a("pw/pw_grids.F",2191)
      IF(.NOT.(pw_grid%ref_count > 0))CALL cp__a("pw/pw_grids.F",2192)
      pw_grid%ref_count = pw_grid%ref_count+1
   END SUBROUTINE pw_grid_retain

! **************************************************************************************************
!> \brief releases the given pw grid
!> \param pw_grid the pw grid to release
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
!> \note
!>      see doc/ReferenceCounting.html
! **************************************************************************************************
   SUBROUTINE pw_grid_release(pw_grid)

      TYPE(pw_grid_type), POINTER              :: pw_grid

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_release', &
                                     routineP = moduleN//':'//routineN


      IF (ASSOCIATED(pw_grid)) THEN
         IF(.NOT.(pw_grid%ref_count > 0))CALL cp__a("pw/pw_grids.F",2218)
         pw_grid%ref_count = pw_grid%ref_count-1
         IF (pw_grid%ref_count == 0) THEN
            IF (ASSOCIATED(pw_grid%gidx)) THEN
               DEALLOCATE (pw_grid%gidx)
            END IF
            IF (ASSOCIATED(pw_grid%g)) THEN
               DEALLOCATE (pw_grid%g)
            END IF
            IF (ASSOCIATED(pw_grid%gsq)) THEN
               DEALLOCATE (pw_grid%gsq)
            END IF
            IF (ASSOCIATED(pw_grid%g_hat)) THEN
               DEALLOCATE (pw_grid%g_hat)
            END IF
            IF (ASSOCIATED(pw_grid%g_hatmap)) THEN
               DEALLOCATE (pw_grid%g_hatmap)
            END IF
            IF (ASSOCIATED(pw_grid%grays)) THEN
               DEALLOCATE (pw_grid%grays)
            END IF
            IF (ASSOCIATED(pw_grid%mapl%pos)) THEN
               DEALLOCATE (pw_grid%mapl%pos)
            END IF
            IF (ASSOCIATED(pw_grid%mapm%pos)) THEN
               DEALLOCATE (pw_grid%mapm%pos)
            END IF
            IF (ASSOCIATED(pw_grid%mapn%pos)) THEN
               DEALLOCATE (pw_grid%mapn%pos)
            END IF
            IF (ASSOCIATED(pw_grid%mapl%neg)) THEN
               DEALLOCATE (pw_grid%mapl%neg)
            END IF
            IF (ASSOCIATED(pw_grid%mapm%neg)) THEN
               DEALLOCATE (pw_grid%mapm%neg)
            END IF
            IF (ASSOCIATED(pw_grid%mapn%neg)) THEN
               DEALLOCATE (pw_grid%mapn%neg)
            END IF
            IF (ASSOCIATED(pw_grid%para%bo)) THEN
               DEALLOCATE (pw_grid%para%bo)
            END IF
            IF (pw_grid%para%mode == PW_MODE_DISTRIBUTED) THEN
               IF (ASSOCIATED(pw_grid%para%yzp)) THEN
                  DEALLOCATE (pw_grid%para%yzp)
               END IF
               IF (ASSOCIATED(pw_grid%para%yzq)) THEN
                  DEALLOCATE (pw_grid%para%yzq)
               END IF
               IF (ASSOCIATED(pw_grid%para%nyzray)) THEN
                  DEALLOCATE (pw_grid%para%nyzray)
               END IF
            END IF
            ! also release groups
            CALL mp_comm_free(pw_grid%para%group)
            IF (PRODUCT(pw_grid%para%rs_dims) /= 0) &
               CALL mp_comm_free(pw_grid%para%rs_group)
            IF (ASSOCIATED(pw_grid%para%pos_of_x)) THEN
               DEALLOCATE (pw_grid%para%pos_of_x)
            END IF

            IF (ASSOCIATED(pw_grid)) THEN
               DEALLOCATE (pw_grid)
            END IF
         END IF
      END IF
      NULLIFY (pw_grid)
   END SUBROUTINE pw_grid_release

END MODULE pw_grids

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \note
!>      If parallel mode is distributed certain combination of
!>      "in_use" and "in_space" can not be used.
!>      For performance reasons it would be better to have the loops
!>      over g-vectros in the gather/scatter routines in new subprograms
!>      with the actual arrays (also the adressing) in the parameter list
!> \par History
!>      JGH (29-Dec-2000) : Changes for parallel use
!>      JGH (13-Mar-2001) : added timing calls
!>      JGH (26-Feb-2003) : OpenMP enabled
!>      JGH (17-Nov-2007) : Removed mass arrays
!>      JGH (01-Dec-2007) : Removed and renamed routines
!>      03.2008 [tlaino] : Splitting pw_types into pw_types and pw_methods
!> \author apsi
! **************************************************************************************************
MODULE pw_types

   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type,&
                                              cp_to_string
   USE kinds,                           ONLY: dp
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_grids,                        ONLY: pw_grid_release,&
                                              pw_grid_retain
   USE timings,                         ONLY: print_stack
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: pw_type, pw_p_type
   PUBLIC :: pw_retain, pw_release, pw_create

! **************************************************************************************************
   TYPE pw_type
      REAL(KIND=dp), DIMENSION(:), POINTER :: cr
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: cr3d
      COMPLEX(KIND=dp), DIMENSION(:), POINTER :: cc
      COMPLEX(KIND=dp), DIMENSION(:, :, :), POINTER :: cc3d

      INTEGER :: in_use ! Which data is used [r1d/c1d/r3d/c3d]
      INTEGER :: in_space ! Real/Reciprocal space
      INTEGER :: id_nr ! unique identifier
      INTEGER :: ref_count ! reference count

      TYPE(pw_grid_type), POINTER :: pw_grid
   END TYPE pw_type

! **************************************************************************************************
   TYPE pw_p_type
      TYPE(pw_type), POINTER :: pw
   END TYPE pw_p_type

   ! Flags for the structure member 'in_use'
   INTEGER, PARAMETER, PUBLIC :: REALDATA1D = 301, COMPLEXDATA1D = 302
   INTEGER, PARAMETER, PUBLIC :: REALDATA3D = 303, COMPLEXDATA3D = 304, NODATA = 305

   ! Flags for the structure member 'in_space'
   INTEGER, PARAMETER, PUBLIC :: NOSPACE = 371, REALSPACE = 372, RECIPROCALSPACE = 373
   INTEGER, PUBLIC, PARAMETER :: SQUARE = 391, SQUAREROOT = 392

   ! to generate unique id_nr
   INTEGER, SAVE, PRIVATE :: last_pw_id_nr = 0
   INTEGER, SAVE, PRIVATE :: allocated_pw_count = 0

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pw_types'
   LOGICAL, PARAMETER, PRIVATE :: debug_this_module = .FALSE.

CONTAINS

! **************************************************************************************************
!> \brief retains a pw type
!> \param pw the pw to retain
!> \par History
!>      03.2003 created [fawzi]
!> \author fawzi
!> \note
!>      see doc/ReferenceCounting.html
! **************************************************************************************************
   SUBROUTINE pw_retain(pw)
      TYPE(pw_type), POINTER                             :: pw

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(pw)))CALL cp__a("pw/pw_types.F",92)
      IF(.NOT.(pw%ref_count > 0))CALL cp__a("pw/pw_types.F",93)
      pw%ref_count = pw%ref_count+1
   END SUBROUTINE pw_retain

! **************************************************************************************************
!> \brief releases the given pw
!> \param pw the pw to release
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
!> \note
!>      see doc/ReferenceCounting.html
! **************************************************************************************************
   SUBROUTINE pw_release(pw)
      TYPE(pw_type), POINTER                             :: pw

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(pw)) THEN
         IF(.NOT.(pw%ref_count > 0))CALL cp__a("pw/pw_types.F",112)
         pw%ref_count = pw%ref_count-1
         IF (pw%ref_count == 0) THEN
            pw%ref_count = 1

            allocated_pw_count = allocated_pw_count-1
            SELECT CASE (pw%in_use)
            CASE (REALDATA1D)
               DEALLOCATE (pw%cr)
            CASE (COMPLEXDATA1D)
               DEALLOCATE (pw%cc)
            CASE (REALDATA3D)
               IF (ASSOCIATED(pw%cr3d)) THEN
                  !FM optimizations of pools might have removed the 3d field to cache it
                  DEALLOCATE (pw%cr3d)
               END IF
            CASE (COMPLEXDATA3D)
               DEALLOCATE (pw%cc3d)
            CASE (NODATA)
            CASE default
               CALL cp__b("pw/pw_types.F",132,"unknown data type "//cp_to_string(pw%in_use))
            END SELECT
            CALL pw_grid_release(pw%pw_grid)
            pw%ref_count = 0
            DEALLOCATE (pw)
         END IF
      END IF
      NULLIFY (pw)
   END SUBROUTINE pw_release

! **************************************************************************************************
!> \brief allocates and initializes pw_type
!> \param pw the type that will bw allocated and initialized
!> \param pw_grid ...
!> \param use_data which kind of data will be used
!> \param in_space in which space the pw is (real or reciprocal)
!> \param cr3d_ptr pointer with the cr3d data (make sense only if
!>        use_data==REALDATA3D)
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pw_create(pw, pw_grid, use_data, in_space, cr3d_ptr)
      TYPE(pw_type), POINTER                             :: pw
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      INTEGER, INTENT(in)                                :: use_data
      INTEGER, INTENT(in), OPTIONAL                      :: in_space
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: cr3d_ptr

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      INTEGER, DIMENSION(:, :), POINTER                  :: bounds
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      IF(.NOT.(.NOT. ASSOCIATED(pw)))CALL cp__a("pw/pw_types.F",169)
      ALLOCATE (pw)
      logger => cp_get_default_logger()
      IF (debug_this_module) THEN
         WRITE (cp_logger_get_default_unit_nr(logger), "('*** allocated pw ***')")
         IF (PRESENT(cr3d_ptr)) THEN
            IF (ASSOCIATED(cr3d_ptr)) THEN
               WRITE (cp_logger_get_default_unit_nr(logger), "('*** cr3d associated ***')")
            END IF
         END IF
         CALL print_stack(cp_logger_get_default_unit_nr(logger))
      END IF

      IF (PRESENT(cr3d_ptr)) THEN
         IF (ASSOCIATED(cr3d_ptr)) THEN
            IF(.NOT.(use_data == REALDATA3D))CALL cp__a("pw/pw_types.F",184)
         END IF
      END IF

      last_pw_id_nr = last_pw_id_nr+1
      pw%id_nr = last_pw_id_nr
      pw%ref_count = 1
      NULLIFY (pw%pw_grid)
      pw%in_use = use_data
      pw%pw_grid => pw_grid
      CALL pw_grid_retain(pw%pw_grid)
      pw%in_space = NOSPACE
      bounds => pw%pw_grid%bounds_local

      allocated_pw_count = allocated_pw_count+1

      NULLIFY (pw%cr, pw%cc, pw%cr3d, pw%cc3d)

      SELECT CASE (use_data)
      CASE (REALDATA1D)
         ALLOCATE (pw%cr(pw%pw_grid%ngpts_cut_local))

      CASE (COMPLEXDATA1D)
         ALLOCATE (pw%cc(pw%pw_grid%ngpts_cut_local))

      CASE (REALDATA3D)
         IF (PRESENT(cr3d_ptr)) THEN
            IF (ASSOCIATED(cr3d_ptr)) THEN
               IF (ALL(bounds(1, :) <= bounds(2, :))) THEN
                  IF(.NOT.(ALL(LBOUND(cr3d_ptr) == bounds(1, :))))CALL cp__a("pw/pw_types.F",213)
                  IF(.NOT.(ALL(UBOUND(cr3d_ptr) == bounds(2, :))))CALL cp__a("pw/pw_types.F",214)
               END IF
               pw%cr3d => cr3d_ptr
            END IF
         END IF
         IF (.NOT. ASSOCIATED(pw%cr3d)) THEN
            ALLOCATE (pw%cr3d( &
                      bounds(1, 1):bounds(2, 1), &
                      bounds(1, 2):bounds(2, 2), &
                      bounds(1, 3):bounds(2, 3)))
         END IF

      CASE (COMPLEXDATA3D)
         ALLOCATE (pw%cc3d( &
                   bounds(1, 1):bounds(2, 1), &
                   bounds(1, 2):bounds(2, 2), &
                   bounds(1, 3):bounds(2, 3)))
      CASE (NODATA)
      CASE default
         CALL cp__b("pw/pw_types.F",233,"unknown data type")
      END SELECT
      IF (PRESENT(in_space)) pw%in_space = in_space
      CALL timestop(handle)
   END SUBROUTINE pw_create

END MODULE pw_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utility routines for the memory handling.
!>        Uses stop_memory to handle exceptions
!> \par History
!>      none
!> \author Matthias Krack (25.06.1999)
! **************************************************************************************************
MODULE memory_utilities

   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp,&
                                              dp_size,&
                                              int_8,&
                                              int_8_size,&
                                              int_size
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'memory_utilities'

   PUBLIC :: reallocate

   INTERFACE reallocate
      MODULE PROCEDURE reallocate_c1, reallocate_c2, reallocate_c3, reallocate_c4, &
         reallocate_i1, reallocate_i2, reallocate_i3, reallocate_i4, &
         reallocate_r1, reallocate_r2, reallocate_r3, reallocate_r4, &
         reallocate_r5, reallocate_s1, reallocate_l1, reallocate_8i1, &
         reallocate_8i2
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief (Re)Allocate a complex vector with a new dimension
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \par History
!>      none
!> \author Matthias Krack (28.11.2005,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_c1(p, lb1_new, ub1_new)

      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_c1', routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: zero = (0.0_dp, 0.0_dp)
      INTEGER, PARAMETER                                 :: t_size = 2*dp_size

      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: work

    INTEGER                                  :: istat, lb1, lb1_old, &
                                                ub1,  ub1_old
    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       ALLOCATE (work(lb1:ub1),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,10,&
                           "work",t_size*(ub1-lb1+1))
       END IF
       work(lb1:ub1) = p(lb1:ub1)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new),STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,19,&
                        "p",t_size*(ub1_new-lb1_new+1))
    END IF
    p(:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1) = work(lb1:ub1)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_c1

! **************************************************************************************************
!> \brief (Re)Allocate a complex array 2D with a new dimension
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \par History
!>      none
!> \author Matthias Krack (28.11.2005,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_c2(p, lb1_new, ub1_new, lb2_new, ub2_new)

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_c2', routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: zero = (0.0_dp, 0.0_dp)
      INTEGER, PARAMETER                                 :: t_size = 2*dp_size

      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :)     :: work

    INTEGER                                  :: istat, lb1, lb1_old, lb2, &
                                                lb2_old, ub1, ub1_old, ub2, ub2_old

    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb2_old = LBOUND(p,2)
       ub2_old = UBOUND(p,2)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       lb2 = MAX(lb2_new,lb2_old)
       ub2 = MIN(ub2_new,ub2_old)
       ALLOCATE (work(lb1:ub1,lb2:ub2),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,15,&
                           "work",t_size*(ub1-lb1+1)*&
                                         (ub2-lb2+1))

       END IF
       work(lb1:ub1,lb2:ub2) = p(lb1:ub1,lb2:ub2)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new),STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,26,&
                        "p",t_size*(ub1_new-lb1_new+1)*&
                                   (ub2_new-lb2_new+1))
    END IF
    p(:,:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1,lb2:ub2) = work(lb1:ub1,lb2:ub2)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_c2

! **************************************************************************************************
!> \brief (Re)Allocate a complex array 3D with a new dimension
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \param lb3_new ...
!> \param ub3_new ...
!> \par History
!>      none
!> \author Matthias Krack (28.11.2005,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_c3(p, lb1_new, ub1_new, lb2_new, ub2_new, lb3_new, ub3_new)

      COMPLEX(KIND=dp), DIMENSION(:, :, :), POINTER      :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new, &
                                                            lb3_new, ub3_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_c3', routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: zero = (0.0_dp, 0.0_dp)
      INTEGER, PARAMETER                                 :: t_size = 2*dp_size

      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)  :: work

    INTEGER :: istat, lb1, lb1_old, lb2, lb2_old, lb3, lb3_old, &
      ub1, ub1_old, ub2, ub2_old, ub3, ub3_old

    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb2_old = LBOUND(p,2)
       ub2_old = UBOUND(p,2)
       lb3_old = LBOUND(p,3)
       ub3_old = UBOUND(p,3)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       lb2 = MAX(lb2_new,lb2_old)
       ub2 = MIN(ub2_new,ub2_old)
       lb3 = MAX(lb3_new,lb3_old)
       ub3 = MIN(ub3_new,ub3_old)
       ALLOCATE (work(lb1:ub1,lb2:ub2,lb3:ub3),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,19,&
                           "work",t_size*(ub1-lb1+1)*&
                                         (ub2-lb2+1)*&
                                         (ub3-lb3+1))
       END IF
       work(lb1:ub1,lb2:ub2,lb3:ub3) = p(lb1:ub1,lb2:ub2,lb3:ub3)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new,lb3_new:ub3_new),STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,30,&
                        "p",t_size*(ub1_new-lb1_new+1)*&
                                   (ub2_new-lb2_new+1)*&
                                   (ub3_new-lb3_new+1))
    END IF
    p(:,:,:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1,lb2:ub2,lb3:ub3) = work(lb1:ub1,lb2:ub2,lb3:ub3)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_c3

! **************************************************************************************************
!> \brief (Re)Allocate a complex array 4D with a new dimension
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \param lb3_new ...
!> \param ub3_new ...
!> \param lb4_new ...
!> \param ub4_new ...
!> \par History
!>      none
!> \author Matthias Krack (28.11.2005,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_c4(p, lb1_new, ub1_new, lb2_new, ub2_new, lb3_new, ub3_new, &
                            lb4_new, ub4_new)

      COMPLEX(KIND=dp), DIMENSION(:, :, :, :), POINTER   :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new, &
                                                            lb3_new, ub3_new, lb4_new, ub4_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_c4', routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: zero = (0.0_dp, 0.0_dp)
      INTEGER, PARAMETER                                 :: t_size = 2*dp_size

      COMPLEX(KIND=dp), ALLOCATABLE, &
         DIMENSION(:, :, :, :)                           :: work

    INTEGER :: istat, lb1, lb1_old, lb2, lb2_old, lb3, lb3_old, lb4, lb4_old, &
      ub1, ub1_old, ub2, ub2_old, ub3, ub3_old, ub4, ub4_old

    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb2_old = LBOUND(p,2)
       ub2_old = UBOUND(p,2)
       lb3_old = LBOUND(p,3)
       ub3_old = UBOUND(p,3)
       lb4_old = LBOUND(p,4)
       ub4_old = UBOUND(p,4)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       lb2 = MAX(lb2_new,lb2_old)
       ub2 = MIN(ub2_new,ub2_old)
       lb3 = MAX(lb3_new,lb3_old)
       ub3 = MIN(ub3_new,ub3_old)
       lb4 = MAX(lb4_new,lb4_old)
       ub4 = MIN(ub4_new,ub4_old)
       ALLOCATE (work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,23,&
                           "work",t_size*(ub1-lb1+1)*&
                                         (ub2-lb2+1)*&
                                         (ub3-lb3+1)*&
                                         (ub4-lb4+1))
       END IF
       work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4) = p(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new,lb3_new:ub3_new,lb4_new:ub4_new),&
              STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,36,&
                        "p",t_size*(ub1_new-lb1_new+1)*&
                                   (ub2_new-lb2_new+1)*&
                                   (ub3_new-lb3_new+1)*&
                                   (ub4_new-lb4_new+1))
    END IF
    p(:,:,:,:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4) = work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_c4

! **************************************************************************************************
!> \brief (Re)Allocate an integer vector with a new dimension.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_i1(p, lb1_new, ub1_new)

      INTEGER, DIMENSION(:), POINTER                     :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_i1', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: t_size = int_size, zero = 0

      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: work

    INTEGER                                  :: istat, lb1, lb1_old, &
                                                ub1,  ub1_old
    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       ALLOCATE (work(lb1:ub1),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,10,&
                           "work",t_size*(ub1-lb1+1))
       END IF
       work(lb1:ub1) = p(lb1:ub1)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new),STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,19,&
                        "p",t_size*(ub1_new-lb1_new+1))
    END IF
    p(:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1) = work(lb1:ub1)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_i1

! **************************************************************************************************
!> \brief (Re)Allocate an integer array with a new dimension.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_i2(p, lb1_new, ub1_new, lb2_new, ub2_new)

      INTEGER, DIMENSION(:, :), POINTER                  :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_i2', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: t_size = int_size, zero = 0

      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: work

    INTEGER                                  :: istat, lb1, lb1_old, lb2, &
                                                lb2_old, ub1, ub1_old, ub2, ub2_old

    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb2_old = LBOUND(p,2)
       ub2_old = UBOUND(p,2)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       lb2 = MAX(lb2_new,lb2_old)
       ub2 = MIN(ub2_new,ub2_old)
       ALLOCATE (work(lb1:ub1,lb2:ub2),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,15,&
                           "work",t_size*(ub1-lb1+1)*&
                                         (ub2-lb2+1))

       END IF
       work(lb1:ub1,lb2:ub2) = p(lb1:ub1,lb2:ub2)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new),STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,26,&
                        "p",t_size*(ub1_new-lb1_new+1)*&
                                   (ub2_new-lb2_new+1))
    END IF
    p(:,:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1,lb2:ub2) = work(lb1:ub1,lb2:ub2)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_i2

! **************************************************************************************************
!> \brief (Re)Allocate an integer array 3D with a new dimension.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \param lb3_new ...
!> \param ub3_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_i3(p, lb1_new, ub1_new, lb2_new, ub2_new, lb3_new, ub3_new)

      INTEGER, DIMENSION(:, :, :), POINTER               :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new, &
                                                            lb3_new, ub3_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_i3', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: t_size = int_size, zero = 0

      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: work

    INTEGER :: istat, lb1, lb1_old, lb2, lb2_old, lb3, lb3_old, &
      ub1, ub1_old, ub2, ub2_old, ub3, ub3_old

    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb2_old = LBOUND(p,2)
       ub2_old = UBOUND(p,2)
       lb3_old = LBOUND(p,3)
       ub3_old = UBOUND(p,3)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       lb2 = MAX(lb2_new,lb2_old)
       ub2 = MIN(ub2_new,ub2_old)
       lb3 = MAX(lb3_new,lb3_old)
       ub3 = MIN(ub3_new,ub3_old)
       ALLOCATE (work(lb1:ub1,lb2:ub2,lb3:ub3),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,19,&
                           "work",t_size*(ub1-lb1+1)*&
                                         (ub2-lb2+1)*&
                                         (ub3-lb3+1))
       END IF
       work(lb1:ub1,lb2:ub2,lb3:ub3) = p(lb1:ub1,lb2:ub2,lb3:ub3)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new,lb3_new:ub3_new),STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,30,&
                        "p",t_size*(ub1_new-lb1_new+1)*&
                                   (ub2_new-lb2_new+1)*&
                                   (ub3_new-lb3_new+1))
    END IF
    p(:,:,:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1,lb2:ub2,lb3:ub3) = work(lb1:ub1,lb2:ub2,lb3:ub3)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_i3

! **************************************************************************************************
!> \brief (Re)Allocate an integer array 3D with a new dimension.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \param lb3_new ...
!> \param ub3_new ...
!> \param lb4_new ...
!> \param ub4_new ...
!> \par History
!>      none
!> \author Matthias Krack (04.10.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_i4(p, lb1_new, ub1_new, lb2_new, ub2_new, lb3_new, ub3_new, &
                            lb4_new, ub4_new)

      INTEGER, DIMENSION(:, :, :, :), POINTER            :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new, &
                                                            lb3_new, ub3_new, lb4_new, ub4_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_i4', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: t_size = int_size, zero = 0

      INTEGER, ALLOCATABLE, DIMENSION(:, :, :, :)        :: work

    INTEGER :: istat, lb1, lb1_old, lb2, lb2_old, lb3, lb3_old, lb4, lb4_old, &
      ub1, ub1_old, ub2, ub2_old, ub3, ub3_old, ub4, ub4_old

    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb2_old = LBOUND(p,2)
       ub2_old = UBOUND(p,2)
       lb3_old = LBOUND(p,3)
       ub3_old = UBOUND(p,3)
       lb4_old = LBOUND(p,4)
       ub4_old = UBOUND(p,4)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       lb2 = MAX(lb2_new,lb2_old)
       ub2 = MIN(ub2_new,ub2_old)
       lb3 = MAX(lb3_new,lb3_old)
       ub3 = MIN(ub3_new,ub3_old)
       lb4 = MAX(lb4_new,lb4_old)
       ub4 = MIN(ub4_new,ub4_old)
       ALLOCATE (work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,23,&
                           "work",t_size*(ub1-lb1+1)*&
                                         (ub2-lb2+1)*&
                                         (ub3-lb3+1)*&
                                         (ub4-lb4+1))
       END IF
       work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4) = p(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new,lb3_new:ub3_new,lb4_new:ub4_new),&
              STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,36,&
                        "p",t_size*(ub1_new-lb1_new+1)*&
                                   (ub2_new-lb2_new+1)*&
                                   (ub3_new-lb3_new+1)*&
                                   (ub4_new-lb4_new+1))
    END IF
    p(:,:,:,:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4) = work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_i4

! **************************************************************************************************
!> \brief (Re)Allocate an integer (int_8) vector with a new dimension.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_8i1(p, lb1_new, ub1_new)

      INTEGER(KIND=int_8), DIMENSION(:), POINTER         :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_8i1', routineP = moduleN//':'//routineN
      INTEGER(KIND=int_8), PARAMETER                     :: zero = 0
      INTEGER, PARAMETER                                 :: t_size = int_8_size

      INTEGER(KIND=int_8), ALLOCATABLE, DIMENSION(:)     :: work

    INTEGER                                  :: istat, lb1, lb1_old, &
                                                ub1,  ub1_old
    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       ALLOCATE (work(lb1:ub1),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,10,&
                           "work",t_size*(ub1-lb1+1))
       END IF
       work(lb1:ub1) = p(lb1:ub1)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new),STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,19,&
                        "p",t_size*(ub1_new-lb1_new+1))
    END IF
    p(:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1) = work(lb1:ub1)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_8i1

! **************************************************************************************************
!> \brief (Re)Allocate an integer (int_8) array with a new dimension.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_8i2(p, lb1_new, ub1_new, lb2_new, ub2_new)

      INTEGER(kind=int_8), DIMENSION(:, :), POINTER      :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_8i2', routineP = moduleN//':'//routineN
      INTEGER(KIND=int_8), PARAMETER                     :: zero = 0
      INTEGER, PARAMETER                                 :: t_size = int_8_size

      INTEGER(KIND=int_8), ALLOCATABLE, DIMENSION(:, :)  :: work

    INTEGER                                  :: istat, lb1, lb1_old, lb2, &
                                                lb2_old, ub1, ub1_old, ub2, ub2_old

    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb2_old = LBOUND(p,2)
       ub2_old = UBOUND(p,2)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       lb2 = MAX(lb2_new,lb2_old)
       ub2 = MIN(ub2_new,ub2_old)
       ALLOCATE (work(lb1:ub1,lb2:ub2),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,15,&
                           "work",t_size*(ub1-lb1+1)*&
                                         (ub2-lb2+1))

       END IF
       work(lb1:ub1,lb2:ub2) = p(lb1:ub1,lb2:ub2)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new),STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,26,&
                        "p",t_size*(ub1_new-lb1_new+1)*&
                                   (ub2_new-lb2_new+1))
    END IF
    p(:,:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1,lb2:ub2) = work(lb1:ub1,lb2:ub2)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_8i2

! **************************************************************************************************
!> \brief (Re)Allocate an real vector with a new dimension.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_r1(p, lb1_new, ub1_new)

      REAL(KIND=dp), DIMENSION(:), POINTER               :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_r1', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: t_size = dp_size
      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: work

    INTEGER                                  :: istat, lb1, lb1_old, &
                                                ub1,  ub1_old
    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       ALLOCATE (work(lb1:ub1),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,10,&
                           "work",t_size*(ub1-lb1+1))
       END IF
       work(lb1:ub1) = p(lb1:ub1)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new),STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,19,&
                        "p",t_size*(ub1_new-lb1_new+1))
    END IF
    p(:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1) = work(lb1:ub1)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_r1

! **************************************************************************************************
!> \brief (Re)Allocate an real array with new dimensions.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_r2(p, lb1_new, ub1_new, lb2_new, ub2_new)

      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_r2', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: t_size = dp_size
      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: work

    INTEGER                                  :: istat, lb1, lb1_old, lb2, &
                                                lb2_old, ub1, ub1_old, ub2, ub2_old

    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb2_old = LBOUND(p,2)
       ub2_old = UBOUND(p,2)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       lb2 = MAX(lb2_new,lb2_old)
       ub2 = MIN(ub2_new,ub2_old)
       ALLOCATE (work(lb1:ub1,lb2:ub2),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,15,&
                           "work",t_size*(ub1-lb1+1)*&
                                         (ub2-lb2+1))

       END IF
       work(lb1:ub1,lb2:ub2) = p(lb1:ub1,lb2:ub2)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new),STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,26,&
                        "p",t_size*(ub1_new-lb1_new+1)*&
                                   (ub2_new-lb2_new+1))
    END IF
    p(:,:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1,lb2:ub2) = work(lb1:ub1,lb2:ub2)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_r2

! **************************************************************************************************
!> \brief (Re)Allocate an real array 3D with new dimensions.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \param lb3_new ...
!> \param ub3_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_r3(p, lb1_new, ub1_new, lb2_new, ub2_new, lb3_new, ub3_new)

      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new, &
                                                            lb3_new, ub3_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_r3', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: t_size = dp_size
      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: work

    INTEGER :: istat, lb1, lb1_old, lb2, lb2_old, lb3, lb3_old, &
      ub1, ub1_old, ub2, ub2_old, ub3, ub3_old

    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb2_old = LBOUND(p,2)
       ub2_old = UBOUND(p,2)
       lb3_old = LBOUND(p,3)
       ub3_old = UBOUND(p,3)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       lb2 = MAX(lb2_new,lb2_old)
       ub2 = MIN(ub2_new,ub2_old)
       lb3 = MAX(lb3_new,lb3_old)
       ub3 = MIN(ub3_new,ub3_old)
       ALLOCATE (work(lb1:ub1,lb2:ub2,lb3:ub3),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,19,&
                           "work",t_size*(ub1-lb1+1)*&
                                         (ub2-lb2+1)*&
                                         (ub3-lb3+1))
       END IF
       work(lb1:ub1,lb2:ub2,lb3:ub3) = p(lb1:ub1,lb2:ub2,lb3:ub3)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new,lb3_new:ub3_new),STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,30,&
                        "p",t_size*(ub1_new-lb1_new+1)*&
                                   (ub2_new-lb2_new+1)*&
                                   (ub3_new-lb3_new+1))
    END IF
    p(:,:,:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1,lb2:ub2,lb3:ub3) = work(lb1:ub1,lb2:ub2,lb3:ub3)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_r3

! **************************************************************************************************
!> \brief (Re)Allocate an real array 4D with new dimensions.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \param lb3_new ...
!> \param ub3_new ...
!> \param lb4_new ...
!> \param ub4_new ...
!> \par History
!>      none
!> \author Matthias Krack (04.10.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_r4(p, lb1_new, ub1_new, lb2_new, ub2_new, lb3_new, ub3_new, &
                            lb4_new, ub4_new)

      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new, &
                                                            lb3_new, ub3_new, lb4_new, ub4_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_r4', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: t_size = dp_size
      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: work

    INTEGER :: istat, lb1, lb1_old, lb2, lb2_old, lb3, lb3_old, lb4, lb4_old, &
      ub1, ub1_old, ub2, ub2_old, ub3, ub3_old, ub4, ub4_old

    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb2_old = LBOUND(p,2)
       ub2_old = UBOUND(p,2)
       lb3_old = LBOUND(p,3)
       ub3_old = UBOUND(p,3)
       lb4_old = LBOUND(p,4)
       ub4_old = UBOUND(p,4)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       lb2 = MAX(lb2_new,lb2_old)
       ub2 = MIN(ub2_new,ub2_old)
       lb3 = MAX(lb3_new,lb3_old)
       ub3 = MIN(ub3_new,ub3_old)
       lb4 = MAX(lb4_new,lb4_old)
       ub4 = MIN(ub4_new,ub4_old)
       ALLOCATE (work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,23,&
                           "work",t_size*(ub1-lb1+1)*&
                                         (ub2-lb2+1)*&
                                         (ub3-lb3+1)*&
                                         (ub4-lb4+1))
       END IF
       work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4) = p(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new,lb3_new:ub3_new,lb4_new:ub4_new),&
              STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,36,&
                        "p",t_size*(ub1_new-lb1_new+1)*&
                                   (ub2_new-lb2_new+1)*&
                                   (ub3_new-lb3_new+1)*&
                                   (ub4_new-lb4_new+1))
    END IF
    p(:,:,:,:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4) = work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_r4

! **************************************************************************************************
!> \brief (Re)Allocate an real array 5D with new dimensions.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lb2_new ...
!> \param ub2_new ...
!> \param lb3_new ...
!> \param ub3_new ...
!> \param lb4_new ...
!> \param ub4_new ...
!> \param lb5_new ...
!> \param ub5_new ...
!> \par History
!>      none
!> \author Matthias Krack (04.10.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_r5(p, lb1_new, ub1_new, lb2_new, ub2_new, lb3_new, ub3_new, &
                            lb4_new, ub4_new, lb5_new, ub5_new)

      REAL(KIND=dp), DIMENSION(:, :, :, :, :), POINTER   :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new, lb2_new, ub2_new, &
                                                            lb3_new, ub3_new, lb4_new, ub4_new, &
                                                            lb5_new, ub5_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_r5', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: t_size = dp_size
      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      REAL(KIND=dp), ALLOCATABLE, &
         DIMENSION(:, :, :, :, :)                        :: work

    INTEGER :: istat, lb1, lb1_old, lb2, lb2_old, lb3, lb3_old, lb4, lb4_old, &
      lb5, lb5_old, ub1, ub1_old, ub2, ub2_old, ub3, ub3_old, ub4, ub4_old, ub5, ub5_old

    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb2_old = LBOUND(p,2)
       ub2_old = UBOUND(p,2)
       lb3_old = LBOUND(p,3)
       ub3_old = UBOUND(p,3)
       lb4_old = LBOUND(p,4)
       ub4_old = UBOUND(p,4)
       lb5_old = LBOUND(p,5)
       ub5_old = UBOUND(p,5)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       lb2 = MAX(lb2_new,lb2_old)
       ub2 = MIN(ub2_new,ub2_old)
       lb3 = MAX(lb3_new,lb3_old)
       ub3 = MIN(ub3_new,ub3_old)
       lb4 = MAX(lb4_new,lb4_old)
       ub4 = MIN(ub4_new,ub4_old)
       lb5 = MAX(lb5_new,lb5_old)
       ub5 = MIN(ub5_new,ub5_old)
       ALLOCATE (work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4,lb5:ub5),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,27,&
                           "work",t_size*(ub1-lb1+1)*&
                                         (ub2-lb2+1)*&
                                         (ub3-lb3+1)*&
                                         (ub4-lb4+1)*&
                                         (ub5-lb5+1))
       END IF
       work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4,lb5:ub5) = p(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4,lb5:ub5)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new,lb2_new:ub2_new,lb3_new:ub3_new,lb4_new:ub4_new,lb5_new:ub5_new),&
              STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,41,&
                        "p",t_size*(ub1_new-lb1_new+1)*&
                                   (ub2_new-lb2_new+1)*&
                                   (ub3_new-lb3_new+1)*&
                                   (ub4_new-lb4_new+1)*&
                                   (ub5_new-lb5_new+1))
    END IF
    p(:,:,:,:,:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4,lb5:ub5) = work(lb1:ub1,lb2:ub2,lb3:ub3,lb4:ub4,lb5:ub5)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_r5

! **************************************************************************************************
!> \brief (Re)Allocate a logical vector with a new dimension.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
! **************************************************************************************************
   SUBROUTINE reallocate_l1(p, lb1_new, ub1_new)

      LOGICAL, DIMENSION(:), POINTER                     :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_l1', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: t_size = int_size
      LOGICAL, PARAMETER                                 :: zero = .FALSE.

      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: work

    INTEGER                                  :: istat, lb1, lb1_old, &
                                                ub1,  ub1_old
    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       ALLOCATE (work(lb1:ub1),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,10,&
                           "work",t_size*(ub1-lb1+1))
       END IF
       work(lb1:ub1) = p(lb1:ub1)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new),STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,19,&
                        "p",t_size*(ub1_new-lb1_new+1))
    END IF
    p(:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1) = work(lb1:ub1)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_l1

! **************************************************************************************************
!> \brief (Re)Allocate a vector of string variables
!> \param p_short ...
!> \param lb_new ...
!> \param ub_new ...
!> \param p_long ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
!> \note
!>      (Maybe outdated) This routine doesnt work on SUN/Solaris!!!
!>                       It should probably not be used. (Thomas Chassaing)
! **************************************************************************************************
   SUBROUTINE reallocate_s1(p_short, lb_new, ub_new, p_long)

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), OPTIONAL, POINTER                 :: p_short
      INTEGER, INTENT(IN)                                :: lb_new, ub_new
      CHARACTER(LEN=default_path_length), DIMENSION(:), &
         OPTIONAL, POINTER                               :: p_long

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_s1', routineP = moduleN//':'//routineN

      IF (PRESENT(p_short)) THEN
         CALL reallocate_ss1(p_short, lb_new, ub_new)
      ELSE IF (PRESENT(p_long)) THEN
         CALL reallocate_ls1(p_long, lb_new, ub_new)
      ELSE
         CALL cp__b("common/memory_utilities.F",511,"At least one of the two optional arguments is required")
      END IF

   END SUBROUTINE reallocate_s1

! **************************************************************************************************
!> \brief (Re)Allocate a vector of string variables of default string length.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \par History
!>      none
!> \author Thomas Chassaing
!> \author Matthias Krack (18.07.2002,MK)
!> \note
!>      (Maybe outdated) This routine doesnt work on SUN/Solaris!!!
!>                       It should probably not be used. (Thomas Chassaing)
! **************************************************************************************************
   SUBROUTINE reallocate_ss1(p, lb1_new, ub1_new)

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_ss1', routineP = moduleN//':'//routineN
      CHARACTER(LEN=default_string_length), PARAMETER    :: zero = ""
      INTEGER, PARAMETER :: t_size = int_size*default_string_length

      CHARACTER(LEN=default_string_length), &
         ALLOCATABLE, DIMENSION(:)                       :: work

    INTEGER                                  :: istat, lb1, lb1_old, &
                                                ub1,  ub1_old
    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       ALLOCATE (work(lb1:ub1),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,10,&
                           "work",t_size*(ub1-lb1+1))
       END IF
       work(lb1:ub1) = p(lb1:ub1)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new),STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,19,&
                        "p",t_size*(ub1_new-lb1_new+1))
    END IF
    p(:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1) = work(lb1:ub1)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_ss1

! **************************************************************************************************
!> \brief (Re)Allocate a vector of string variables of default path length.
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \par History
!>      none
!> \author Matthias Krack (18.07.2002,MK)
!> \note
!>      (Maybe outdated) This routine doesnt work on SUN/Solaris!!!
!>                       It should probably not be used. (Thomas Chassaing)
! **************************************************************************************************
   SUBROUTINE reallocate_ls1(p, lb1_new, ub1_new)

      CHARACTER(LEN=default_path_length), DIMENSION(:), &
         POINTER                                         :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reallocate_ls1', routineP = moduleN//':'//routineN
      CHARACTER(LEN=default_path_length), PARAMETER      :: zero = ""
      INTEGER, PARAMETER :: t_size = int_size*default_path_length

      CHARACTER(LEN=default_path_length), ALLOCATABLE, &
         DIMENSION(:)                                    :: work

    INTEGER                                  :: istat, lb1, lb1_old, &
                                                ub1,  ub1_old
    IF (ASSOCIATED(p)) THEN
       lb1_old = LBOUND(p,1)
       ub1_old = UBOUND(p,1)
       lb1 = MAX(lb1_new,lb1_old)
       ub1 = MIN(ub1_new,ub1_old)
       ALLOCATE (work(lb1:ub1),STAT=istat)
       IF (istat /= 0) THEN
          CALL stop_memory(routineN,moduleN,10,&
                           "work",t_size*(ub1-lb1+1))
       END IF
       work(lb1:ub1) = p(lb1:ub1)
       DEALLOCATE (p)
    END IF

    ALLOCATE (p(lb1_new:ub1_new),STAT=istat)
    IF (istat /= 0) THEN
       CALL stop_memory(routineN,moduleN,19,&
                        "p",t_size*(ub1_new-lb1_new+1))
    END IF
    p(:) = zero

    IF (ASSOCIATED(p).AND.ALLOCATED(work)) THEN
       p(lb1:ub1) = work(lb1:ub1)
       DEALLOCATE (work)
    END IF

   END SUBROUTINE reallocate_ls1

END MODULE memory_utilities
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief routines to handle the output, The idea is to remove the
!>      decision of wheter to output and what to output from the code
!>      that does the output, and centralize it here.
!> \note
!>      These were originally together with the log handling routines,
!>      but have been spawned off. Some dependencies are still there,
!>      and some of the comments about log handling also applies to output
!>      handling: @see cp_log_handling
!> \par History
!>      12.2001 created [fawzi]
!>      08.2002 updated to new logger [fawzi]
!>      10.2004 big rewrite of the output methods, connected to the new
!>              input, and iteration_info [fawzi]
!>      08.2005 property flags [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_output_handling
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_iter_types,                   ONLY: cp_iteration_info_release,&
                                              cp_iteration_info_retain,&
                                              cp_iteration_info_type,&
                                              each_desc_labels,&
                                              each_possible_labels
   USE cp_log_handling,                 ONLY: cp_logger_generate_filename,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_get_unit_nr,&
                                              cp_logger_type,&
                                              cp_to_string
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length
   USE machine,                         ONLY: m_mov
   USE memory_utilities,                ONLY: reallocate
   USE string_utilities,                ONLY: compress,&
                                              s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_output_handling'
   PUBLIC :: cp_print_key_should_output, cp_iterate, cp_add_iter_level, cp_rm_iter_level
   PUBLIC :: cp_iter_string, cp_print_key_section_create
   PUBLIC :: cp_print_key_unit_nr, cp_print_key_finished_output
   PUBLIC :: cp_print_key_generate_filename, cp_printkey_is_on

   INTEGER, PARAMETER, PUBLIC               :: add_last_no = 0, &
                                               add_last_numeric = 1, &
                                               add_last_symbolic = 2
   INTEGER, PARAMETER, PUBLIC               :: silent_print_level = 0, &
                                               low_print_level = 1, &
                                               medium_print_level = 2, &
                                               high_print_level = 3, &
                                               debug_print_level = 4

!! flags controlling the printing and storing of a property.
!!
!! cp_out_none: do not calculate the property
!! cp_out_file_if  : if the printkey says it calculate and output the property
!! cp_out_store_if : if the printkey says it calculate and store in memory
!!                   the property
!! cp_out_file_each: calculate and output the property with the same periodicity
!!                   as said in the printkey (irrespective of the activation of
!!                   the printkey)
!! cp_out_store_each: calculate and store the property with the same periodicity
!!                   as said in the printkey (irrespective of the activation of
!!                   the printkey)
!! cp_out_file: always calculate and output the property
!! cp_out_store: always calculate and store in memory the property
!! cp_out_calc: just calculate the value (indipendently from the fact that there
!!              should be output)
!! cp_out_default: the default value for proprety flags (cp_out_file_if)
!!
!! this flags can be ior-ed together:
!! ior(cp_out_file_if,cp_out_store_if): if the printkey says it both print
!!                                          and store the property
!!
!! there is no guarantee that a proprety is not stored if it is not necessary
!! not all printkeys have a control flag
   INTEGER, PUBLIC, PARAMETER :: cp_p_file_if = 3, cp_p_store_if = 4, &
                                 cp_p_store = 2, cp_p_file = 1, cp_p_file_each = 5, cp_p_store_each = 6, cp_p_calc = 7
   INTEGER, PUBLIC, PARAMETER :: cp_out_none = 0, cp_out_file_if = IBSET(0, cp_p_file_if), &
                                 cp_out_store_if = IBSET(0, cp_p_store_if), cp_out_file = IBSET(0, cp_p_file), &
                                 cp_out_store = IBSET(0, cp_p_store), cp_out_calc = IBSET(0, cp_p_calc), &
                                 cp_out_file_each = IBSET(0, cp_p_file_each), &
                                 cp_out_store_each = IBSET(0, cp_p_store_each), &
                                 cp_out_default = cp_out_file_if

! **************************************************************************************************
!> \brief stores the flags_env controlling the output of properties
!> \param ref_count reference count (see doc/ReferenceCounting.html)
!> \param id_nr identification number (unique to each istance)
!> \param n_flags number of flags stored in this type
!> \param names names of the stored flags
!> \param control_val value of the flag
!> \param input the input (with all the printkeys)
!> \param logger logger and iteration information (to know if output is needed)
!> \param strict if flags that were not stored can be read
!> \param default_val default value of the flags that are not explicitly
!>        stored
!> \note
!>      Two features of this object should be:
!>        1) easy state storage, one should be able to store the state of the
!>           flags, to some changes to them just for one (or few) force evaluations
!>           and then reset the original state. The actual implementation is good
!>           in this respect
!>        2) work well with subsections. This is a problem at the moment, as
!>           if you pass just a subsection of the input the control flags get lost.
!>        A better implementation should be done storing the flags also in the
!>        input itself to be transparent
!> \author fawzi
! **************************************************************************************************
   TYPE cp_out_flags_type
      INTEGER :: ref_count, id_nr, n_flags
      CHARACTER(default_string_length), DIMENSION(:), POINTER :: names
      INTEGER, DIMENSION(:), POINTER :: control_val
      TYPE(section_vals_type), POINTER :: input
      TYPE(cp_logger_type), POINTER :: logger
      LOGICAL :: strict
      INTEGER :: default_val
   END TYPE cp_out_flags_type

CONTAINS

! **************************************************************************************************
!> \brief creates a print_key section
!> \param print_key_section the print key to create
!> \param name the name of the print key
!> \param description the description of the print key
!> \param print_level print level starting at which the printing takes place
!>        (defaults to debug_print_level)
!> \param each_iter_names ...
!> \param each_iter_values ...
!> \param add_last ...
!> \param filename ...
!> \param common_iter_levels ...
!> \param citations ...
!> \param unit_str specifies an unit of measure for output quantity. If not
!>        provided the control is totally left to how the output was coded
!>        (i.e. USERS have no possibility to change it)
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE cp_print_key_section_create(print_key_section, name, description, &
                                          print_level, each_iter_names, each_iter_values, add_last, filename, &
                                          common_iter_levels, citations, unit_str)
      TYPE(section_type), POINTER                        :: print_key_section
      CHARACTER(len=*), INTENT(IN)                       :: name, description
      INTEGER, INTENT(IN), OPTIONAL                      :: print_level
      CHARACTER(LEN=*), DIMENSION(:), INTENT(IN), &
         OPTIONAL                                        :: each_iter_names
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: each_iter_values
      INTEGER, INTENT(IN), OPTIONAL                      :: add_last
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: filename
      INTEGER, INTENT(IN), OPTIONAL                      :: common_iter_levels
      INTEGER, DIMENSION(:), OPTIONAL                    :: citations
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: unit_str

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_print_key_section_create', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: my_filename
      INTEGER                                            :: i_each, i_iter, my_add_last, &
                                                            my_comm_iter_levels, my_print_level, &
                                                            my_value
      LOGICAL                                            :: check, ext_each
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(print_key_section)))CALL cp__a("input/cp_output_handling.F",187)
      my_print_level = debug_print_level
      IF (PRESENT(print_level)) my_print_level = print_level

      CALL section_create(print_key_section, name=name, description=description, &
                          n_keywords=2, n_subsections=0, repeats=.FALSE., &
                          citations=citations)

      NULLIFY (keyword, subsection)
      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="Level starting at which this property is printed", &
                          usage="silent", &
                          default_i_val=my_print_level, lone_keyword_i_val=silent_print_level, &
                          enum_c_vals=s2a("on", "off", "silent", "low", "medium", "high", "debug"), &
                          enum_i_vals=(/silent_print_level-1, debug_print_level+1, &
                                        silent_print_level, low_print_level, &
                                        medium_print_level, high_print_level, debug_print_level/))
      CALL section_add_keyword(print_key_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="__CONTROL_VAL", &
                          description=' hidden parameter that controls storage, printing,...'// &
                          ' of the print_key', &
                          default_i_val=cp_out_default)
      CALL section_add_keyword(print_key_section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(subsection, name="EACH", &
                          description="This section specifies how often this proprety is printed."// &
                          "Each keyword inside this section is mapping to a specific iteration level and "// &
                          "the value of each of these keywords is matched with the iteration level during "// &
                          "the calculation. How to handle the last iteration is treated "// &
                          "separately in ADD_LAST (this mean that each iteration level (MD, GEO_OPT, etc..), "// &
                          "though equal to 0, might print the last iteration). If an iteration level is specified "// &
                          "that is not present in the flow of the calculation it is just ignored.", &
                          n_keywords=2, n_subsections=0, repeats=.FALSE., &
                          citations=citations)

      ! Enforce the presence or absence of both.. or give an error
      check = (PRESENT(each_iter_names)) .EQV. (PRESENT(each_iter_values))
      IF(.NOT.(check))CALL cp__a("input/cp_output_handling.F",227)
      ext_each = (PRESENT(each_iter_names)) .AND. (PRESENT(each_iter_values))

      DO i_each = 1, SIZE(each_possible_labels)
         my_value = 1
         IF (ext_each) THEN
            check = SUM(INDEX(each_iter_names, each_possible_labels(i_each))) <= 1
            IF(.NOT.(check))CALL cp__a("input/cp_output_handling.F",234)
            DO i_iter = 1, SIZE(each_iter_names)
               IF (INDEX(TRIM(each_iter_names(i_iter)), TRIM(each_possible_labels(i_each))) /= 0) THEN
                  my_value = each_iter_values(i_iter)
               END IF
            END DO
         END IF
         CALL keyword_create(keyword, name=TRIM(each_possible_labels(i_each)), &
                             description=TRIM(each_desc_labels(i_each)), &
                             usage=TRIM(each_possible_labels(i_each))//" <INTEGER>", &
                             default_i_val=my_value)
         CALL section_add_keyword(subsection, keyword)
         CALL keyword_release(keyword)
      END DO
      CALL section_add_subsection(print_key_section, subsection)
      CALL section_release(subsection)

      my_add_last = add_last_no
      IF (PRESENT(add_last)) THEN
         my_add_last = add_last
      END IF
      CALL keyword_create(keyword, name="ADD_LAST", &
                          description="If the last iteration should be added, and if it "// &
                          "should be marked symbolically (with lowercase letter l) or with "// &
                          "the iteration number. "// &
                          "Not every iteration level is able to identify the last iteration "// &
                          "early enough to be able to output. When this keyword is activated "// &
                          "all iteration levels are checked for the last iteration step.", &
                          usage="ADD_LAST (NO|NUMERIC|SYMBOLIC)", &
                          enum_c_vals=s2a("no", "numeric", "symbolic"), &
                          enum_i_vals=(/add_last_no, add_last_numeric, add_last_symbolic/), &
                          enum_desc=s2a("Do not mark last iteration specifically", &
                                        "Mark last iteration with its iteration number", &
                                        "Mark last iteration with lowercase letter l"), &
                          default_i_val=my_add_last)
      CALL section_add_keyword(print_key_section, keyword)
      CALL keyword_release(keyword)

      my_comm_iter_levels = 0
      IF (PRESENT(common_iter_levels)) my_comm_iter_levels = common_iter_levels
      CALL keyword_create(keyword, name="COMMON_ITERATION_LEVELS", &
                          description="How many iterations levels should be written"// &
                          " in the same file (no extra information about the actual"// &
                          " iteration level is written to the file)", &
                          usage="COMMON_ITERATION_LEVELS <INTEGER>", &
                          default_i_val=my_comm_iter_levels)
      CALL section_add_keyword(print_key_section, keyword)
      CALL keyword_release(keyword)

      my_filename = ""
      IF (PRESENT(filename)) my_filename = filename
      CALL keyword_create(keyword, name="FILENAME", &
                          description=' controls part of the filename for output. '// &
                          ' use __STD_OUT__ (exactly as written here) for the screen or standard logger. '// &
                          ' use filename to obtain projectname-filename. '// &
                          ' use ./filename to get filename.'// &
                          ' A middle name (if present), iteration numbers'// &
                          ' and extension are always added to the filename.'// &
                          ' if you want to avoid it use =filename, in this'// &
                          ' case the filename is always exactly as typed.'// &
                          ' Please note that this can lead to clashes of'// &
                          ' filenames.', &
                          usage="FILENAME ./filename ", &
                          default_lc_val=my_filename)
      CALL section_add_keyword(print_key_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LOG_PRINT_KEY", &
                          description="This keywords enables the logger for the print_key (a message is printed on "// &
                          "screen everytime data, controlled by this print_key, are written)", &
                          usage="LOG_PRINT_KEY <LOGICAL>", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key_section, keyword)
      CALL keyword_release(keyword)

      IF (PRESENT(unit_str)) THEN
         CALL keyword_create(keyword, name="UNIT", &
                             description='Specify the unit of measurement for the quantity in output. '// &
                             "All available CP2K units can be used.", &
                             usage="UNIT angstrom", default_c_val=TRIM(unit_str))
         CALL section_add_keyword(print_key_section, keyword)
         CALL keyword_release(keyword)
      END IF
   END SUBROUTINE cp_print_key_section_create

! **************************************************************************************************
!> \brief returns what should be done with the given proprety
!>      if btest(res,cp_p_store) then the property should be stored in memory
!>      if btest(res,cp_p_file) then the property should be print ed to a file
!>      if res==0 then nothing should be done
!> \param iteration_info information about the actual iteration level
!> \param basis_section section that contains the printkey
!> \param print_key_path path to the printkey- "%" between sections, and
!>        optionally a "/" and a logical flag to check). Might be empty.
!> \param used_print_key here the print_key that was used is returned
!> \param first_time if it ist the first time that an output is written
!>        (not fully correct, but most of the time)
!> \retval res ...
!> \author fawzi
!> \note
!>      not all the propreties support can be stored
! **************************************************************************************************
   FUNCTION cp_print_key_should_output(iteration_info, basis_section, &
                                       print_key_path, used_print_key, first_time) &
      RESULT(res)
      TYPE(cp_iteration_info_type), POINTER              :: iteration_info
      TYPE(section_vals_type), POINTER                   :: basis_section
      CHARACTER(len=*), INTENT(IN), OPTIONAL             :: print_key_path
      TYPE(section_vals_type), OPTIONAL, POINTER         :: used_print_key
      LOGICAL, INTENT(OUT), OPTIONAL                     :: first_time
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_print_key_should_output', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: end_str, my_control_val, to_path
      LOGICAL                                            :: flags, is_iter, is_on
      TYPE(section_vals_type), POINTER                   :: print_key

      res = 0
      IF (PRESENT(first_time)) first_time = .FALSE.
      IF(.NOT.(ASSOCIATED(basis_section)))CALL cp__a("input/cp_output_handling.F",354)
      IF(.NOT.(basis_section%ref_count > 0))CALL cp__a("input/cp_output_handling.F",355)
      IF (PRESENT(used_print_key)) NULLIFY (used_print_key)
      ! IF (failure) THEN
      !    IF (iteration_info%print_level>=debug_print_level) res=cp_out_default
      !    RETURN
      ! END IF

      IF (PRESENT(print_key_path)) THEN
         end_str = LEN_TRIM(print_key_path)
         to_path = INDEX(print_key_path, "/")
         IF (to_path < 1) THEN
            to_path = end_str+1
         END IF

         IF (to_path > 1) THEN
            print_key => section_vals_get_subs_vals(basis_section, &
                                                    print_key_path(1:(to_path-1)))
         ELSE
            print_key => basis_section
         END IF
         IF(.NOT.(ASSOCIATED(print_key)))CALL cp__a("input/cp_output_handling.F",375)
         IF(.NOT.(print_key%ref_count > 0))CALL cp__a("input/cp_output_handling.F",376)
         IF (to_path+1 < end_str) THEN
            CALL section_vals_val_get(print_key, print_key_path((to_path+1):end_str), &
                                      l_val=flags)
         ELSE
            flags = .TRUE.
         END IF
      ELSE
         print_key => basis_section
         flags = .TRUE.
      END IF
      IF (PRESENT(used_print_key)) used_print_key => print_key

      IF (.NOT. flags) RETURN

      CALL section_vals_val_get(print_key, "__CONTROL_VAL", &
                                i_val=my_control_val)
      is_on = cp_printkey_is_on(iteration_info, print_key)

      ! a shortcut for most common case
      IF (my_control_val == cp_out_default .AND. .NOT. is_on) RETURN

      is_iter = cp_printkey_is_iter(iteration_info, print_key, first_time=first_time)

      IF (BTEST(my_control_val, cp_p_store)) THEN
         res = IBSET(res, cp_p_store)
      ELSE IF (BTEST(my_control_val, cp_p_store_if) .AND. is_iter .AND. is_on) THEN
         res = IBSET(res, cp_p_store)
      ELSE IF (BTEST(my_control_val, cp_p_store_each) .AND. is_iter) THEN
         res = IBSET(res, cp_p_store)
      END IF

      IF (BTEST(my_control_val, cp_p_file)) THEN
         res = IBSET(res, cp_p_file)
      ELSE IF (BTEST(my_control_val, cp_p_file_if) .AND. is_iter .AND. is_on) THEN
         res = IBSET(res, cp_p_file)
      ELSE IF (BTEST(my_control_val, cp_p_file_each) .AND. is_iter) THEN
         res = IBSET(res, cp_p_file)
      END IF
      IF (BTEST(my_control_val, cp_p_calc) .OR. res /= 0) THEN
         res = IBSET(res, cp_p_calc)
      END IF
   END FUNCTION cp_print_key_should_output

! **************************************************************************************************
!> \brief returns true if the printlevel activates this printkey
!>      does not look if this iteration it should be printed
!> \param iteration_info information about the actual iteration level
!> \param print_key the section values of the key to be printed
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION cp_printkey_is_on(iteration_info, print_key) RESULT(res)
      TYPE(cp_iteration_info_type), POINTER              :: iteration_info
      TYPE(section_vals_type), POINTER                   :: print_key
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_printkey_is_on', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: print_level

      IF(.NOT.(ASSOCIATED(iteration_info)))CALL cp__a("input/cp_output_handling.F",438)
      IF(.NOT.(iteration_info%ref_count > 0))CALL cp__a("input/cp_output_handling.F",439)
      IF (.NOT. ASSOCIATED(print_key)) THEN
         res = (iteration_info%print_level > debug_print_level)
      ELSE
         IF(.NOT.(print_key%ref_count > 0))CALL cp__a("input/cp_output_handling.F",443)
         CALL section_vals_val_get(print_key, "_SECTION_PARAMETERS_", i_val=print_level)
         res = iteration_info%print_level >= print_level
      END IF
   END FUNCTION cp_printkey_is_on

! **************************************************************************************************
!> \brief returns if the actual iteration matches those selected by the
!>      given printkey. Does not check it the prinkey is active (at the
!>      actual print_level)
!> \param iteration_info information about the actual iteration level
!> \param print_key the section values of the key to be printed
!> \param first_time returns if it is the first time that output is written
!>        (not fully correct, but most of the time)
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION cp_printkey_is_iter(iteration_info, print_key, first_time) &
      RESULT(res)
      TYPE(cp_iteration_info_type), POINTER              :: iteration_info
      TYPE(section_vals_type), POINTER                   :: print_key
      LOGICAL, INTENT(OUT), OPTIONAL                     :: first_time
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_printkey_is_iter', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: add_last, ilevel, iter_nr, ival
      LOGICAL                                            :: first, level_passed

      IF(.NOT.(ASSOCIATED(iteration_info)))CALL cp__a("input/cp_output_handling.F",473)
      IF(.NOT.(iteration_info%ref_count > 0))CALL cp__a("input/cp_output_handling.F",474)
      IF (.NOT. ASSOCIATED(print_key)) THEN
         res = (iteration_info%print_level > debug_print_level)
         first = ALL(iteration_info%iteration(1:iteration_info%n_rlevel) == 1)
      ELSE
         IF(.NOT.(print_key%ref_count > 0))CALL cp__a("input/cp_output_handling.F",479)
         res = .FALSE.
         first = .FALSE.
         CALL section_vals_val_get(print_key, "ADD_LAST", i_val=add_last)
         res = .TRUE.
         first = .TRUE.
         DO ilevel = 1, iteration_info%n_rlevel
            level_passed = .FALSE.
            CALL section_vals_val_get(print_key, "EACH%"//TRIM(iteration_info%level_name(ilevel)), &
                                      i_val=ival)
            IF (ival > 0) THEN
               iter_nr = iteration_info%iteration(ilevel)
               IF (iter_nr/ival > 1) first = .FALSE.
               IF (MODULO(iter_nr, ival) == 0) THEN
                  level_passed = .TRUE.
               END IF
            END IF
            IF (add_last == add_last_numeric .OR. add_last == add_last_symbolic) THEN
               IF (iteration_info%last_iter(ilevel)) THEN
                  level_passed = .TRUE.
               END IF
            END IF
            IF (.NOT. level_passed) res = .FALSE.
         END DO
      END IF
      first = first .AND. res
      IF (PRESENT(first_time)) first_time = first
   END FUNCTION cp_printkey_is_iter

! **************************************************************************************************
!> \brief returns the iteration string, a string that is useful to create
!>      unique filenames (once you trim it)
!> \param iter_info the iteration info from where to take the iteration
!>        number
!> \param print_key the print key to optionally show the last iteration
!>        symbolically
!> \param for_file if the string is to be used for file generation
!>        (and should consequently ignore some iteration levels depending
!>        on COMMON_ITERATION_LEVELS).
!>        Defaults to false.
!> \retval res ...
!> \author fawzi
!> \note
!>      If the root level is 1 removes it
! **************************************************************************************************
   FUNCTION cp_iter_string(iter_info, print_key, for_file) RESULT(res)
      TYPE(cp_iteration_info_type), POINTER              :: iter_info
      TYPE(section_vals_type), OPTIONAL, POINTER         :: print_key
      LOGICAL, INTENT(IN), OPTIONAL                      :: for_file
      CHARACTER(len=default_string_length)               :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_iter_string', routineP = moduleN//':'//routineN

      INTEGER                                            :: add_last, c_i_level, ilevel, n_rlevel, &
                                                            s_level
      LOGICAL                                            :: my_for_file
      TYPE(section_vals_type), POINTER                   :: my_print_key

      res = ""
      my_for_file = .FALSE.
      IF (PRESENT(for_file)) my_for_file = for_file
      IF(.NOT.(ASSOCIATED(iter_info)))CALL cp__a("input/cp_output_handling.F",540)
      IF(.NOT.(iter_info%ref_count > 0))CALL cp__a("input/cp_output_handling.F",541)
      NULLIFY (my_print_key)
      IF (PRESENT(print_key)) my_print_key => print_key
      s_level = 1
      IF (ASSOCIATED(my_print_key)) THEN
         CALL section_vals_val_get(my_print_key, "ADD_LAST", i_val=add_last)
         CALL section_vals_val_get(my_print_key, "COMMON_ITERATION_LEVELS", i_val=c_i_level)
         n_rlevel = iter_info%n_rlevel
         IF (my_for_file) n_rlevel = MIN(n_rlevel, MAX(0, n_rlevel-c_i_level))
         DO ilevel = s_level, n_rlevel
            IF (iter_info%last_iter(ilevel)) THEN
               IF (add_last == add_last_symbolic) THEN
                  WRITE (res(9*ilevel-8:9*ilevel), "('l_')")
               ELSE
                  WRITE (res(9*ilevel-8:9*ilevel), "(i8,'_')") iter_info%iteration(ilevel)
               END IF
            ELSE
               WRITE (res(9*ilevel-8:9*ilevel), "(i8,'_')") iter_info%iteration(ilevel)
            END IF
         END DO
      ELSE
         DO ilevel = s_level, iter_info%n_rlevel
            WRITE (res(9*ilevel-8:9*ilevel), "(i8,'_')") iter_info%iteration(ilevel)
         END DO
      END IF
      CALL compress(res, .TRUE.)
      IF (LEN_TRIM(res) > 0) THEN
         res(LEN_TRIM(res):LEN_TRIM(res)) = " "
      END IF
   END FUNCTION cp_iter_string

! **************************************************************************************************
!> \brief adds one to the actual iteration
!> \param iteration_info the iteration info to update
!> \param last if this iteration is the last one (defaults to false)
!> \param iter_nr ...
!> \param increment ...
!> \param iter_nr_out ...
!> \author fawzi
!> \note
!>      this is supposed to be called at the beginning of each iteration
! **************************************************************************************************
   SUBROUTINE cp_iterate(iteration_info, last, iter_nr, increment, iter_nr_out)
      TYPE(cp_iteration_info_type), POINTER              :: iteration_info
      LOGICAL, INTENT(IN), OPTIONAL                      :: last
      INTEGER, INTENT(IN), OPTIONAL                      :: iter_nr, increment
      INTEGER, INTENT(OUT), OPTIONAL                     :: iter_nr_out

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_iterate', routineP = moduleN//':'//routineN

      INTEGER                                            :: my_increment
      LOGICAL                                            :: my_last

      my_last = .FALSE.
      my_increment = 1
      IF (PRESENT(last)) my_last = last
      IF (PRESENT(increment)) my_increment = increment
      IF (PRESENT(iter_nr_out)) iter_nr_out = -1

      IF(.NOT.(ASSOCIATED(iteration_info)))CALL cp__a("input/cp_output_handling.F",600)
      IF(.NOT.(iteration_info%ref_count > 0))CALL cp__a("input/cp_output_handling.F",601)
      IF (PRESENT(iter_nr)) THEN
         iteration_info%iteration(iteration_info%n_rlevel) = iter_nr
      ELSE
         iteration_info%iteration(iteration_info%n_rlevel) = &
            iteration_info%iteration(iteration_info%n_rlevel)+my_increment
      END IF
      ! If requested provide the value of the iteration level
      IF (PRESENT(iter_nr_out)) iter_nr_out = iteration_info%iteration(iteration_info%n_rlevel)

      ! Possibly setup the LAST flag
      iteration_info%last_iter(iteration_info%n_rlevel) = my_last
   END SUBROUTINE cp_iterate

! **************************************************************************************************
!> \brief Adds an iteration level
!> \param iteration_info the iteration info to which an iteration level has
!>        to be added
!> \param level_name the name of this level, for pretty printing only, right now
!> \param n_rlevel_new number of iteration levels after this call
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE cp_add_iter_level(iteration_info, level_name, n_rlevel_new)
      TYPE(cp_iteration_info_type), POINTER              :: iteration_info
      CHARACTER(LEN=*), INTENT(IN)                       :: level_name
      INTEGER, INTENT(OUT), OPTIONAL                     :: n_rlevel_new

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_add_iter_level', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: found

      IF(.NOT.(ASSOCIATED(iteration_info)))CALL cp__a("input/cp_output_handling.F",634)
      IF(.NOT.(iteration_info%ref_count > 0))CALL cp__a("input/cp_output_handling.F",635)
      found = .FALSE.
      DO i = 1, SIZE(each_possible_labels)
         IF (TRIM(level_name) == TRIM(each_possible_labels(i))) THEN
            found = .TRUE.
            EXIT
         END IF
      END DO
      IF (found) THEN
         CALL cp_iteration_info_retain(iteration_info)
         iteration_info%n_rlevel = iteration_info%n_rlevel+1
         CALL reallocate(iteration_info%iteration, 1, iteration_info%n_rlevel)
         CALL reallocate(iteration_info%level_name, 1, iteration_info%n_rlevel)
         CALL reallocate(iteration_info%last_iter, 1, iteration_info%n_rlevel)
         iteration_info%iteration(iteration_info%n_rlevel) = 0
         iteration_info%level_name(iteration_info%n_rlevel) = level_name
         iteration_info%last_iter(iteration_info%n_rlevel) = .FALSE.
         IF (PRESENT(n_rlevel_new)) n_rlevel_new = iteration_info%n_rlevel
      ELSE
         CALL cp_abort(cp__l("input/cp_output_handling.F",654), &
                       "Trying to create an iteration level ("//TRIM(level_name)//") not defined."// &
                       "Please update the module: cp_iter_types.")
      END IF

   END SUBROUTINE cp_add_iter_level

! **************************************************************************************************
!> \brief Removes an iteration level
!> \param iteration_info the iteration info to which an iteration level has
!>        to be removed
!> \param level_name level_name to be destroyed (if does not match gives an error)
!> \param n_rlevel_att iteration level before the call (to do some checks)
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE cp_rm_iter_level(iteration_info, level_name, n_rlevel_att)
      TYPE(cp_iteration_info_type), POINTER              :: iteration_info
      CHARACTER(LEN=*), INTENT(IN)                       :: level_name
      INTEGER, INTENT(IN), OPTIONAL                      :: n_rlevel_att

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_rm_iter_level', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: check

      IF(.NOT.(ASSOCIATED(iteration_info)))CALL cp__a("input/cp_output_handling.F",679)
      IF(.NOT.(iteration_info%ref_count > 0))CALL cp__a("input/cp_output_handling.F",680)
      IF (PRESENT(n_rlevel_att)) THEN
         IF(.NOT.(n_rlevel_att == iteration_info%n_rlevel))CALL cp__a("input/cp_output_handling.F",682)
      END IF
      CALL cp_iteration_info_release(iteration_info)
      ! This check that the iteration levels are consistently created and destroyed..
      ! Never remove this check..
      check = iteration_info%level_name(iteration_info%n_rlevel) == level_name
      IF(.NOT.(check))CALL cp__a("input/cp_output_handling.F",688)
      iteration_info%n_rlevel = iteration_info%n_rlevel-1
      CALL reallocate(iteration_info%iteration, 1, iteration_info%n_rlevel)
      CALL reallocate(iteration_info%level_name, 1, iteration_info%n_rlevel)
      CALL reallocate(iteration_info%last_iter, 1, iteration_info%n_rlevel)
   END SUBROUTINE cp_rm_iter_level

! **************************************************************************************************
!> \brief Utility function that retuns a unit number to write the print key.
!>     Might open a file with a unique filename, generated from
!>     the print_key name and iteration info.
!>
!>     Normally a valid unit (>0) is returned only if cp_print_key_should_output
!>     says that the print_key should be printed, and if the unit is global
!>     only the io node has a valid unit.
!>     So in many cases you can decide if you should print just checking if
!>     the returned units is bigger than 0.
!>
!>     IMPORTANT you should call cp_finished_output when an iteration output is
!>     finished (to immediately close the file that might have been opened)
!> \param logger the logger for the parallel environment, iteration info
!>        and filename generation
!> \param print_key ...
!> \param middle_name name to be added to the generated filename, useful when
!>        print_key activates different distinct outputs, to be able to
!>        distinguish them
!> \param extension extension to be applied to the filename (including the ".")
!> \param my_local if the unit should be local to this task, or global to the
!>        program (defaults to false).
!> \retval filename ...
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_print_key_generate_filename(logger, print_key, middle_name, extension, &
                                           my_local) RESULT(filename)
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key
      CHARACTER(len=*), INTENT(IN), OPTIONAL             :: middle_name
      CHARACTER(len=*), INTENT(IN)                       :: extension
      LOGICAL, INTENT(IN)                                :: my_local
      CHARACTER(len=default_path_length)                 :: filename

      CHARACTER(len=default_path_length)                 :: outPath, postfix, root
      CHARACTER(len=default_string_length)               :: my_middle_name, outName
      INTEGER                                            :: my_ind1, my_ind2
      LOGICAL                                            :: has_root

      CALL section_vals_val_get(print_key, "FILENAME", c_val=outPath)
      IF (outPath(1:1) == '=') THEN
         filename = outPath(2:LEN_TRIM(outPath))
         RETURN
      END IF
      IF (outPath == "__STD_OUT__") outPath = ""
      outName = outPath
      has_root = .FALSE.
      my_ind1 = INDEX(outPath, "/")
      my_ind2 = LEN_TRIM(outPath)
      IF (my_ind1 /= 0) THEN
         has_root = .TRUE.
         DO WHILE (INDEX(outPath(my_ind1+1:my_ind2), "/") /= 0)
            my_ind1 = INDEX(outPath(my_ind1+1:my_ind2), "/")+my_ind1
         END DO
         IF (my_ind1 == my_ind2) THEN
            outName = ""
         ELSE
            outName = outPath(my_ind1+1:my_ind2)
         END IF
      END IF

      IF (PRESENT(middle_name)) THEN
         IF (outName /= "") THEN
            my_middle_name = "-"//TRIM(outName)//"-"//middle_name
         ELSE
            my_middle_name = "-"//middle_name
         END IF
      ELSE
         IF (outName /= "") THEN
            my_middle_name = "-"//TRIM(outName)
         ELSE
            my_middle_name = ""
         END IF
      ENDIF

      IF (.NOT. has_root) THEN
         root = TRIM(logger%iter_info%project_name)//TRIM(my_middle_name)
      ELSE IF (outName == "") THEN
         root = outPath(1:my_ind1)//TRIM(logger%iter_info%project_name)//TRIM(my_middle_name)
      ELSE
         root = outPath(1:my_ind1)//my_middle_name(2:LEN_TRIM(my_middle_name))
      END IF

      ! use the cp_iter_string as a postfix
      postfix = "-"//TRIM(cp_iter_string(logger%iter_info, print_key=print_key, for_file=.TRUE.))
      IF (TRIM(postfix) == "-") postfix = ""

      ! and add the extension
      postfix = TRIM(postfix)//extension
      ! and let the logger generate the filename
      CALL cp_logger_generate_filename(logger, res=filename, &
                                       root=root, postfix=postfix, local=my_local)

   END FUNCTION cp_print_key_generate_filename

! **************************************************************************************************
!> \brief ...
!> \param logger ...
!> \param basis_section ...
!> \param print_key_path ...
!> \param extension ...
!> \param middle_name ...
!> \param local ...
!> \param log_filename ...
!> \param ignore_should_output ...
!> \param file_form ...
!> \param file_position ...
!> \param file_action ...
!> \param file_status ...
!> \param do_backup ...
!> \param on_file ...
!> \param is_new_file true if this rank created a new (or rewound) file, false otherwise
!> \retval res ...
! **************************************************************************************************
   FUNCTION cp_print_key_unit_nr(logger, basis_section, print_key_path, extension, &
                                 middle_name, local, log_filename, ignore_should_output, file_form, file_position, &
                                 file_action, file_status, do_backup, on_file, is_new_file) RESULT(res)
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: basis_section
      CHARACTER(len=*), INTENT(IN), OPTIONAL             :: print_key_path
      CHARACTER(len=*), INTENT(IN)                       :: extension
      CHARACTER(len=*), INTENT(IN), OPTIONAL             :: middle_name
      LOGICAL, INTENT(IN), OPTIONAL                      :: local, log_filename, ignore_should_output
      CHARACTER(len=*), INTENT(IN), OPTIONAL             :: file_form, file_position, file_action, &
                                                            file_status
      LOGICAL, INTENT(IN), OPTIONAL                      :: do_backup, on_file
      LOGICAL, INTENT(OUT), OPTIONAL                     :: is_new_file
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_print_key_unit_nr', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: filename, filename_bak, filename_bak_1, &
                                                            filename_bak_2
      CHARACTER(len=default_string_length)               :: my_file_action, my_file_form, &
                                                            my_file_position, my_file_status, &
                                                            outPath
      INTEGER                                            :: c_i_level, f_backup_level, i, &
                                                            my_backup_level, my_nbak, nbak, &
                                                            s_backup_level, unit_nr
      LOGICAL                                            :: do_log, found, my_do_backup, my_local, &
                                                            my_on_file, my_should_output
      TYPE(cp_iteration_info_type), POINTER              :: iteration_info
      TYPE(section_vals_type), POINTER                   :: print_key

      my_local = .FALSE.
      my_do_backup = .FALSE.
      found = .FALSE.
      res = -1
      my_file_form = "FORMATTED"
      my_file_position = "APPEND"
      my_file_action = "WRITE"
      my_file_status = "UNKNOWN"
      my_on_file = .FALSE.
      IF (PRESENT(file_form)) my_file_form = file_form
      IF (PRESENT(file_position)) my_file_position = file_position
      IF (PRESENT(file_action)) my_file_action = file_action
      IF (PRESENT(file_status)) my_file_status = file_status
      IF (PRESENT(do_backup)) my_do_backup = do_backup
      IF (PRESENT(on_file)) my_on_file = on_file
      IF (PRESENT(local)) my_local = local
      IF (PRESENT(is_new_file)) is_new_file = .FALSE.
      NULLIFY (print_key)
      IF(.NOT.(ASSOCIATED(basis_section)))CALL cp__a("input/cp_output_handling.F",858)
      IF(.NOT.(ASSOCIATED(logger)))CALL cp__a("input/cp_output_handling.F",859)
      IF(.NOT.(basis_section%ref_count > 0))CALL cp__a("input/cp_output_handling.F",860)
      IF(.NOT.(logger%ref_count > 0))CALL cp__a("input/cp_output_handling.F",861)
      my_should_output = BTEST(cp_print_key_should_output(logger%iter_info, &
                                                          basis_section, print_key_path, used_print_key=print_key), cp_p_file)
      IF (PRESENT(ignore_should_output)) my_should_output = my_should_output .OR. ignore_should_output
      IF (.NOT. my_should_output) RETURN
      IF (my_local .OR. &
          logger%para_env%mepos == logger%para_env%source) THEN

         CALL section_vals_val_get(print_key, "FILENAME", c_val=outPath)
         IF (outPath == '__STD_OUT__' .AND. .NOT. my_on_file) THEN
            res = cp_logger_get_default_unit_nr(logger, local=my_local)
         ELSE
            !
            ! complex logic to build filename:
            !   1)  Try to avoid '--' and '-.'
            !   2)  If outPath contains '/' (as in ./filename) do not prepend the project_name
            !
            ! if it is actually a full path, use it as the root
            filename = cp_print_key_generate_filename(logger, print_key, middle_name, extension, &
                                                      my_local)
            ! Give back info about a possible existence of the file if required
            IF (PRESENT(is_new_file)) THEN
               INQUIRE (FILE=filename, EXIST=found)
               is_new_file = .NOT. found
               IF (my_file_position == "REWIND") is_new_file = .TRUE.
            END IF
            ! Check is we have to log any operation performed on the file..
            do_log = .FALSE.
            IF (PRESENT(log_filename)) THEN
               do_log = log_filename
            ELSE
               CALL section_vals_val_get(print_key, "LOG_PRINT_KEY", l_val=do_log)
            END IF
            ! If required do a backup
            IF (my_do_backup) THEN
               INQUIRE (FILE=filename, EXIST=found)
               CALL section_vals_val_get(print_key, "BACKUP_COPIES", i_val=nbak)
               IF (nbak /= 0) THEN
                  iteration_info => logger%iter_info
                  s_backup_level = 0
                  IF (ASSOCIATED(print_key%ibackup)) s_backup_level = SIZE(print_key%ibackup)
                  CALL section_vals_val_get(print_key, "COMMON_ITERATION_LEVELS", i_val=c_i_level)
                  my_backup_level = MAX(1, iteration_info%n_rlevel-c_i_level+1)
                  f_backup_level = MAX(s_backup_level, my_backup_level)
                  IF (f_backup_level > s_backup_level) THEN
                     CALL reallocate(print_key%ibackup, 1, f_backup_level)
                     DO i = s_backup_level+1, f_backup_level
                        print_key%ibackup(i) = 0
                     END DO
                  END IF
                  IF (found) THEN
                     print_key%ibackup(my_backup_level) = print_key%ibackup(my_backup_level)+1
                     my_nbak = print_key%ibackup(my_backup_level)
                     ! Recent backup copies correspond to lower backup indexes
                     DO i = MIN(nbak, my_nbak), 2, -1
                        filename_bak_1 = TRIM(filename)//".bak-"//ADJUSTL(cp_to_string(i))
                        filename_bak_2 = TRIM(filename)//".bak-"//ADJUSTL(cp_to_string(i-1))
                        IF (do_log) THEN
                           unit_nr = cp_logger_get_unit_nr(logger, local=my_local)
                           IF (unit_nr > 0) &
                              WRITE (unit_nr, *) "Moving file "//TRIM(filename_bak_2)// &
                              " into file "//TRIM(filename_bak_1)//"."
                        END IF
                        INQUIRE (FILE=filename_bak_2, EXIST=found)
                        IF (.NOT. found) THEN
                           IF (do_log) THEN
                              unit_nr = cp_logger_get_unit_nr(logger, local=my_local)
                              IF (unit_nr > 0) &
                                 WRITE (unit_nr, *) "File "//TRIM(filename_bak_2)//" not existing.."
                           END IF
                        ELSE
                           CALL m_mov(TRIM(filename_bak_2), TRIM(filename_bak_1))
                        END IF
                     END DO
                     ! The last backup is always the one with index 1
                     filename_bak = TRIM(filename)//".bak-"//ADJUSTL(cp_to_string(1))
                     IF (do_log) THEN
                        unit_nr = cp_logger_get_unit_nr(logger, local=my_local)
                        IF (unit_nr > 0) &
                           WRITE (unit_nr, *) "Moving file "//TRIM(filename)//" into file "//TRIM(filename_bak)//"."
                     END IF
                     CALL m_mov(TRIM(filename), TRIM(filename_bak))
                  ELSE
                     ! Zero the backup history for this new iteration level..
                     print_key%ibackup(my_backup_level) = 0
                  END IF
               END IF
            END IF
            CALL open_file(file_name=filename, file_status=my_file_status, &
                           file_form=my_file_form, file_action=my_file_action, &
                           file_position=my_file_position, unit_number=res)
            IF (do_log) THEN
               unit_nr = cp_logger_get_unit_nr(logger, local=my_local)
               IF (unit_nr > 0) &
                  WRITE (unit_nr, *) "Writing "//TRIM(print_key%section%name)//" "// &
                  TRIM(cp_iter_string(logger%iter_info))//" to "// &
                  TRIM(filename)
            END IF
         END IF
      ELSE
         res = -1
      END IF
   END FUNCTION cp_print_key_unit_nr

! **************************************************************************************************
!> \brief should be called after you finish working with a unit obtained with
!>      cp_print_key_unit_nr, so that the file that might have been opened
!>      can be closed.
!>
!>      the inputs should be exactly the same of the corresponding
!>      cp_print_key_unit_nr
!> \param unit_nr ...
!> \param logger ...
!> \param basis_section ...
!> \param print_key_path ...
!> \param local ...
!> \param ignore_should_output ...
!> \param on_file ...
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      closes if the corresponding filename of the printkey is
!>      not __STD_OUT__
! **************************************************************************************************
   SUBROUTINE cp_print_key_finished_output(unit_nr, logger, basis_section, &
                                           print_key_path, local, ignore_should_output, on_file)
      INTEGER, INTENT(INOUT)                             :: unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: basis_section
      CHARACTER(len=*), INTENT(IN), OPTIONAL             :: print_key_path
      LOGICAL, INTENT(IN), OPTIONAL                      :: local, ignore_should_output, on_file

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_print_key_finished_output', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: outPath
      LOGICAL                                            :: my_local, my_on_file, my_should_output
      TYPE(section_vals_type), POINTER                   :: print_key

      my_local = .FALSE.
      my_on_file = .FALSE.
      NULLIFY (print_key)
      IF (PRESENT(local)) my_local = local
      IF (PRESENT(on_file)) my_on_file = on_file
      IF(.NOT.(ASSOCIATED(basis_section)))CALL cp__a("input/cp_output_handling.F",1006)
      IF(.NOT.(ASSOCIATED(logger)))CALL cp__a("input/cp_output_handling.F",1007)
      IF(.NOT.(basis_section%ref_count > 0))CALL cp__a("input/cp_output_handling.F",1008)
      IF(.NOT.(logger%ref_count > 0))CALL cp__a("input/cp_output_handling.F",1009)
      my_should_output = BTEST(cp_print_key_should_output(logger%iter_info, basis_section, &
                                                          print_key_path, used_print_key=print_key), cp_p_file)
      IF (PRESENT(ignore_should_output)) my_should_output = my_should_output .OR. ignore_should_output
      IF (my_should_output .AND. (my_local .OR. &
                                  logger%para_env%source == logger%para_env%mepos)) THEN
         CALL section_vals_val_get(print_key, "FILENAME", c_val=outPath)
         IF (my_on_file .OR. outPath .NE. '__STD_OUT__') THEN
            IF(.NOT.(unit_nr > 0))CALL cp__a("input/cp_output_handling.F",1017)
            CALL close_file(unit_nr, "KEEP")
            unit_nr = -1
         ELSE
            unit_nr = -1
         ENDIF
      END IF
      IF(.NOT.(unit_nr == -1))CALL cp__a("input/cp_output_handling.F",1024)
      unit_nr = -1
   END SUBROUTINE cp_print_key_finished_output

END MODULE cp_output_handling

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief collects all constants needed in input so that they can be used without
!>      circular dependencies
!> \par History
!>      02.2005 created [fawzi&Joost]
!>      10.2008 Teodoro Laino [tlaino] - University of Zurich
!>              Cleaned this file of all arrays of strings. Here must be kept
!>              only integer/real global constants
! **************************************************************************************************
MODULE input_constants

   USE kinds,                           ONLY: dp

   IMPLICIT NONE
   PRIVATE

   ! Tags for development/release versions
   INTEGER, PARAMETER, PUBLIC               :: id_development_version = 0, &
                                               id_release_version = 1

   ! Program Names
   INTEGER, PARAMETER, PUBLIC               :: do_atom = 1, &
                                               do_farming = 2, &
                                               do_test = 3, &
                                               do_cp2k = 4, &
                                               do_optimize_input = 5, &
                                               do_opt_basis = 6, &
                                               do_tree_mc = 7, &
                                               do_tree_mc_ana = 8, &
                                               do_swarm = 9

   INTEGER, PARAMETER, PUBLIC               :: opt_force_matching = 1

   INTEGER, PARAMETER, PUBLIC               :: do_full_density = 0, &
                                               do_spin_density = 1

   INTEGER, PARAMETER, PUBLIC               :: do_no_et = 0, &
                                               do_et_ddapc = 1, &
                                               do_et_becke = 2
   ! Method Names
   INTEGER, PARAMETER, PUBLIC               :: do_qs = 1, &
                                               do_fist = 2, &
                                               do_qmmm = 3, &
                                               do_qmmmx = 4, &
                                               do_eip = 5, &
                                               do_mixed = 6

   ! QMMM
   REAL(KIND=dp), PARAMETER, PUBLIC         :: RADIUS_QMMM_DEFAULT = 0.80_dp, & ! Angstrom
                                               ALPHA_IMOMM_DEFAULT = 1.38_dp, &
                                               CHARGE_SCALE_FACTOR = 1.00_dp
   ! BAND
   INTEGER, PARAMETER, PUBLIC               :: band_md_opt = 0, &
                                               band_diis_opt = 1

   INTEGER, PARAMETER, PUBLIC               :: do_band_cartesian = 0, &
                                               do_band_collective = 1
   ! Weight_types
   INTEGER, PARAMETER, PUBLIC               :: weight_type_unit = 0, &
                                               weight_type_mass = 1

   ! diag methods
   INTEGER, PARAMETER, PUBLIC               :: do_diag_syevd = 1, &
                                               do_diag_syevx = 2
   ! initial matrix methods methods
   INTEGER, PARAMETER, PUBLIC               :: do_mat_random = 1, &
                                               do_mat_read = 2

   INTEGER, PARAMETER, PUBLIC               :: do_constr_none = 1, &
                                               do_constr_atomic = 2, &
                                               do_constr_molec = 3

   INTEGER, PARAMETER, PUBLIC               :: do_cell_cp2k = 1, &
                                               do_cell_xsc = 2

   INTEGER, PARAMETER, PUBLIC               :: rmsd_all = 0, &
                                               rmsd_list = 1, &
                                               rmsd_weightlist = 2

   INTEGER, PARAMETER, PUBLIC               :: do_rep_interleaved = 0, &
                                               do_rep_blocked = 1

   ! MD Ensemble
   INTEGER, PARAMETER, PUBLIC               :: nve_ensemble = 1, &
                                               nvt_ensemble = 2, &
                                               npt_i_ensemble = 3, &
                                               npt_f_ensemble = 4, &
                                               nph_uniaxial_ensemble = 5, &
                                               nph_uniaxial_damped_ensemble = 6, &
                                               nph_ensemble = 7, &
                                               isokin_ensemble = 8, &
                                               reftraj_ensemble = 9, &
                                               langevin_ensemble = 10, &
                                               npe_f_ensemble = 11, &
                                               npe_i_ensemble = 12, &
                                               nvt_adiabatic_ensemble = 13

   INTEGER, PARAMETER, PUBLIC               :: do_thermo_same_as_part = 100, &
                                               do_thermo_nose = 101, &
                                               do_thermo_csvr = 102, &
                                               do_thermo_gle = 103, &
                                               do_thermo_al = 104

   ! Run Types
   INTEGER, PARAMETER, PUBLIC               :: none_run = 1, &
                                               energy_run = 2, &
                                               energy_force_run = 3, &
                                               mol_dyn_run = 4, &
                                               geo_opt_run = 5, &
                                               mon_car_run = 6, &
                                               electronic_spectra_run = 7, &
                                               debug_run = 8, &
                                               bsse_run = 9, &
                                               linear_response_run = 10, &
                                               pint_run = 11, &
                                               vib_anal = 12, &
                                               do_band = 13, &
                                               cell_opt_run = 14, &
                                               real_time_propagation = 15, &
                                               ehrenfest = 16, &
                                               do_tamc = 17, &
                                               tree_mc_run = 18, &
                                               driver_run = 19

   ! Run Types of Atom Code
   INTEGER, PARAMETER, PUBLIC               :: atom_no_run = 1, &
                                               atom_energy_run = 2, &
                                               atom_basis_run = 3, &
                                               atom_pseudo_run = 4

   ! Integral types in Atom Code
   INTEGER, PARAMETER, PUBLIC               :: do_analytic = 1, &
                                               do_semi_analytic = 2, &
                                               do_numeric = 3
   ! Basis set type for Atom Code
   INTEGER, PARAMETER, PUBLIC               :: gaussian = 1, &
                                               geometrical_gto = 2, &
                                               contracted_gto = 3, &
                                               slater = 4, &
                                               numerical = 5
   ! Method type for Atom Code
   INTEGER, PARAMETER, PUBLIC               :: do_rks_atom = 1, &
                                               do_uks_atom = 2, &
                                               do_rhf_atom = 3, &
                                               do_uhf_atom = 4, &
                                               do_rohf_atom = 5
   ! Relativistic correction for Atom Code
   INTEGER, PARAMETER, PUBLIC               :: do_nonrel_atom = 100, &
                                               do_zoramp_atom = 101, &
                                               do_sczoramp_atom = 102, &
                                               do_dkh0_atom = 110, &
                                               do_dkh1_atom = 111, &
                                               do_dkh2_atom = 112, &
                                               do_dkh3_atom = 113, &
                                               do_dkh4_atom = 114, &
                                               do_dkh5_atom = 115

   ! Pseudopotential type for Atom Code
   INTEGER, PARAMETER, PUBLIC               :: no_pseudo = 0, &
                                               gth_pseudo = 1, &
                                               upf_pseudo = 2, &
                                               ecp_pseudo = 3
   ! Confinement type for Atom Code
   INTEGER, PARAMETER, PUBLIC               :: no_conf = 0, &
                                               poly_conf = 1, &
                                               barrier_conf = 2
   ! BAND
   INTEGER, PARAMETER, PUBLIC               :: do_b_neb = 0, &
                                               do_it_neb = 1, &
                                               do_ci_neb = 2, &
                                               do_d_neb = 3, &
                                               do_sm = 4, &
                                               do_eb = 5

   INTEGER, PARAMETER, PUBLIC               :: pot_neb_full = 0, &
                                               pot_neb_fe = 1, &
                                               pot_neb_me = 2

   INTEGER, PARAMETER, PUBLIC               :: cholesky_off = 0, &
                                               cholesky_reduce = 1, &
                                               cholesky_restore = 2, &
                                               cholesky_inverse = 3, &
                                               cholesky_dbcsr = 4

   INTEGER, PARAMETER, PUBLIC               :: atomic_guess = 1, &
                                               restart_guess = 2, &
                                               random_guess = 3, &
                                               core_guess = 4, &
                                               history_guess = 5, &
                                               no_guess = 6, &
                                               mopac_guess = 7, &
                                               sparse_guess = 8, &
                                               almo_scf_guess = 9, &
                                               molecular_guess = 10

   ! Projection formulas for the maximum overlap method
   INTEGER, PARAMETER, PUBLIC               :: momproj_norm = 0, &
                                               momproj_sum = 1

   INTEGER, PARAMETER, PUBLIC               :: almo_frz_none = 0, &
                                               almo_frz_isolated = 1, &
                                               almo_frz_crystal = 2, &
                                               almo_frz_both = 3

   INTEGER, PARAMETER, PUBLIC               :: pw_interp = 1, &
                                               spline3_nopbc_interp = 2, &
                                               spline3_pbc_interp = 3

   INTEGER, PARAMETER, PUBLIC               :: do_gapw_gcs = 1, &
                                               do_gapw_gct = 2, &
                                               do_gapw_log = 3

   INTEGER, PARAMETER, PUBLIC               :: do_pwgrid_spherical = 1, &
                                               do_pwgrid_ns_fullspace = 2, &
                                               do_pwgrid_ns_halfspace = 3

   INTEGER, PARAMETER, PUBLIC               :: wfi_use_guess_method_nr = 0, &
                                               wfi_use_prev_p_method_nr = 1, &
                                               wfi_use_prev_rho_r_method_nr = 2, &
                                               wfi_linear_wf_method_nr = 3, &
                                               wfi_linear_p_method_nr = 4, &
                                               wfi_linear_ps_method_nr = 5, &
                                               wfi_use_prev_wf_method_nr = 6, &
                                               wfi_ps_method_nr = 7, &
                                               wfi_frozen_method_nr = 8, &
                                               wfi_aspc_nr = 9

   INTEGER, PARAMETER, PUBLIC               :: do_method_undef = 0, &
                                               do_method_gapw = 1, &
                                               do_method_gapw_xc = 2, &
                                               do_method_gpw = 3, &
                                               do_method_lrigpw = 4, &
                                               do_method_mndo = 7, &
                                               do_method_am1 = 8, &
                                               do_method_pm3 = 9, &
                                               do_method_pm6 = 10, &
                                               do_method_mndod = 11, &
                                               do_method_pdg = 12, &
                                               do_method_rm1 = 13, &
                                               do_method_dftb = 14, &
                                               do_method_pchg = 15, &
                                               do_method_pnnl = 16, &
                                               do_method_ofgpw = 17, &
                                               do_method_scptb = 18

   INTEGER, PARAMETER, PUBLIC               :: do_multipole_section_on = 0, &
                                               do_multipole_section_off = 1

   INTEGER, PARAMETER, PUBLIC               :: do_ppl_analytic = 0, &
                                               do_ppl_grid = 1

   INTEGER, PARAMETER, PUBLIC               :: do_se_IS_kdso = 0, &
                                               do_se_IS_kdso_d = 1, &
                                               do_se_IS_slater = 2

   INTEGER, PARAMETER, PUBLIC               :: do_se_lr_none = 0, &
                                               do_se_lr_ewald = 1, &
                                               do_se_lr_ewald_r3 = 2, &
                                               do_se_lr_ewald_gks = 3

   INTEGER, PARAMETER, PUBLIC               :: do_par_atom = 0, &
                                               do_par_grid = 1

   INTEGER, PARAMETER, PUBLIC               :: do_qmmm_center_every_step = 0, &
                                               do_qmmm_center_setup_only = 1, &
                                               do_qmmm_center_never = 2

   INTEGER, PARAMETER, PUBLIC                    :: do_qmmm_center_max_minus_min = 0, &
                                                    do_qmmm_center_pbc_aware = 1
   ! Thermostats
   INTEGER, PARAMETER, PUBLIC               :: do_thermo_no_communication = 0, &
                                               do_thermo_communication = 1, &
                                               do_thermo_only_master = 2

   ! pint coordinate transformations
   INTEGER, PARAMETER, PUBLIC               :: transformation_normal = 1, &
                                               transformation_stage = 2

   ! pint propagator mode
   INTEGER, PARAMETER, PUBLIC               :: propagator_pimd = 1, &
                                               propagator_rpmd = 2
   ! pint propagator mode
   INTEGER, PARAMETER, PUBLIC               :: integrate_numeric = 1, &
                                               integrate_exact = 2

   ! piglet S-matrix initialisations
   INTEGER, PARAMETER, PUBLIC               :: matrix_init_cholesky = 1, &
                                               matrix_init_diagonal = 2

   ! helium pbc unit cell shapes
   INTEGER, PARAMETER, PUBLIC               :: helium_cell_shape_none = 0, &
                                               helium_cell_shape_cube = 1, &
                                               helium_cell_shape_octahedron = 2

   ! helium-solute interaction potentials
   INTEGER, PARAMETER, PUBLIC               :: helium_solute_intpot_none = 0, &
                                               helium_solute_intpot_mwater = 1

   ! helium force selection
   INTEGER, PARAMETER, PUBLIC               :: helium_forces_average = 0, &
                                               helium_forces_last = 1
   ! superfluid density, estimators
   INTEGER, PARAMETER, PUBLIC               :: estimator_none = 1, &
                                               estimator_parea = 2, &
                                               estimator_wnumber = 3, &
                                               estimator_pcycle = 4, &
                                               estimator_weighted = 5, &
                                               estimator_normalization = 6

   ! superfluid density, denominator selector
   INTEGER, PARAMETER, PUBLIC               :: denominator_unity = 1, &
                                               denominator_natoms = 2, &
                                               denominator_inertia = 3, &
                                               denominator_rperp2 = 4

   ! helium-sampling algorithms
   INTEGER, PARAMETER, PUBLIC               :: helium_sampling_ceperley = 0

   ! distribution types for sampling path lengths
   INTEGER, PARAMETER, PUBLIC               :: helium_mdist_singlev = 1, &
                                               helium_mdist_uniform = 2, &
                                               helium_mdist_linear = 3, &
                                               helium_mdist_quadratic = 4, &
                                               helium_mdist_exponential = 5, &
                                               helium_mdist_gaussian = 6

   INTEGER, PARAMETER, PUBLIC               :: perm_plain = 1, &
                                               perm_cycle = 2

   ! Free Energy methods
   INTEGER, PARAMETER, PUBLIC               :: do_fe_meta = 0, &
                                               do_fe_ui = 1, &
                                               do_fe_ac = 2

   INTEGER, PARAMETER, PUBLIC               :: do_wall_none = 0, &
                                               do_wall_reflective = 1, &
                                               do_wall_quadratic = 2, &
                                               do_wall_quartic = 3, &
                                               do_wall_gaussian = 4

   INTEGER, PARAMETER, PUBLIC               :: do_wall_p = 0, &
                                               do_wall_m = 1
   ! Output formats
   INTEGER, PARAMETER, PUBLIC               :: dump_atomic = 1, &
                                               dump_xmol = 2, &
                                               dump_dcd = 3, &
                                               dump_pdb = 4, &
                                               dump_dcd_aligned_cell = 5

   INTEGER, PARAMETER, PUBLIC               :: fmt_id_xyz = 1, &
                                               fmt_id_pdb = 2

   INTEGER, PARAMETER, PUBLIC               :: do_qmmm_none = 0, &
                                               do_qmmm_coulomb = 1, &
                                               do_qmmm_gauss = 2, &
                                               do_qmmm_swave = 3, &
                                               do_qmmm_pcharge = 4, &
                                               do_qmmm_link_imomm = 1, &
                                               do_qmmm_link_gho = 2, &
                                               do_qmmm_link_pseudo = 3, &
                                               do_qmmm_center_none = 0, &
                                               do_qmmm_center_grid = 1, &
                                               do_qmmm_wall_none = 0, &
                                               do_qmmm_wall_reflective = 1, &
                                               do_qmmm_wall_quadratic = 2, &
                                               do_qmmm_image_calcmatrix = 0, &
                                               do_qmmm_image_iter = 1

   INTEGER, PARAMETER, PUBLIC               :: calc_always = 0, &
                                               calc_once = 1, &
                                               calc_once_done = 2

   INTEGER, PARAMETER, PUBLIC               :: do_fm_mom_conserv_none = 0, &
                                               do_fm_mom_conserv_equal_f = 1, &
                                               do_fm_mom_conserv_equal_a = 2

   INTEGER, PARAMETER, PUBLIC               :: do_fm_mom_conserv_core = 0, &
                                               do_fm_mom_conserv_qm = 1, &
                                               do_fm_mom_conserv_buffer = 2

   INTEGER, PARAMETER, PUBLIC               :: do_region_global = 0, &
                                               do_region_molecule = 1, &
                                               do_region_massive = 2, &
                                               do_region_none = 3, &
                                               do_region_defined = 4, &
                                               do_bondparm_covalent = 0, &
                                               do_bondparm_vdw = 1, &
                                               do_skip_11 = 0, &
                                               do_skip_12 = 1, &
                                               do_skip_13 = 2, &
                                               do_skip_14 = 3, &
                                               do_conn_psf = 1, &
                                               do_conn_psf_u = 2, &
                                               do_conn_generate = 3, &
                                               do_conn_off = 4, &
                                               do_conn_mol_set = 5, &
                                               do_conn_g87 = 6, &
                                               do_conn_g96 = 7, &
                                               do_conn_amb7 = 8, &
                                               do_conn_user = 9, &
                                               do_coord_off = 0, &
                                               do_coord_pdb = 1, &
                                               do_coord_xyz = 2, &
                                               do_coord_g96 = 3, &
                                               do_coord_crd = 4, &
                                               do_coord_cif = 5, &
                                               do_coord_xtl = 6, &
                                               do_coord_cp2k = 7

   INTEGER, PARAMETER, PUBLIC               :: do_stress_none = 0, &
                                               do_stress_analytical = 1, &
                                               do_stress_numerical = 2, &
                                               do_stress_diagonal_anal = 3, &
                                               do_stress_diagonal_numer = 4

   INTEGER, PARAMETER, PUBLIC               :: do_mc_traditional = 0, &
                                               do_mc_gemc_nvt = 1, &
                                               do_mc_gemc_npt = 2, &
                                               do_mc_virial = 3

   INTEGER, PARAMETER, PUBLIC               :: do_loc_none = 0, &
                                               do_loc_jacobi = 1, &
                                               do_loc_crazy = 2, &
                                               do_loc_direct = 3, &
                                               do_loc_l1_norm_sd = 4

   INTEGER, PARAMETER, PUBLIC               :: do_loc_min = 0, &
                                               do_loc_max = 1, &
                                               op_loc_berry = 0, &
                                               op_loc_boys = 1, &
                                               op_loc_pipek = 2, &
                                               op_loc_l1_norm = 3, &
                                               state_loc_all = 0, &
                                               state_loc_range = 1, &
                                               state_loc_list = 2, &
                                               energy_loc_range = 3, &
                                               state_loc_none = 4

   INTEGER, PARAMETER, PUBLIC               :: do_loc_homo = 0, &
                                               do_loc_lumo = 1, &
                                               do_loc_both = 2

   INTEGER, PARAMETER, PUBLIC               :: orb_s = 0, &
                                               orb_px = 1, &
                                               orb_py = 2, &
                                               orb_pz = 3, &
                                               orb_dxy = 4, &
                                               orb_dyz = 5, &
                                               orb_dzx = 6, &
                                               orb_dx2 = 7, &
                                               orb_dy2 = 8, &
                                               orb_dz2 = 9

   INTEGER, PARAMETER, PUBLIC               :: use_mom_ref_com = 0, &
                                               use_mom_ref_coac = 1, &
                                               use_mom_ref_user = 2, &
                                               use_mom_ref_zero = 3

   INTEGER, PARAMETER, PUBLIC               :: current_gauge_r = 1, &
                                               current_gauge_r_and_step_func = 2, &
                                               current_gauge_atom = 3, &
                                               current_orb_center_wannier = 1, &
                                               current_orb_center_common = 2, &
                                               current_orb_center_atom = 3, &
                                               current_orb_center_box = 4

   INTEGER, PARAMETER, PUBLIC               :: ot_mini_sd = 1, ot_mini_cg = 2, ot_mini_diis = 3, ot_mini_broyden = 4
   INTEGER, PARAMETER, PUBLIC               :: ot_algo_taylor_or_diag = 1, ot_algo_irac = 2
   INTEGER, PARAMETER, PUBLIC               :: ot_chol_irac = 1, ot_poly_irac = 2, ot_lwdn_irac = 3
   INTEGER, PARAMETER, PUBLIC               :: ls_none = 1, ls_2pnt = 2, ls_3pnt = 3, ls_gold = 4, ls_fit = 5
   INTEGER, PARAMETER, PUBLIC               :: ot_precond_none = 0, &
                                               ot_precond_full_single = 1, &
                                               ot_precond_full_kinetic = 2, &
                                               ot_precond_s_inverse = 3, &
                                               ot_precond_full_all = 5, &
                                               ot_precond_full_single_inverse = 6
   INTEGER, PARAMETER, PUBLIC               :: ot_precond_solver_default = 1, &
                                               ot_precond_solver_inv_chol = 2, &
                                               ot_precond_solver_direct = 3, &
                                               ot_precond_solver_update = 4

   INTEGER, PARAMETER, PUBLIC               :: diag_standard = 1, &
                                               diag_ot = 2, &
                                               diag_block_krylov = 3, &
                                               diag_block_davidson = 4, &
                                               diag_filter_matrix = 5

   INTEGER, PUBLIC, PARAMETER               :: default_lbfgs_method_id = 1, &
                                               default_bfgs_method_id = 2, &
                                               default_cg_method_id = 3

   INTEGER, PUBLIC, PARAMETER               :: default_minimization_method_id = 1, &
                                               default_ts_method_id = 2, &
                                               default_cell_method_id = 3, &
                                               default_shellcore_method_id = 4

   INTEGER, PUBLIC, PARAMETER               :: none_ts_method_id = 0, &
                                               default_dimer_method_id = 1

   INTEGER, PUBLIC, PARAMETER               :: do_first_rotation_step = 0, &
                                               do_second_rotation_step = 1, &
                                               do_third_rotation_step = 2

   INTEGER, PUBLIC, PARAMETER               :: default_cell_geo_opt_id = 0, &
                                               default_cell_md_id = 1, &
                                               default_cell_direct_id = 2

   INTEGER, PARAMETER, PUBLIC               :: xc_funct_no_shortcut = 0, &
                                               xc_funct_blyp = 1, &
                                               xc_funct_pade = 2, &
                                               xc_funct_pbe = 3, &
                                               xc_funct_tpss = 4, &
                                               xc_funct_hcth120 = 5, &
                                               xc_funct_olyp = 6, &
                                               xc_funct_bp = 7, &
                                               xc_none = 8, &
                                               xc_funct_xwpbe = 9, &
                                               xc_funct_b3lyp = 10, &
                                               xc_funct_pbe0 = 11, &
                                               xc_funct_beefvdw = 12
   INTEGER, PARAMETER, PUBLIC               :: sic_none = 0, &
                                               sic_mauri_us = 1, &
                                               sic_mauri_spz = 2, &
                                               sic_ad = 3, &
                                               sic_eo = 4
   INTEGER, PARAMETER, PUBLIC               :: sic_list_all = 1, &
                                               sic_list_unpaired = 2
   INTEGER, PARAMETER, PUBLIC               :: tddfpt_singlet = 0, &
                                               tddfpt_triplet = 1
   INTEGER, PARAMETER, PUBLIC               :: tddfpt_lanczos = 0, &
                                               tddfpt_davidson = 1
   INTEGER, PARAMETER, PUBLIC               :: oe_none = 0, &
                                               oe_lb = 1, &
                                               oe_gllb = 2, &
                                               oe_saop = 3, &
                                               oe_sic = 4
   INTEGER, PARAMETER, PUBLIC               :: no_excitations = 0, &
                                               tddfpt_excitations = 1
   INTEGER, PARAMETER, PUBLIC               :: xc_vdw_fun_none = 100, &
                                               xc_vdw_fun_pairpot = 101, &
                                               xc_vdw_fun_nonloc = 102
   INTEGER, PARAMETER, PUBLIC               :: vdw_pairpot_dftd2 = 1, &
                                               vdw_pairpot_dftd3 = 2, &
                                               vdw_pairpot_dftd3bj = 3
   INTEGER, PARAMETER, PUBLIC               :: vdw_nl_DRSLL = 1, &
                                               vdw_nl_LMKLL = 2, &
                                               vdw_nl_RVV10 = 3

   ! Input constants for relativistic calculations
   INTEGER, PARAMETER, PUBLIC               :: rel_none = 0, &
                                               rel_dkh = 1, &
                                               rel_zora = 2
   INTEGER, PARAMETER, PUBLIC               :: rel_zora_full = 100, &
                                               rel_zora_mp = 101, &
                                               rel_sczora_mp = 102
   INTEGER, PARAMETER, PUBLIC               :: rel_trans_full = 0, &
                                               rel_trans_molecule = 1, &
                                               rel_trans_atom = 2
   INTEGER, PARAMETER, PUBLIC               :: rel_pot_full = 0, &
                                               rel_pot_erfc = 1

   INTEGER, PARAMETER, PUBLIC               :: do_add = 1, &
                                               do_remove = 2

   INTEGER, PARAMETER, PUBLIC               :: xc_pot_energy_none = 0, &
                                               xc_pot_energy_xc_functional = 1, &
                                               xc_pot_energy_sum_eigenvalues = 2

   INTEGER, PARAMETER, PUBLIC               :: xc_pot_none = 1000, &
                                               xc_pot_saop = 1001

   ! Hirshfeld partitioning
   INTEGER, PARAMETER, PUBLIC               :: shape_function_gaussian = 1, &
                                               shape_function_density = 2
   INTEGER, PARAMETER, PUBLIC               :: ref_charge_atomic = 100, &
                                               ref_charge_mulliken = 200

   ! MAO
   INTEGER, PARAMETER, PUBLIC               :: mao_basis_orb = 2000, &
                                               mao_basis_prim = 2001, &
                                               mao_basis_ext = 2002

   ! X-Ray spectra Calculation
   INTEGER, PARAMETER, PUBLIC               :: xas_none = 0, &
                                               xas_tp_hh = 1, &
                                               xas_tp_fh = 2, &
                                               xes_tp_val = 3, &
                                               xas_tp_xhh = 4, &
                                               xas_tp_xfh = 5, &
                                               xas_dscf = 6, &
                                               xas_tp_flex = 7
   INTEGER, PARAMETER, PUBLIC               :: xas_1s_type = 1, &
                                               xas_2s_type = 2, &
                                               xas_2p_type = 3, &
                                               xas_3s_type = 4, &
                                               xas_3p_type = 5, &
                                               xas_3d_type = 6, &
                                               xas_4s_type = 7, &
                                               xas_4p_type = 8, &
                                               xas_4d_type = 9, &
                                               xas_4f_type = 10
   INTEGER, PARAMETER, PUBLIC               :: xas_dip_len = 1, &
                                               xas_dip_vel = 2, &
                                               xas_dip_len2 = 3
   INTEGER, PARAMETER, PUBLIC               :: xas_scf_default = 0, &
                                               xas_scf_general = 1

   ! Linear Response for properties
   INTEGER, PARAMETER, PUBLIC               :: lr_none = 0, &
                                               lr_chemshift = 1, &
                                               lr_gtensor = 2, &
                                               lr_current = 3

   ! EIP models
   INTEGER, PARAMETER, PUBLIC               :: use_bazant_eip = 1, &
                                               use_lenosky_eip = 2

   ! ddapc restraint forms
   INTEGER, PARAMETER, PUBLIC               :: do_ddapc_restraint = 773, &
                                               do_ddapc_constraint = 774

   ! outer scf types
   INTEGER, PARAMETER, PUBLIC               :: outer_scf_ddapc_constraint = 123, &
                                               outer_scf_s2_constraint = 124, &
                                               outer_scf_becke_constraint = 125, &
                                               outer_scf_none = 126

   ! outer scf optimizers
   INTEGER, PARAMETER, PUBLIC               :: outer_scf_optimizer_sd = 1001, &
                                               outer_scf_optimizer_diis = 1002, &
                                               outer_scf_optimizer_none = 1003, &
                                               outer_scf_optimizer_bisect = 1004

   ! s2 restraint forms
   INTEGER, PARAMETER, PUBLIC               :: do_s2_restraint = 872, &
                                               do_s2_constraint = 873

   ! ROKS schemes
   INTEGER, PARAMETER, PUBLIC               :: general_roks = 1, &
                                               high_spin_roks = 2

   ! mixing of force envs
   INTEGER, PARAMETER, PUBLIC               :: mix_linear_combination = 701, &
                                               mix_minimum = 702, &
                                               mix_coupled = 703, &
                                               mix_restrained = 704, &
                                               mix_generic = 705, &
                                               mix_mymix = 706

   ! cost models for distribution 2d
   INTEGER, PARAMETER, PUBLIC               :: model_block_count = 453, &
                                               model_block_surface = 454, &
                                               model_block_lmax = 455

   ! clustering methods
   INTEGER, PARAMETER, PUBLIC               :: clustering_none = 0, &
                                               clustering_box = 1, &
                                               clustering_h_on_heavier = 2, &
                                               clustering_small_on_larger = 3

   ! HFX potential type
   INTEGER, PARAMETER, PUBLIC               :: do_hfx_potential_coulomb = 1, &
                                               do_hfx_potential_short = 2, &
                                               do_hfx_potential_long = 3, &
                                               do_hfx_potential_mix_cl = 4, &
                                               do_hfx_potential_gaussian = 5, &
                                               do_hfx_potential_mix_lg = 6, &
                                               do_hfx_potential_id = 7, &
                                               do_hfx_potential_truncated = 8, &
                                               do_hfx_potential_mix_cl_trunc = 9, &
                                               do_mp2_potential_TShPSC = 1009

   ! HFX periodic number of shells
   INTEGER, PARAMETER, PUBLIC               :: do_hfx_auto_shells = -1

   ! adiabatic hybrid type
   INTEGER, PARAMETER, PUBLIC               :: do_adiabatic_hybrid_mcy3 = 1

   ! adiabatic model type
   INTEGER, PARAMETER, PUBLIC               :: do_adiabatic_model_pade = 1

   ! HFX evaluation type
   INTEGER, PARAMETER, PUBLIC               :: hfx_do_eval_energy = 1, &
                                               hfx_do_eval_forces = 2

   ! mode selctive vibrational analysis
   INTEGER, PARAMETER, PUBLIC               :: ms_guess_bfgs = 1, &
                                               ms_guess_atomic = 2, &
                                               ms_guess_restart = 3, &
                                               ms_guess_restart_vec = 4, &
                                               ms_guess_molden = 5

   ! BASIS SET PARAMETER
   ! This parameter identifies basis sets
   INTEGER, PARAMETER, PUBLIC               :: use_orb_basis_set = 1, &
                                               use_aux_fit_basis_set = 2, &
                                               use_aux_basis_set = 3, &
                                               use_ri_aux_basis_set = 4, &
                                               use_lri_basis_set = 5
   ! ADMM PARAMETER
   ! This parameter defines the method to use for the purification
   INTEGER, PARAMETER, PUBLIC               :: do_admm_purify_none = 1, &
                                               do_admm_purify_cauchy = 2, &
                                               do_admm_purify_cauchy_subspace = 3, &
                                               do_admm_purify_mo_diag = 4, &
                                               do_admm_purify_mo_no_diag = 5, &
                                               do_admm_purify_mcweeny = 6, &
                                               do_admm_purify_none_dm = 7

   ! This parameter defines when to use blocking
   INTEGER, PARAMETER, PUBLIC               :: do_admm_basis_projection = 0, &
                                               do_admm_blocking_purify_full = 1, &
                                               do_admm_blocked_projection = 2, &
                                               do_admm_charge_constrained_projection = 3

   ! This parameter defines when to use blocking
   INTEGER, PARAMETER, PUBLIC               :: do_admm_aux_exch_func_default = 0, &
                                               do_admm_aux_exch_func_pbex = 1, &
                                               do_admm_aux_exch_func_none = 2, &
                                               do_admm_aux_exch_func_opt = 3, &
                                               do_admm_aux_exch_func_bee = 4

   ! Define scaling of the exchange correction
   INTEGER, PARAMETER, PUBLIC               :: do_admm_exch_scaling_none = 0, &
                                               do_admm_exch_scaling_merlot = 1

   ! Define DFT+U method identifier codes
   INTEGER, PARAMETER, PUBLIC               :: plus_u_lowdin = 1, &
                                               plus_u_mulliken = 2, &
                                               plus_u_mulliken_charges = 3

   ! Define identifier codes for smearing method
   INTEGER, PARAMETER, PUBLIC               :: smear_energy_window = 1, &
                                               smear_fermi_dirac = 2, &
                                               smear_list = 3

   ! Define identifier codes for mixing method

   INTEGER, PARAMETER, PUBLIC               :: no_mix = 0, &
                                               direct_p_mix = 1, &
                                               kerker_mix = 2, &
                                               pulay_mix = 3, &
                                               broy_mix = 4, &
                                               broy_mix_new = 5, &
                                               multisec_mix = 6

   ! DIAGONALIZATION library
   INTEGER, PARAMETER, PUBLIC               :: do_diag_sl = 1, &
                                               do_diag_sl2 = 2, &
                                               do_diag_elpa = 3
   ! ELPA kernel
   INTEGER, PARAMETER, PUBLIC               :: elpa_gen = 1, &
                                               elpa_gen_sim = 2, &
                                               elpa_bgp = 3, &
                                               elpa_bgq = 4, &
                                               elpa_sse = 5, &
                                               elpa_avx_b2 = 6, &
                                               elpa_avx_b4 = 7, &
                                               elpa_avx_b6 = 8

   ! FFT library
   ! these might need sync with fft_lib.F
   INTEGER, PARAMETER, PUBLIC               :: do_fft_sg = 1, &
                                               do_fft_fftw3 = 3
   INTEGER, PARAMETER, PUBLIC               :: fftw_plan_estimate = 1, &
                                               fftw_plan_measure = 2, &
                                               fftw_plan_patient = 3, &
                                               fftw_plan_exhaustive = 4

   !real time propagation
   INTEGER, PARAMETER, PUBLIC               :: do_taylor = 1, &
                                               do_pade = 2, &
                                               do_arnoldi = 3, &
                                               do_bch = 4

   INTEGER, PARAMETER, PUBLIC               :: do_etrs = 1, &
                                               do_cn = 2, &
                                               do_em = 3

   INTEGER, PARAMETER, PUBLIC               :: use_scf_wfn = 1, &
                                               use_restart_wfn = 2, &
                                               use_rt_restart = 3

   INTEGER, PARAMETER, PUBLIC               :: constant_env = 1, &
                                               gaussian_env = 2, &
                                               ramp_env = 3, &
                                               custom_env = 4

   ! how to solve polarizable force fields
   INTEGER, PARAMETER, PUBLIC               :: do_fist_pol_none = 1, &
                                               do_fist_pol_sc = 2, &
                                               do_fist_pol_cg = 3

   ! some dm_ls_scf parameters
   INTEGER, PARAMETER, PUBLIC               :: ls_s_preconditioner_none = 7, &
                                               ls_s_preconditioner_atomic = 13, &
                                               ls_s_preconditioner_molecular = 17

   INTEGER, PARAMETER, PUBLIC               :: ls_cluster_atomic = 111, &
                                               ls_cluster_molecular = 112, &
                                               ls_cluster_pao = 113

   INTEGER, PARAMETER, PUBLIC               :: ls_s_inversion_hotelling = 3, &
                                               ls_s_inversion_sign_sqrt = 5, &
                                               ls_s_inversion_none = 6

   INTEGER, PARAMETER, PUBLIC               :: ls_scf_ns = 17, ls_scf_trs4 = 18, &
                                               ls_scf_tc2 = 19, ls_scf_pexsi = 20

   INTEGER, PARAMETER, PUBLIC               :: ls_scf_line_search_3point = 3, &
                                               ls_scf_line_search_3point_2d = 6
   ! some ZMP paramenters
   INTEGER, PARAMETER, PUBLIC               :: use_coulomb = 1, &
                                               use_diff = 2, &
                                               use_no = 3

   ! parameters for ALMO methods
   INTEGER, PARAMETER, PUBLIC   :: almo_domain_layout_orbital = 1, &
                                   almo_domain_layout_atomic = 10, &
                                   almo_domain_layout_molecular = 100

   INTEGER, PARAMETER, PUBLIC   :: almo_mat_distr_atomic = 1, &
                                   almo_mat_distr_molecular = 2

   INTEGER, PARAMETER, PUBLIC   :: almo_constraint_block_diagonal = 0, &
                                   almo_constraint_distance = 1, &
                                   almo_constraint_ao_overlap = 2

   ! methods to treat electron delocalization
   INTEGER, PARAMETER, PUBLIC   :: almo_deloc_none = 0, &
                                   almo_deloc_x = 10, &
                                   almo_deloc_xalmo_x = 11, &
                                   almo_deloc_xk = 12, &
                                   almo_deloc_xalmo_scf = 13, &
                                   almo_deloc_xalmo_1diag = 14, &
                                   almo_deloc_scf = 100, &
                                   almo_deloc_x_then_scf = 110

   INTEGER, PARAMETER, PUBLIC   :: almo_scf_dm_sign = 1, &
                                   almo_scf_diag = 2, &
                                   almo_scf_pcg = 3

   ! optimizer parameters
   INTEGER, PARAMETER, PUBLIC   :: cg_zero = 0, &
                                   cg_polak_ribiere = 1, &
                                   cg_fletcher_reeves = 2, &
                                   cg_hestenes_stiefel = 3, &
                                   cg_fletcher = 4, &
                                   cg_liu_storey = 5, &
                                   cg_dai_yuan = 6, &
                                   cg_hager_zhang = 7
   INTEGER, PARAMETER, PUBLIC   :: optimizer_diis = 1, &
                                   optimizer_pcg = 2
   INTEGER, PARAMETER, PUBLIC   :: prec_zero = 0, &
                                   prec_default = -1, &
                                   prec_ks_plus_s = 4
   INTEGER, PARAMETER, PUBLIC   :: xalmo_case_block_diag = 0, &
                                   xalmo_case_fully_deloc = 1, &
                                   xalmo_case_normal = -1

   ! parameters for CT methods
   INTEGER, PARAMETER, PUBLIC   :: tensor_orthogonal = 1, &
                                   tensor_up_down = 2
   INTEGER, PARAMETER, PUBLIC   :: virt_full = 1, &
                                   virt_minimal = 2, &
                                   virt_occ_size = 3, &
                                   virt_number = 4

   ! some MP2 parameters
   INTEGER, PARAMETER, PUBLIC               :: mp2_method_none = 0, &
                                               mp2_method_laplace = 2, &
                                               mp2_method_direct = 4, &
                                               mp2_method_gpw = 6, &
                                               ri_mp2_method_gpw = 7, &
                                               ri_rpa_method_gpw = 8, &
                                               ri_mp2_laplace = 9, &
                                               mp2_ri_optimize_basis = 10, &
                                               wfc_mm_style_gemm = 11, &
                                               wfc_mm_style_syrk = 12, &
                                               ri_coulomb = 13, &
                                               ri_overlap = 14

   ! G0W0 parameter
   INTEGER, PARAMETER, PUBLIC               :: ri_rpa_g0w0_crossing_none = 0, &
                                               ri_rpa_g0w0_crossing_z_shot = 1, &
                                               ri_rpa_g0w0_crossing_newton = 2

   ! periodic RESP parameters
   INTEGER, PARAMETER, PUBLIC               :: do_resp_x_dir = 0, &
                                               do_resp_y_dir = 1, &
                                               do_resp_z_dir = 2, &
                                               do_resp_minus_x_dir = 3, &
                                               do_resp_minus_y_dir = 4, &
                                               do_resp_minus_z_dir = 5, &
                                               use_cambridge_vdw_radii = 0, &
                                               use_uff_vdw_radii = 1

   ! basis optimization parameters
   INTEGER, PARAMETER, PUBLIC               :: do_opt_all = 0, &
                                               do_opt_none = 1, &
                                               do_opt_coeff = 2, &
                                               do_opt_exps = 3

   ! LRI paramters for inverse of overlap
   INTEGER, PARAMETER, PUBLIC               :: do_lri_inv = 0, &
                                               do_lri_pseudoinv_svd = 1, &
                                               do_lri_pseudoinv_diag = 2, &
                                               do_lri_inv_auto = 3

   ! LRI basis optimization parameters
   INTEGER, PARAMETER, PUBLIC               :: do_lri_opt_all = 0, &
                                               do_lri_opt_coeff = 1, &
                                               do_lri_opt_exps = 2

   ! callgraph parameters
   INTEGER, PARAMETER, PUBLIC               :: callgraph_none = 0, &
                                               callgraph_master = 1, &
                                               callgraph_all = 2

   ! kg coloring parameters
   INTEGER, PARAMETER, PUBLIC               :: kg_color_dsatur = 0, &
                                               kg_color_greedy = 1
   ! kg kinetic energy embedding potential
   INTEGER, PARAMETER, PUBLIC               :: kg_tnadd_embed = 100, &
                                               kg_tnadd_atomic = 200

   ! swarm parameters
   INTEGER, PARAMETER, PUBLIC               :: swarm_do_glbopt = 1

   ! global opt parameters
   INTEGER, PARAMETER, PUBLIC               :: glbopt_do_minhop = 1, &
                                               glbopt_do_mincrawl = 2

   ! Self-consistent continuum solvation (SCCS) models
   INTEGER, PARAMETER, PUBLIC               :: sccs_andreussi = 0, &
                                               sccs_fattebert_gygi = 1

   ! Derivative method used by the SCCS
   INTEGER, PARAMETER, PUBLIC               :: sccs_derivative_fft = 0, &
                                               sccs_derivative_cd3 = 1, &
                                               sccs_derivative_cd5 = 2, &
                                               sccs_derivative_cd7 = 3

   ! fm matrix multiplication
   INTEGER, PARAMETER, PUBLIC               :: do_pdgemm = 1, &
                                               do_dbcsr = 2

   ! Dispersion DFTB
   INTEGER, PARAMETER, PUBLIC               :: dispersion_uff = 100, &
                                               dispersion_d3 = 200

   ! Transport section
   INTEGER, PARAMETER, PUBLIC               :: transport_writeout = 0, &
                                               transport_diagscalapack = 1, &
                                               transport_localscf = 2, &
                                               transport_negf = 3, &
                                               transport_transmission = 4, &
                                               linsolver_splitsolve = 11, &
                                               linsolver_superlu = 12, &
                                               linsolver_mumps = 13, &
                                               linsolver_full = 14, &
                                               linsolver_banded = 15, &
                                               injmethod_evp = 21, &
                                               injmethod_beyn = 22, &
                                               injsign_positive = 1, &
                                               injsign_negative = -1, &
                                               rlaxisint_GaussChebyshev = 31, &
                                               rlaxisint_trapezoidal = 32, &
                                               rlaxisint_readfromfile = 33

   ! Active Space Section: orbital selection methods
   INTEGER, PARAMETER, PUBLIC               :: casci_canonical = 100, &
                                               wannier_projection = 101, &
                                               mao_projection = 102
   INTEGER, PARAMETER, PUBLIC               :: eri_method_full_gpw = 1, &
                                               eri_method_gpw_ht = 2
   INTEGER, PARAMETER, PUBLIC               :: eri_operator_coulomb = 1, &
                                               eri_operator_yukawa = 2, &
                                               eri_operator_erf = 3, &
                                               eri_operator_erfc = 4, &
                                               eri_operator_gaussian = 5

   INTEGER, PARAMETER, PUBLIC               :: do_eri_gpw = 0, &
                                               do_eri_mme = 1

END MODULE input_constants
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Defines control structures, which contain the parameters and the
!>      settings for the calculations.
! **************************************************************************************************
MODULE xas_control

   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE input_constants,                 ONLY: xas_1s_type,&
                                              xas_dscf,&
                                              xas_tp_fh,&
                                              xas_tp_flex,&
                                              xas_tp_hh,&
                                              xas_tp_xfh,&
                                              xas_tp_xhh,&
                                              xes_tp_val
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE memory_utilities,                ONLY: reallocate
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************
!> \brief A type that holds controling information for a xas calculation
! **************************************************************************************************
   TYPE xas_control_type
      INTEGER                             :: ref_count
      INTEGER                             :: nexc_atoms
      INTEGER                             :: nexc_search
      INTEGER                             :: state_type
      INTEGER                             :: xas_method
      INTEGER                             :: dipole_form
      INTEGER                             :: added_mos
      INTEGER                             :: max_iter_added
      INTEGER                             :: ngauss
      INTEGER                             :: stride
      INTEGER, DIMENSION(:), POINTER      :: exc_atoms
      LOGICAL                             :: cubes, do_centers
      LOGICAL                             :: xas_restart
      INTEGER, DIMENSION(:), POINTER      :: list_cubes
!
      REAL(dp)                            :: eps_added
      REAL(dp)                            :: xes_core_occupation
      REAL(dp)                            :: xes_homo_occupation
      REAL(dp)                            :: nel_tot, xas_core_occupation
   END TYPE xas_control_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xas_control'

! *** Public data types ***

   PUBLIC :: xas_control_type

! *** Public subroutines ***

   PUBLIC :: read_xas_control, write_xas_control, xas_control_create, &
             xas_control_release

CONTAINS

! **************************************************************************************************
!> \brief read from input the instructions for a xes/xas calculation
!> \param xas_control control variables
!>       error
!> \param xas_section ...
!> \par History
!>      04.2005 created [MI]
! **************************************************************************************************
   SUBROUTINE read_xas_control(xas_control, xas_section)

      TYPE(xas_control_type)                             :: xas_control
      TYPE(section_vals_type), POINTER                   :: xas_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_xas_control', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ir, n_rep, nex_at
      INTEGER, DIMENSION(:), POINTER                     :: list
      LOGICAL                                            :: hempty, was_present

      was_present = .FALSE.

      NULLIFY (list)

      CALL section_vals_val_get(xas_section, "METHOD", &
                                i_val=xas_control%xas_method)

      CALL section_vals_val_get(xas_section, "DIPOLE_FORM", &
                                i_val=xas_control%dipole_form)

      CALL section_vals_val_get(xas_section, "RESTART", &
                                l_val=xas_control%xas_restart)

      CALL section_vals_val_get(xas_section, "STATE_TYPE", &
                                i_val=xas_control%state_type)

      CALL section_vals_val_get(xas_section, "STATE_SEARCH", &
                                i_val=xas_control%nexc_search)

      CALL section_vals_val_get(xas_section, "XAS_CORE", &
                                r_val=xas_control%xas_core_occupation)

      CALL section_vals_val_get(xas_section, "XAS_TOT_EL", &
                                r_val=xas_control%nel_tot)

      CALL section_vals_val_get(xas_section, "XES_CORE", &
                                r_val=xas_control%xes_core_occupation)

      CALL section_vals_val_get(xas_section, "XES_EMPTY_HOMO", &
                                l_val=hempty)
      IF (hempty) THEN
         xas_control%xes_homo_occupation = 0
      ELSE
         xas_control%xes_homo_occupation = 1
      END IF

! It should be further generalized
      IF (.NOT. ASSOCIATED(xas_control%exc_atoms)) THEN
         CALL section_vals_val_get(xas_section, "ATOMS_LIST", &
                                   n_rep_val=n_rep)

         IF (n_rep > 0) THEN
            nex_at = 0
            DO ir = 1, n_rep
               NULLIFY (list)
               CALL section_vals_val_get(xas_section, "ATOMS_LIST", &
                                         i_rep_val=ir, i_vals=list)

               IF (ASSOCIATED(list)) THEN
                  CALL reallocate(xas_control%exc_atoms, 1, nex_at+SIZE(list))
                  DO i = 1, SIZE(list)
                     xas_control%exc_atoms(i+nex_at) = list(i)
                  END DO
                  xas_control%nexc_atoms = nex_at+SIZE(list)
                  nex_at = nex_at+SIZE(list)
               END IF
            END DO ! ir
         END IF
      END IF

      IF (.NOT. ASSOCIATED(xas_control%exc_atoms)) THEN
         xas_control%nexc_atoms = 1
         ALLOCATE (xas_control%exc_atoms(1))
         xas_control%exc_atoms(1) = 1
      END IF

      CALL section_vals_val_get(xas_section, "ADDED_MOS", &
                                i_val=xas_control%added_mos)

      CALL section_vals_val_get(xas_section, "MAX_ITER_ADDED", &
                                i_val=xas_control%max_iter_added)

      CALL section_vals_val_get(xas_section, "EPS_ADDED", &
                                r_val=xas_control%eps_added)

      CALL section_vals_val_get(xas_section, "NGAUSS", &
                                i_val=xas_control%ngauss)

   END SUBROUTINE read_xas_control

! **************************************************************************************************
!> \brief write on the instructions for a xes/xas calculation
!> \param xas_control control variables
!>       error
!> \param dft_section ...
!> \par History
!>      12.2005 created [MI]
! **************************************************************************************************
   SUBROUTINE write_xas_control(xas_control, dft_section)

      TYPE(xas_control_type)                             :: xas_control
      TYPE(section_vals_type), POINTER                   :: dft_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_xas_control', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: output_unit
      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, dft_section, &
                                         "PRINT%DFT_CONTROL_PARAMETERS", extension=".Log")
      IF (output_unit > 0) THEN
         SELECT CASE (xas_control%xas_method)
         CASE (xas_tp_hh)
            WRITE (UNIT=output_unit, FMT="(/,T2,A,T40,A)") &
               "XAS| Method:", &
               "      Transition potential with half hole"
         CASE (xas_tp_xhh)
            WRITE (UNIT=output_unit, FMT="(/,T2,A,T40,A)") &
               "XAS| Method:", &
               "      Transition potential with excited half hole"
         CASE (xas_tp_fh)
            WRITE (UNIT=output_unit, FMT="(/,T2,A,T40,A)") &
               "XAS| Method:", &
               "      Transition potential with full hole"
         CASE (xas_tp_xfh)
            WRITE (UNIT=output_unit, FMT="(/,T2,A,T40,A)") &
               "XAS| Method:", &
               "      Transition potential with excited full hole"
         CASE (xes_tp_val)
            WRITE (UNIT=output_unit, FMT="(/,T2,A,T40,A)") &
               "XAS| Method:", &
               " Only XES with full core and hole in lumo"
         CASE (xas_tp_flex)
            WRITE (UNIT=output_unit, FMT="(/,T2,A,T25,A)") &
               "XAS| Method:", &
               "      Transition potential with occupation of core state given from input"
         CASE (xas_dscf)
            WRITE (UNIT=output_unit, FMT="(/,T2,A,T40,A)") &
               "XAS| Method:", &
               "         DSCF for the first excited state"
         CASE default
            CALL cp__b("xas_control.F",226,"unknown xas method "//TRIM(ADJUSTL(cp_to_string(xas_control%xas_method))))
         END SELECT
         IF (xas_control%xas_restart) THEN
            WRITE (UNIT=output_unit, FMT="(/,T2,A,T30,A)") &
               "XAS|", " Orbitals read from atom-specific restart file when available"
         END IF
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, dft_section, &
                                        "PRINT%DFT_CONTROL_PARAMETERS")
   END SUBROUTINE write_xas_control

!****f* xas_control/xas_control_create/retain/release *

! **************************************************************************************************
!> \brief create retain release the xas_control_type
!> \param xas_control ...
!> \par History
!>      04.2005 created [MI]
! **************************************************************************************************
   SUBROUTINE xas_control_create(xas_control)

      TYPE(xas_control_type), POINTER                    :: xas_control

      CHARACTER(len=*), PARAMETER :: routineN = 'xas_control_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(xas_control)))CALL cp__a("xas_control.F",252)
      ALLOCATE (xas_control)

      xas_control%ref_count = 1
      xas_control%xas_method = xas_tp_hh
      xas_control%nexc_atoms = 1
      xas_control%nexc_search = -1
      xas_control%state_type = xas_1s_type
      xas_control%xas_restart = .FALSE.
      xas_control%added_mos = 0
      xas_control%xes_core_occupation = 1.0_dp
      xas_control%xes_homo_occupation = 1.0_dp
      NULLIFY (xas_control%exc_atoms)
      xas_control%cubes = .FALSE.
      xas_control%do_centers = .FALSE.
      NULLIFY (xas_control%list_cubes)

   END SUBROUTINE xas_control_create

! **************************************************************************************************
!> \brief ...
!> \param xas_control ...
! **************************************************************************************************
   SUBROUTINE xas_control_release(xas_control)

      TYPE(xas_control_type), POINTER                    :: xas_control

      CHARACTER(len=*), PARAMETER :: routineN = 'xas_control_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(xas_control)) THEN
         IF(.NOT.(xas_control%ref_count > 0))CALL cp__a("xas_control.F",283)
         xas_control%ref_count = xas_control%ref_count-1
         IF (xas_control%ref_count == 0) THEN
            IF (ASSOCIATED(xas_control%exc_atoms)) THEN
               DEALLOCATE (xas_control%exc_atoms)
            END IF
            IF (ASSOCIATED(xas_control%list_cubes)) THEN
               DEALLOCATE (xas_control%list_cubes)
            END IF
            DEALLOCATE (xas_control)
         END IF
      END IF

   END SUBROUTINE xas_control_release

! **************************************************************************************************
!> \brief ...
!> \param xas_control ...
! **************************************************************************************************
   SUBROUTINE xas_control_retain(xas_control)

      TYPE(xas_control_type), POINTER                    :: xas_control

      CHARACTER(len=*), PARAMETER :: routineN = 'xas_control_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(xas_control)))CALL cp__a("xas_control.F",309)
      xas_control%ref_count = xas_control%ref_count+1

   END SUBROUTINE xas_control_retain

END MODULE xas_control
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Defines control structures, which contain the parameters and the
!>      settings for the DFT-based calculations.
! **************************************************************************************************
MODULE cp_control_types
   USE cp_fm_types,                     ONLY: cp_fm_p_type,&
                                              cp_fm_release
   USE input_constants,                 ONLY: do_full_density
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE pw_types,                        ONLY: pw_p_type
   USE xas_control,                     ONLY: xas_control_release,&
                                              xas_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************
! \brief Control parameters for pw grids
! **************************************************************************************************
   TYPE pw_grid_option
      LOGICAL :: spherical
      LOGICAL :: fullspace
      INTEGER, DIMENSION(2) :: distribution_layout
      INTEGER :: blocked
   END TYPE pw_grid_option

! **************************************************************************************************
! \brief Control parameters for REAL_TIME_PROPAGATION calculations
! **************************************************************************************************
   TYPE rtp_control_type
      LOGICAL                              :: converged
      REAL(KIND=dp)                      :: eps_ener
      INTEGER                              :: max_iter
      INTEGER                              :: mat_exp
      INTEGER                              :: propagator
      LOGICAL                              :: fixed_ions
      INTEGER                              :: initial_wfn
      REAL(dp)                             :: eps_exp
      LOGICAL                              :: initial_step
      LOGICAL                              :: hfx_redistribute
      INTEGER                              :: aspc_order
      INTEGER                              :: sc_check_start
      LOGICAL                              :: apply_delta_pulse
      LOGICAL                              :: periodic
      LOGICAL                              :: linear_scaling
      LOGICAL                              :: orthonormal
      LOGICAL                              :: write_restart
      INTEGER                              :: mcweeny_max_iter
      INTEGER                              :: acc_ref
      REAL(dp)                             :: mcweeny_eps
      INTEGER, DIMENSION(3)                :: delta_pulse_direction
      REAL(KIND=dp)                        :: delta_pulse_scale
   END TYPE rtp_control_type
! **************************************************************************************************
! \brief Control parameters for DFTB calculations
! **************************************************************************************************
   TYPE dftb_control_type
      LOGICAL                              :: self_consistent
      LOGICAL                              :: orthogonal_basis
      LOGICAL                              :: dispersion
      INTEGER                              :: dispersion_type
      LOGICAL                              :: dftb3_diagonal
      LOGICAL                              :: hb_sr_damp
      REAL(KIND=dp)                      :: hb_sr_para
      REAL(KIND=dp)                      :: eps_disp
      REAL(KIND=dp)                      :: epscn
      REAL(KIND=dp)                      :: rcdisp
      REAL(KIND=dp), DIMENSION(3)        :: sd3
      LOGICAL                              :: do_ewald
      CHARACTER(LEN=default_path_length)   :: sk_file_path
      CHARACTER(LEN=default_path_length)   :: sk_file_list
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:, :), POINTER         :: sk_pair_list
      CHARACTER(LEN=default_path_length)   :: uff_force_field
      CHARACTER(LEN=default_path_length)   :: dispersion_parameter_file
   END TYPE dftb_control_type

! **************************************************************************************************
! \brief Control parameters for SCPTB calculations
! **************************************************************************************************
   TYPE scptb_control_type
      LOGICAL                              :: dispersion
      LOGICAL                              :: do_ewald
      LOGICAL                              :: do_scc
      LOGICAL                              :: do_scp
      INTEGER                              :: sto_ng
      REAL(KIND=dp)                      :: rcdisp
      REAL(KIND=dp)                      :: epscn
      REAL(KIND=dp), DIMENSION(3)        :: sd3
      CHARACTER(LEN=default_path_length)   :: parameter_file
      CHARACTER(LEN=default_path_length)   :: dispersion_parameter_file
      REAL(KIND=dp)                      :: epspair
   END TYPE scptb_control_type

! **************************************************************************************************
! \brief Control parameters for semi empirical calculations
! **************************************************************************************************
   TYPE semi_empirical_control_type
      LOGICAL               :: orthogonal_basis
      LOGICAL               :: analytical_gradients
      LOGICAL               :: force_kdsod_EX
      LOGICAL               :: do_ewald, do_ewald_r3, do_ewald_gks
      INTEGER               :: integral_screening, periodic_type
      INTEGER               :: max_multipole
      INTEGER               :: ga_ncells
      REAL(KIND=dp)       :: delta
      ! Dispersion pair potential
      LOGICAL               :: dispersion
      REAL(KIND=dp)                      :: rcdisp
      REAL(KIND=dp)                      :: epscn
      REAL(KIND=dp), DIMENSION(3)        :: sd3
      CHARACTER(LEN=default_path_length)   :: dispersion_parameter_file
      ! Parameters controlling the evaluation of the integrals
      REAL(KIND=dp)       :: cutoff_lrc, taper_lrc, range_lrc
      REAL(KIND=dp)       :: cutoff_cou, taper_cou, range_cou
      REAL(KIND=dp)       :: cutoff_exc, taper_exc, range_exc
      REAL(KIND=dp)       :: taper_scr, range_scr
   END TYPE semi_empirical_control_type

! **************************************************************************************************
! \brief Control parameters for GAPW method within QUICKSTEP ***
! **************************************************************************************************
   TYPE gapw_control_type
      REAL(KIND=dp)                 :: eps_fit, &
                                       eps_iso, &
                                       eps_Vrho0, &
                                       eps_svd, &
                                       eps_cpc
      INTEGER                         :: ladd_rho0, &
                                         lmax_rho0, &
                                         lmax_sphere, &
                                         quadrature
      LOGICAL                         :: lrho1_eq_lrho0
      LOGICAL                         :: alpha0_hard_from_input, &
                                         force_paw, &
                                         non_paw_atoms, &
                                         nopaw_as_gpw
      REAL(KIND=dp)                 :: alpha0_hard
      REAL(KIND=dp)                 :: max_rad_local
   END TYPE gapw_control_type
! **************************************************************************************************
! \brief parameters for calculations involving a time dependent electric field
! **************************************************************************************************
   TYPE efield_type
      REAL(KIND=dp)                 :: actual_time
      REAL(KIND=dp), DIMENSION(:), POINTER :: polarisation
      INTEGER                         :: envelop_id
      REAL(kind=dp), DIMENSION(:), &
         POINTER :: envelop_r_vars
      INTEGER, DIMENSION(:), &
         POINTER :: envelop_i_vars
      REAL(kind=dp)                   :: strength
      REAL(KIND=dp)                   :: phase_offset
      REAL(KIND=dp)                   :: wavelength
   END TYPE efield_type

   TYPE efield_p_type
      TYPE(efield_type), POINTER       :: efield
   END TYPE efield_p_type
! **************************************************************************************************
! \brief parameters for calculations involving a time dependent electric field
! **************************************************************************************************
   TYPE period_efield_type
      LOGICAL                         :: displacement_field
      REAL(KIND=dp), DIMENSION(:), POINTER :: polarisation
      REAL(kind=dp)                   :: strength
   END TYPE period_efield_type

! **************************************************************************************************
! \brief some parameters useful for mulliken_restraints
! **************************************************************************************************
   TYPE mulliken_restraint_type
      INTEGER                         :: ref_count
      REAL(KIND=dp)                 :: strength
      REAL(KIND=dp)                 :: TARGET
      INTEGER                         :: natoms
      INTEGER, POINTER, DIMENSION(:)  :: atoms
   END TYPE mulliken_restraint_type

! **************************************************************************************************
! \brief some parameters useful for ddapc_restraints
! **************************************************************************************************
   TYPE ddapc_restraint_type
      INTEGER                         :: ref_count
      REAL(KIND=dp)                 :: strength
      REAL(KIND=dp)                 :: TARGET
      REAL(KIND=dp)                 :: ddapc_order_p
      INTEGER                         :: functional_form
      INTEGER                         :: natoms
      INTEGER, POINTER, DIMENSION(:)  :: atoms
      REAL(KIND=dp), POINTER, DIMENSION(:)  :: coeff
      INTEGER                         :: density_type
   END TYPE ddapc_restraint_type

! **************************************************************************************************
! \brief provides a vector of pointers to ddapc_restraint_type
! **************************************************************************************************
   TYPE ddapc_restraint_p_type
      TYPE(ddapc_restraint_type), POINTER:: ddapc_restraint_control
   END TYPE ddapc_restraint_p_type

! **************************************************************************************************
! \brief some parameters useful for becke_restraints
! **************************************************************************************************
   TYPE becke_restraint_type
      INTEGER                         :: ref_count
      REAL(KIND=dp)                 :: strength
      REAL(KIND=dp)                 :: TARGET
      REAL(KIND=dp)                 :: becke_order_p
      INTEGER                         :: functional_form
      INTEGER                         :: natoms
      INTEGER, POINTER, DIMENSION(:)  :: atoms
      TYPE(pw_p_type)                 :: becke_pot
      LOGICAL                         :: need_pot
      REAL(KIND=dp), POINTER, DIMENSION(:)  :: coeff
      INTEGER                         :: density_type
   END TYPE becke_restraint_type

! **************************************************************************************************
! \brief some parameters useful for s2_restraints
! **************************************************************************************************
   TYPE s2_restraint_type
      INTEGER                         :: ref_count
      REAL(KIND=dp)                 :: strength
      REAL(KIND=dp)                 :: TARGET
      REAL(KIND=dp)                 :: s2_order_p
      INTEGER                         :: functional_form
   END TYPE s2_restraint_type

! **************************************************************************************************
! \brief some parameters useful for auxiliary density matrix method
! **************************************************************************************************
   TYPE admm_block_type
      INTEGER, DIMENSION(:), ALLOCATABLE :: list
   END TYPE admm_block_type

   TYPE admm_control_type
      REAL(KIND=dp)                 :: eps_filter
      INTEGER                         :: purification_method
      INTEGER                         :: method
      LOGICAL                         :: charge_constrain
      INTEGER                         :: scaling_model
      INTEGER                         :: aux_exch_func
      TYPE(admm_block_type), DIMENSION(:), &
         ALLOCATABLE :: blocks
   END TYPE admm_control_type

! **************************************************************************************************
! \brief Control parameters for a QUICKSTEP and KIM-GORDON calculation ***
!        eps_pgf_orb: Cutoff value for the interaction of the primitive
!                     Gaussian-type functions (primitive basis functions).
! **************************************************************************************************
   TYPE qs_control_type
      CHARACTER(LEN=10)                             :: method
      INTEGER                                       :: method_id
      REAL(KIND=dp)                               :: eps_core_charge, &
                                                     eps_kg_orb, &
                                                     eps_pgf_orb, &
                                                     eps_ppl, &
                                                     eps_ppnl, &
                                                     eps_rho_gspace, &
                                                     eps_rho_rspace, &
                                                     eps_filter_matrix, &
                                                     eps_gvg_rspace, &
                                                     progression_factor, &
                                                     relative_cutoff
      LOGICAL                                       :: do_almo_scf
      LOGICAL                                       :: do_ls_scf
      LOGICAL                                       :: do_kg
      LOGICAL                                       :: commensurate_mgrids
      LOGICAL                                       :: realspace_mgrids
      LOGICAL                                       :: map_consistent
      LOGICAL                                       :: gapw, gapw_xc, gpw, pao
      LOGICAL                                       :: lrigpw
      LOGICAL                                       :: lri_optbas
      LOGICAL                                       :: ofgpw
      LOGICAL                                       :: dftb
      LOGICAL                                       :: scptb
      LOGICAL                                       :: semi_empirical
      LOGICAL                                       :: mulliken_restraint
      LOGICAL                                       :: ddapc_restraint
      LOGICAL                                       :: ddapc_restraint_is_spin
      LOGICAL                                       :: ddapc_explicit_potential
      LOGICAL                                       :: becke_restraint
      LOGICAL                                       :: et_coupling_calc
      LOGICAL                                       :: s2_restraint
      INTEGER                                       :: do_ppl_method
      INTEGER                                       :: wf_interpolation_method_nr
      INTEGER                                       :: wf_extrapolation_order
      REAL(KIND=dp)                               :: cutoff
      REAL(KIND=dp), DIMENSION(:), POINTER        :: e_cutoff
      TYPE(mulliken_restraint_type), POINTER       :: mulliken_restraint_control
      TYPE(ddapc_restraint_p_type), DIMENSION(:), &
         POINTER                                     :: ddapc_restraint_control
      TYPE(becke_restraint_type), POINTER          :: becke_control
      TYPE(s2_restraint_type), POINTER             :: s2_restraint_control
      TYPE(dftb_control_type), POINTER              :: dftb_control
      TYPE(scptb_control_type), POINTER             :: scptb_control
      TYPE(semi_empirical_control_type), POINTER    :: se_control
      TYPE(gapw_control_type), POINTER             :: gapw_control
      TYPE(pw_grid_option)                         :: pw_grid_opt
      LOGICAL                                       :: skip_load_balance_distributed
   END TYPE qs_control_type

! **************************************************************************************************
! \brief Control parameters for the SCCS models
! **************************************************************************************************
   TYPE sccs_control_type
      LOGICAL                                      :: sccs_activated
      INTEGER                                      :: derivative_method, &
                                                      max_iter, &
                                                      method_id, &
                                                      ref_count
      REAL(KIND=dp)                                :: alpha_solvent, &
                                                      beta, &
                                                      beta_solvent, &
                                                      delta_rho, &
                                                      eps_sccs, &
                                                      eps_scf, &
                                                      epsilon_solvent, &
                                                      gamma_solvent, &
                                                      mixing, &
                                                      rho_zero, &
                                                      rho_max, &
                                                      rho_min
   END TYPE sccs_control_type

! **************************************************************************************************
! \brief Control parameters for a TIME-DEPENDENT PERTURBATION calculation
! \par  ATTRIBUTES
!   - n_ev       : number of eigenvalues to calculate
!   - n_reortho  : how many time to reorthogonalize (in the lanczos algorithm)
!   - do_kernel  : wether to evaluate the kernel (this is a debugging option)
!   - res_etype : { SINGLET | TRIPLET } which excitations
!                  to calculate
!   - lumos_eigenvalues : holds the eigenvalues of the lumos (if calculated in QS)
!
! \par  NOTES
!   The lumos are helpfull in choosing a initial vector for the TDDFPT
!   calculation, since they can be used to construct the solutions of the
!   TDDFPT operator without the perturbation kernel.
! **************************************************************************************************
   TYPE tddfpt_control_type
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER :: lumos
      REAL(KIND=dp)                             :: tolerance
      INTEGER                                   :: n_ev
      INTEGER                                   :: max_kv
      INTEGER                                   :: n_restarts
      INTEGER                                   :: n_reortho
      LOGICAL                                   :: do_kernel
      LOGICAL                                   :: lsd_singlets
      LOGICAL                                   :: invert_S
      LOGICAL                                   :: precond
      LOGICAL                                   :: drho_by_collocation
      LOGICAL                                   :: use_kinetic_energy_density
      INTEGER                                   :: res_etype
      INTEGER                                   :: diag_method
      INTEGER                                   :: oe_corr
      INTEGER                                   :: sic_method_id
      INTEGER                                   :: sic_list_id
      REAL(KIND=dp)                             :: sic_scaling_a, sic_scaling_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: lumos_eigenvalues
   END TYPE tddfpt_control_type

! **************************************************************************************************
! \brief Control parameters for a DFT calculation
! **************************************************************************************************
   TYPE dft_control_type
      TYPE(admm_control_type), POINTER           :: admm_control
      TYPE(period_efield_type), POINTER          :: period_efield
      TYPE(qs_control_type), POINTER             :: qs_control
      TYPE(rtp_control_type), POINTER            :: rtp_control
      TYPE(sccs_control_type), POINTER           :: sccs_control
      TYPE(tddfpt_control_type), POINTER         :: tddfpt_control
      TYPE(xas_control_type), POINTER            :: xas_control
      TYPE(efield_p_type), POINTER, DIMENSION(:) :: efield_fields
      INTEGER                                    :: nspins, &
                                                    charge, &
                                                    multiplicity, &
                                                    sic_method_id, &
                                                    ref_count, &
                                                    id_nr, &
                                                    plus_u_method_id, &
                                                    dir_surf_dip, &
                                                    nimages = 1
      INTEGER                                    :: sic_list_id
      REAL(KIND=dp)                              :: relax_multiplicity, &
                                                    sic_scaling_a, &
                                                    sic_scaling_b
      LOGICAL                                    :: do_tddfpt_calculation, &
                                                    do_xas_calculation, &
                                                    drho_by_collocation, &
                                                    use_kinetic_energy_density, &
                                                    restricted, &
                                                    roks, &
                                                    uks, &
                                                    lsd, &
                                                    dft_plus_u, &
                                                    apply_efield, &
                                                    apply_efield_field, &
                                                    apply_period_efield, &
                                                    apply_external_potential, &
                                                    eval_external_potential, &
                                                    do_admm, &
                                                    do_admm_dm, &
                                                    do_admm_mo, &
                                                    smear, &
                                                    low_spin_roks, &
                                                    apply_external_density, &
                                                    read_external_density, &
                                                    apply_external_vxc, &
                                                    read_external_vxc, &
                                                    correct_surf_dip, &
                                                    do_sccs
   END TYPE dft_control_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_control_types'
   INTEGER, SAVE :: last_dft_control_id = 0

   ! Public data types

   PUBLIC :: dft_control_type, &
             qs_control_type, &
             gapw_control_type, &
             tddfpt_control_type, &
             efield_type, &
             mulliken_restraint_type, &
             ddapc_restraint_type, &
             dftb_control_type, &
             scptb_control_type, &
             becke_restraint_type, &
             semi_empirical_control_type, &
             s2_restraint_type, &
             admm_control_type, &
             rtp_control_type, &
             sccs_control_type

   ! Public subroutines

   PUBLIC :: dft_control_retain, &
             dft_control_release, &
             dft_control_create, &
             tddfpt_control_create, &
             admm_control_create, &
             ddapc_control_create, &
             sccs_control_create

CONTAINS

! **************************************************************************************************
!> \brief create  the mulliken_restraint_type
!> \param mulliken_restraint_control ...
!> \par History
!>      02.2005 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE mulliken_control_create(mulliken_restraint_control)
      TYPE(mulliken_restraint_type), POINTER             :: mulliken_restraint_control

      CHARACTER(len=*), PARAMETER :: routineN = 'mulliken_control_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(mulliken_restraint_control)))CALL cp__a("cp_control_types.F",471)
      ALLOCATE (mulliken_restraint_control)

      mulliken_restraint_control%ref_count = 1
      mulliken_restraint_control%strength = 0.1_dp
      mulliken_restraint_control%target = 1.0_dp
      mulliken_restraint_control%natoms = 0
      NULLIFY (mulliken_restraint_control%atoms)
   END SUBROUTINE mulliken_control_create

! **************************************************************************************************
!> \brief release the mulliken_restraint_type
!> \param mulliken_restraint_control ...
!> \par History
!>      02.2005 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE mulliken_control_release(mulliken_restraint_control)
      TYPE(mulliken_restraint_type), POINTER             :: mulliken_restraint_control

      CHARACTER(len=*), PARAMETER :: routineN = 'mulliken_control_release', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(mulliken_restraint_control)))CALL cp__a("cp_control_types.F",493)
      IF(.NOT.(mulliken_restraint_control%ref_count > 0))CALL cp__a("cp_control_types.F",494)
      mulliken_restraint_control%ref_count = mulliken_restraint_control%ref_count-1
      IF (mulliken_restraint_control%ref_count == 0) THEN
         IF (ASSOCIATED(mulliken_restraint_control%atoms)) &
            DEALLOCATE (mulliken_restraint_control%atoms)
         mulliken_restraint_control%ref_count = 0
         mulliken_restraint_control%strength = 0.0_dp
         mulliken_restraint_control%target = 0.0_dp
         mulliken_restraint_control%natoms = 0
         DEALLOCATE (mulliken_restraint_control)
      ENDIF
   END SUBROUTINE mulliken_control_release

! **************************************************************************************************
!> \brief retain the mulliken_restraint_type
!> \param mulliken_restraint_control ...
!> \par History
!>      02.2005 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE mulliken_control_retain(mulliken_restraint_control)
      TYPE(mulliken_restraint_type), POINTER             :: mulliken_restraint_control

      CHARACTER(len=*), PARAMETER :: routineN = 'mulliken_control_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(mulliken_restraint_control)))CALL cp__a("cp_control_types.F",519)

      mulliken_restraint_control%ref_count = mulliken_restraint_control%ref_count+1
   END SUBROUTINE mulliken_control_retain

! **************************************************************************************************
!> \brief create the ddapc_restraint_type
!> \param ddapc_restraint_control ...
!> \par History
!>      02.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE ddapc_control_create(ddapc_restraint_control)
      TYPE(ddapc_restraint_type), POINTER                :: ddapc_restraint_control

      CHARACTER(len=*), PARAMETER :: routineN = 'ddapc_control_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(ddapc_restraint_control)))CALL cp__a("cp_control_types.F",536)
      ALLOCATE (ddapc_restraint_control)

      ddapc_restraint_control%density_type = do_full_density
      ddapc_restraint_control%ref_count = 1
      ddapc_restraint_control%strength = 0.1_dp
      ddapc_restraint_control%ddapc_order_p = 0.0_dp
      ddapc_restraint_control%functional_form = -1
      ddapc_restraint_control%target = 1.0_dp
      ddapc_restraint_control%natoms = 0
      NULLIFY (ddapc_restraint_control%atoms)
      NULLIFY (ddapc_restraint_control%coeff)

   END SUBROUTINE ddapc_control_create

! **************************************************************************************************
!> \brief release the ddapc_restraint_type
!> \param ddapc_restraint_control ...
!> \par History
!>      02.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE ddapc_control_release(ddapc_restraint_control)
      TYPE(ddapc_restraint_type), POINTER                :: ddapc_restraint_control

      CHARACTER(len=*), PARAMETER :: routineN = 'ddapc_control_release', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ddapc_restraint_control)))CALL cp__a("cp_control_types.F",563)
      IF(.NOT.(ddapc_restraint_control%ref_count > 0))CALL cp__a("cp_control_types.F",564)
      ddapc_restraint_control%ref_count = ddapc_restraint_control%ref_count-1
      IF (ddapc_restraint_control%ref_count == 0) THEN
         IF (ASSOCIATED(ddapc_restraint_control%atoms)) &
            DEALLOCATE (ddapc_restraint_control%atoms)
         IF (ASSOCIATED(ddapc_restraint_control%coeff)) &
            DEALLOCATE (ddapc_restraint_control%coeff)
         ddapc_restraint_control%ref_count = 0
         ddapc_restraint_control%strength = 0.0_dp
         ddapc_restraint_control%target = 0.0_dp
         ddapc_restraint_control%natoms = 0
         DEALLOCATE (ddapc_restraint_control)
      ENDIF
   END SUBROUTINE ddapc_control_release

! **************************************************************************************************
!> \brief retain the ddapc_restraint_type
!> \param ddapc_restraint_control ...
!> \par History
!>      02.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE ddapc_control_retain(ddapc_restraint_control)
      TYPE(ddapc_restraint_type), POINTER                :: ddapc_restraint_control

      CHARACTER(len=*), PARAMETER :: routineN = 'ddapc_control_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ddapc_restraint_control)))CALL cp__a("cp_control_types.F",591)

      ddapc_restraint_control%ref_count = ddapc_restraint_control%ref_count+1
   END SUBROUTINE ddapc_control_retain

! **************************************************************************************************
!> \brief create the becke_restraint_type
!> \param becke_control ...
!> \par History
!>      02.2007 created [Florian Schiffmann]
! **************************************************************************************************
   SUBROUTINE becke_control_create(becke_control)
      TYPE(becke_restraint_type), POINTER                :: becke_control

      CHARACTER(len=*), PARAMETER :: routineN = 'becke_control_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(becke_control)))CALL cp__a("cp_control_types.F",608)
      ALLOCATE (becke_control)

      becke_control%ref_count = 1
      becke_control%density_type = do_full_density
      becke_control%strength = 0.1_dp
      becke_control%becke_order_p = 0.0_dp
      becke_control%functional_form = -1
      becke_control%target = 1.0_dp
      becke_control%natoms = 0
      becke_control%need_pot = .TRUE.
      NULLIFY (becke_control%atoms)
      NULLIFY (becke_control%coeff)
   END SUBROUTINE becke_control_create

! **************************************************************************************************
!> \brief release the becke_restraint_type
!> \param becke_control ...
!> \par History
!>      02.2007 created [Florian Schiffmann]
! **************************************************************************************************
   SUBROUTINE becke_control_release(becke_control)
      TYPE(becke_restraint_type), POINTER                :: becke_control

      CHARACTER(len=*), PARAMETER :: routineN = 'becke_control_release', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(becke_control)))CALL cp__a("cp_control_types.F",635)
      IF(.NOT.(becke_control%ref_count > 0))CALL cp__a("cp_control_types.F",636)
      becke_control%ref_count = becke_control%ref_count-1
      IF (becke_control%ref_count == 0) THEN
         IF (ASSOCIATED(becke_control%atoms)) &
            DEALLOCATE (becke_control%atoms)
         IF (ASSOCIATED(becke_control%coeff)) &
            DEALLOCATE (becke_control%coeff)
         becke_control%ref_count = 0
         becke_control%strength = 0.0_dp
         becke_control%target = 0.0_dp
         becke_control%natoms = 0
         DEALLOCATE (becke_control)
      ENDIF
   END SUBROUTINE becke_control_release

! **************************************************************************************************
!> \brief retain the becke_restraint_type
!> \param becke_control ...
!> \par History
!>      02.2007 created [Florian Schiffmann]
! **************************************************************************************************
   SUBROUTINE becke_control_retain(becke_control)
      TYPE(becke_restraint_type), POINTER                :: becke_control

      CHARACTER(len=*), PARAMETER :: routineN = 'becke_control_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(becke_control)))CALL cp__a("cp_control_types.F",663)
      becke_control%ref_count = becke_control%ref_count+1
   END SUBROUTINE becke_control_retain

! **************************************************************************************************
!> \brief create the s2_restraint_type
!> \param s2_restraint_control ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE s2_control_create(s2_restraint_control)
      TYPE(s2_restraint_type), POINTER                   :: s2_restraint_control

      CHARACTER(len=*), PARAMETER :: routineN = 's2_control_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(s2_restraint_control)))CALL cp__a("cp_control_types.F",679)
      ALLOCATE (s2_restraint_control)

      s2_restraint_control%ref_count = 1
      s2_restraint_control%strength = 0.1_dp
      s2_restraint_control%s2_order_p = 0.0_dp
      s2_restraint_control%functional_form = -1
      s2_restraint_control%target = 1.0_dp
   END SUBROUTINE s2_control_create

! **************************************************************************************************
!> \brief release the s2_restraint_type
!> \param s2_restraint_control ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE s2_control_release(s2_restraint_control)
      TYPE(s2_restraint_type), POINTER                   :: s2_restraint_control

      CHARACTER(len=*), PARAMETER :: routineN = 's2_control_release', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(s2_restraint_control)))CALL cp__a("cp_control_types.F",701)
      IF(.NOT.(s2_restraint_control%ref_count > 0))CALL cp__a("cp_control_types.F",702)
      s2_restraint_control%ref_count = s2_restraint_control%ref_count-1
      IF (s2_restraint_control%ref_count == 0) THEN
         s2_restraint_control%ref_count = 0
         s2_restraint_control%strength = 0.0_dp
         s2_restraint_control%target = 0.0_dp
         DEALLOCATE (s2_restraint_control)
      ENDIF
   END SUBROUTINE s2_control_release

! **************************************************************************************************
!> \brief retain the s2_restraint_type
!> \param s2_restraint_control ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE s2_control_retain(s2_restraint_control)
      TYPE(s2_restraint_type), POINTER                   :: s2_restraint_control

      CHARACTER(len=*), PARAMETER :: routineN = 's2_control_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(s2_restraint_control)))CALL cp__a("cp_control_types.F",724)
      s2_restraint_control%ref_count = s2_restraint_control%ref_count+1
   END SUBROUTINE s2_control_retain

! **************************************************************************************************
!> \brief allocates and perform a very basic initialization
!> \param dft_control the object to create
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE dft_control_create(dft_control)
      TYPE(dft_control_type), POINTER                    :: dft_control

      CHARACTER(len=*), PARAMETER :: routineN = 'dft_control_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(dft_control)))CALL cp__a("cp_control_types.F",741)
      ALLOCATE (dft_control)
      dft_control%ref_count = 1
      last_dft_control_id = last_dft_control_id+1
      dft_control%id_nr = last_dft_control_id
      NULLIFY (dft_control%xas_control)
      NULLIFY (dft_control%qs_control)
      NULLIFY (dft_control%tddfpt_control)
      NULLIFY (dft_control%efield_fields)
      NULLIFY (dft_control%period_efield)
      NULLIFY (dft_control%admm_control)
      NULLIFY (dft_control%rtp_control)
      NULLIFY (dft_control%sccs_control)
      dft_control%do_sccs = .FALSE.
      CALL qs_control_create(dft_control%qs_control)
   END SUBROUTINE dft_control_create

! **************************************************************************************************
!> \brief ...
!> \param dft_control ...
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE dft_control_retain(dft_control)
      TYPE(dft_control_type), POINTER                    :: dft_control

      CHARACTER(len=*), PARAMETER :: routineN = 'dft_control_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(dft_control)))CALL cp__a("cp_control_types.F",771)
      IF(.NOT.(dft_control%ref_count > 0))CALL cp__a("cp_control_types.F",772)
      dft_control%ref_count = dft_control%ref_count+1
   END SUBROUTINE dft_control_retain

! **************************************************************************************************
!> \brief ...
!> \param dft_control ...
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE dft_control_release(dft_control)
      TYPE(dft_control_type), POINTER                    :: dft_control

      CHARACTER(len=*), PARAMETER :: routineN = 'dft_control_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(dft_control)) THEN
         IF(.NOT.(dft_control%ref_count > 0))CALL cp__a("cp_control_types.F",790)
         dft_control%ref_count = dft_control%ref_count-1
         IF (dft_control%ref_count == 0) THEN
            CALL qs_control_release(dft_control%qs_control)
            CALL tddfpt_control_release(dft_control%tddfpt_control)
            CALL xas_control_release(dft_control%xas_control)
            CALL admm_control_release(dft_control%admm_control)
            CALL efield_fields_release(dft_control%efield_fields)
            CALL sccs_control_release(dft_control%sccs_control)
            IF (ASSOCIATED(dft_control%period_efield)) THEN
               DEALLOCATE (dft_control%period_efield%polarisation)
               DEALLOCATE (dft_control%period_efield)
            END IF
            IF (ASSOCIATED(dft_control%rtp_control)) THEN
               DEALLOCATE (dft_control%rtp_control)
            END IF
            DEALLOCATE (dft_control)
         END IF
      END IF

   END SUBROUTINE dft_control_release

! **************************************************************************************************
!> \brief ...
!> \param gapw_control ...
! **************************************************************************************************
   SUBROUTINE gapw_control_create(gapw_control)
      TYPE(gapw_control_type), POINTER                   :: gapw_control

      CHARACTER(len=*), PARAMETER :: routineN = 'gapw_control_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(gapw_control)))CALL cp__a("cp_control_types.F",822)
      ALLOCATE (gapw_control)
   END SUBROUTINE gapw_control_create

! **************************************************************************************************
!> \brief ...
!> \param qs_control ...
! **************************************************************************************************
   SUBROUTINE qs_control_create(qs_control)
      TYPE(qs_control_type), POINTER                     :: qs_control

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_control_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(qs_control)))CALL cp__a("cp_control_types.F",836)
      ALLOCATE (qs_control)

      NULLIFY (qs_control%e_cutoff)
      NULLIFY (qs_control%gapw_control)
      NULLIFY (qs_control%mulliken_restraint_control)
      NULLIFY (qs_control%ddapc_restraint_control)
      NULLIFY (qs_control%s2_restraint_control)
      NULLIFY (qs_control%se_control)
      NULLIFY (qs_control%dftb_control)
      NULLIFY (qs_control%scptb_control)
      NULLIFY (qs_control%becke_control)
      NULLIFY (qs_control%ddapc_restraint_control)

      CALL mulliken_control_create(qs_control%mulliken_restraint_control)
      CALL becke_control_create(qs_control%becke_control)
      CALL s2_control_create(qs_control%s2_restraint_control)
      CALL gapw_control_create(qs_control%gapw_control)
      CALL se_control_create(qs_control%se_control)
      CALL dftb_control_create(qs_control%dftb_control)
      CALL scptb_control_create(qs_control%scptb_control)
   END SUBROUTINE qs_control_create

! **************************************************************************************************
!> \brief ...
!> \param qs_control ...
! **************************************************************************************************
   SUBROUTINE qs_control_release(qs_control)
      TYPE(qs_control_type), POINTER                     :: qs_control

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_control_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(qs_control)) THEN
         CALL mulliken_control_release(qs_control%mulliken_restraint_control)
         CALL s2_control_release(qs_control%s2_restraint_control)
         CALL se_control_release(qs_control%se_control)
         CALL dftb_control_release(qs_control%dftb_control)
         CALL scptb_control_release(qs_control%scptb_control)
         CALL becke_control_release(qs_control%becke_control)

         IF (ASSOCIATED(qs_control%e_cutoff)) THEN
            DEALLOCATE (qs_control%e_cutoff)
         END IF
         IF (ASSOCIATED(qs_control%gapw_control)) THEN
            DEALLOCATE (qs_control%gapw_control)
         END IF
         IF (ASSOCIATED(qs_control%ddapc_restraint_control)) THEN
            DO i = 1, SIZE(qs_control%ddapc_restraint_control)
               CALL ddapc_control_release(qs_control%ddapc_restraint_control(i)%ddapc_restraint_control)
            END DO
            DEALLOCATE (qs_control%ddapc_restraint_control)
         END IF
         DEALLOCATE (qs_control)
      END IF
   END SUBROUTINE qs_control_release

! **************************************************************************************************
!> \brief ...
!> \param tddfpt_control ...
! **************************************************************************************************
   SUBROUTINE tddfpt_control_create(tddfpt_control)
      TYPE(tddfpt_control_type), POINTER                 :: tddfpt_control

      CHARACTER(len=*), PARAMETER :: routineN = 'tddfpt_control_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(tddfpt_control)))CALL cp__a("cp_control_types.F",905)
      ALLOCATE (tddfpt_control)
      NULLIFY (tddfpt_control%lumos)
      NULLIFY (tddfpt_control%lumos_eigenvalues)

   END SUBROUTINE tddfpt_control_create

! **************************************************************************************************
!> \brief ...
!> \param tddfpt_control ...
! **************************************************************************************************
   SUBROUTINE tddfpt_control_release(tddfpt_control)
      TYPE(tddfpt_control_type), POINTER                 :: tddfpt_control

      CHARACTER(len=*), PARAMETER :: routineN = 'tddfpt_control_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin
      LOGICAL                                            :: dummy

      IF (ASSOCIATED(tddfpt_control)) THEN
         IF (ASSOCIATED(tddfpt_control%lumos)) THEN
            DO ispin = 1, SIZE(tddfpt_control%lumos)
               CALL cp_fm_release(tddfpt_control%lumos(ispin)%matrix)
               !MK the following line just avoids a crash of TDDFT runs using
               !MK the sdbg version compiled with the NAG compiler when
               !MK tddfpt_control%lumos is deallocated. This is most likely a
               !MK compiler bug and thus the line might become obsolete
               dummy = ASSOCIATED(tddfpt_control%lumos(ispin)%matrix)
            END DO
            DEALLOCATE (tddfpt_control%lumos)
         END IF
         IF (ASSOCIATED(tddfpt_control%lumos_eigenvalues)) THEN
            DEALLOCATE (tddfpt_control%lumos_eigenvalues)
         END IF
         DEALLOCATE (tddfpt_control)
      END IF
   END SUBROUTINE tddfpt_control_release

! **************************************************************************************************
!> \brief ...
!> \param efield_fields ...
! **************************************************************************************************
   SUBROUTINE efield_fields_release(efield_fields)
      TYPE(efield_p_type), DIMENSION(:), POINTER         :: efield_fields

      CHARACTER(len=*), PARAMETER :: routineN = 'efield_fields_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(efield_fields)) THEN
         DO i = 1, SIZE(efield_fields)
            IF (ASSOCIATED(efield_fields(i)%efield)) THEN
               IF (ASSOCIATED(efield_fields(i)%efield%envelop_r_vars)) THEN
                  DEALLOCATE (efield_fields(i)%efield%envelop_r_vars)
               END IF
               IF (ASSOCIATED(efield_fields(i)%efield%envelop_i_vars)) THEN
                  DEALLOCATE (efield_fields(i)%efield%envelop_i_vars)
               END IF
               IF (ASSOCIATED(efield_fields(i)%efield%polarisation)) THEN
                  DEALLOCATE (efield_fields(i)%efield%polarisation)
               END IF
               DEALLOCATE (efield_fields(i)%efield)
            END IF
         END DO
         DEALLOCATE (efield_fields)
      END IF
   END SUBROUTINE efield_fields_release

! **************************************************************************************************
!> \brief ...
!> \param dftb_control ...
! **************************************************************************************************
   SUBROUTINE dftb_control_create(dftb_control)
      TYPE(dftb_control_type), POINTER                   :: dftb_control

      CHARACTER(len=*), PARAMETER :: routineN = 'dftb_control_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(dftb_control)))CALL cp__a("cp_control_types.F",985)
      ALLOCATE (dftb_control)

      NULLIFY (dftb_control%sk_pair_list)
   END SUBROUTINE dftb_control_create

! **************************************************************************************************
!> \brief ...
!> \param dftb_control ...
! **************************************************************************************************
   SUBROUTINE dftb_control_release(dftb_control)
      TYPE(dftb_control_type), POINTER                   :: dftb_control

      CHARACTER(len=*), PARAMETER :: routineN = 'dftb_control_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(dftb_control)) THEN
         IF (ASSOCIATED(dftb_control%sk_pair_list)) THEN
            DEALLOCATE (dftb_control%sk_pair_list)
         END IF
         DEALLOCATE (dftb_control)
      END IF
   END SUBROUTINE dftb_control_release

! **************************************************************************************************
!> \brief ...
!> \param scptb_control ...
! **************************************************************************************************
   SUBROUTINE scptb_control_create(scptb_control)
      TYPE(scptb_control_type), POINTER                  :: scptb_control

      CHARACTER(len=*), PARAMETER :: routineN = 'scptb_control_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(scptb_control)))CALL cp__a("cp_control_types.F",1019)
      ALLOCATE (scptb_control)

   END SUBROUTINE scptb_control_create

! **************************************************************************************************
!> \brief ...
!> \param scptb_control ...
! **************************************************************************************************
   SUBROUTINE scptb_control_release(scptb_control)
      TYPE(scptb_control_type), POINTER                  :: scptb_control

      CHARACTER(len=*), PARAMETER :: routineN = 'scptb_control_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(scptb_control)) THEN
         DEALLOCATE (scptb_control)
      END IF
   END SUBROUTINE scptb_control_release

! **************************************************************************************************
!> \brief ...
!> \param se_control ...
! **************************************************************************************************
   SUBROUTINE se_control_create(se_control)
      TYPE(semi_empirical_control_type), POINTER         :: se_control

      CHARACTER(len=*), PARAMETER :: routineN = 'se_control_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(se_control)))CALL cp__a("cp_control_types.F",1049)
      ALLOCATE (se_control)
   END SUBROUTINE se_control_create

! **************************************************************************************************
!> \brief ...
!> \param se_control ...
! **************************************************************************************************
   SUBROUTINE se_control_release(se_control)
      TYPE(semi_empirical_control_type), POINTER         :: se_control

      CHARACTER(len=*), PARAMETER :: routineN = 'se_control_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(se_control)) THEN
         DEALLOCATE (se_control)
      END IF
   END SUBROUTINE se_control_release

! **************************************************************************************************
!> \brief ...
!> \param admm_control ...
! **************************************************************************************************
   SUBROUTINE admm_control_create(admm_control)
      TYPE(admm_control_type), POINTER                   :: admm_control

      CHARACTER(len=*), PARAMETER :: routineN = 'admm_control_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(admm_control)))CALL cp__a("cp_control_types.F",1078)
      ALLOCATE (admm_control)

   END SUBROUTINE admm_control_create

! **************************************************************************************************
!> \brief ...
!> \param admm_control ...
! **************************************************************************************************
   SUBROUTINE admm_control_release(admm_control)
      TYPE(admm_control_type), POINTER                   :: admm_control

      CHARACTER(len=*), PARAMETER :: routineN = 'admm_control_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(admm_control)) THEN
         DEALLOCATE (admm_control)
      END IF
   END SUBROUTINE admm_control_release

! **************************************************************************************************
!> \brief   Create sccs_control_type
!> \param sccs_control ...
!> \par     History
!>          - Creation (11.10.2013,MK)
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE sccs_control_create(sccs_control)
      TYPE(sccs_control_type), POINTER                   :: sccs_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'sccs_control_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(sccs_control)))CALL cp__a("cp_control_types.F",1112)
      ALLOCATE (sccs_control)

      sccs_control%ref_count = 1
      sccs_control%derivative_method = 0
      sccs_control%max_iter = 0
      sccs_control%method_id = 0
      sccs_control%alpha_solvent = 0.0_dp
      sccs_control%beta = 0.0_dp
      sccs_control%beta_solvent = 0.0_dp
      sccs_control%delta_rho = 0.0_dp
      sccs_control%eps_sccs = 0.0_dp
      sccs_control%eps_scf = 0.0_dp
      sccs_control%epsilon_solvent = 0.0_dp
      sccs_control%gamma_solvent = 0.0_dp
      sccs_control%mixing = 0.0_dp
      sccs_control%rho_max = 0.0_dp
      sccs_control%rho_min = 0.0_dp
      sccs_control%rho_zero = 0.0_dp
      sccs_control%sccs_activated = .FALSE.

   END SUBROUTINE sccs_control_create

! **************************************************************************************************
!> \brief   Release sccs_control_type
!> \param sccs_control ...
!> \par     History
!>          - Creation (11.10.2013,MK)
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE sccs_control_release(sccs_control)
      TYPE(sccs_control_type), POINTER                   :: sccs_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'sccs_control_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(sccs_control)) THEN
         IF(.NOT.(sccs_control%ref_count > 0))CALL cp__a("cp_control_types.F",1150)
         sccs_control%ref_count = sccs_control%ref_count-1
         IF (sccs_control%ref_count == 0) THEN
            DEALLOCATE (sccs_control)
         END IF
      END IF

   END SUBROUTINE sccs_control_release

END MODULE cp_control_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Accelerator support
!> \author  Ole Schuett
!> \date    2013-04
! **************************************************************************************************
MODULE acc_stream
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'acc_stream'

   PUBLIC :: acc_stream_type
   PUBLIC :: acc_stream_create, acc_stream_destroy
   PUBLIC :: acc_stream_synchronize
   PUBLIC :: acc_stream_priority_range
   PUBLIC :: acc_stream_equal, acc_stream_associated
   PUBLIC :: acc_stream_cptr

   TYPE acc_stream_type
      PRIVATE
      INTEGER :: dummy = 1
   END TYPE acc_stream_type

CONTAINS

! **************************************************************************************************
!> \brief Returns C-pointer of given stream.
!> \param[in] this stream ID
!> \retval res false (accelerator support is not enabled)
!> \author  Ole Schuett
! **************************************************************************************************
   FUNCTION acc_stream_cptr(this) RESULT(res)
      INTEGER, INTENT(in)                                :: this
      LOGICAL                                            :: res

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
      res = .FALSE.
   END FUNCTION acc_stream_cptr


! **************************************************************************************************
!> \brief Fortran-wrapper for cudaStreamCreate.
!> \param[out] this stream
!> \param[in] name stream name
!> \param[in] priority (optional) stream priority
!> \author  Ole Schuett
! **************************************************************************************************
   SUBROUTINE acc_stream_create(this, name, priority)
      TYPE(acc_stream_type), INTENT(OUT) :: this
      CHARACTER(LEN=*), INTENT(IN)             :: name
      INTEGER, INTENT(IN), OPTIONAL            :: priority

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(name))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(priority))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("acc/acc_stream.F",127,"__ACC not compiled in.")
   END SUBROUTINE acc_stream_create

! **************************************************************************************************
!> \brief Fortran-wrapper for cudaStreamDestroy.
!> \param[in,out] this stream
!> \author  Ole Schuett
! **************************************************************************************************
   SUBROUTINE acc_stream_destroy(this)
      TYPE(acc_stream_type), &
         INTENT(INOUT)                          :: this

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("acc/acc_stream.F",156,"__ACC not compiled in.")
   END SUBROUTINE acc_stream_destroy

! **************************************************************************************************
!> \brief Fortran-wrapper for cudaStreamSynchronize.
!> \param[in] this stream
!> \author  Ole Schuett
! **************************************************************************************************
   SUBROUTINE acc_stream_synchronize(this)
      TYPE(acc_stream_type), &
         INTENT(IN)                             :: this

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("acc/acc_stream.F",179,"__ACC not compiled in.")
   END SUBROUTINE acc_stream_synchronize

! **************************************************************************************************
!> \brief Fortran-wrapper for cudaDeviceGetStreamPriorityRange.
!> \param[out] least least priority
!> \param[out] greatest greatest priroity
!> \author  Ole Schuett
! **************************************************************************************************
   SUBROUTINE acc_stream_priority_range(least, greatest)
      INTEGER, INTENT(OUT)                     :: least, greatest

      least = -1; greatest = -1 ! assign intent-out arguments to silence compiler warnings
      CALL cp__b("acc/acc_stream.F",201,"__ACC not compiled in.")
   END SUBROUTINE acc_stream_priority_range

! **************************************************************************************************
!> \brief Checks if two streams are equal
!> \param[in] this first stream
!> \param[in] other second stream
!> \retval res true if equal, false otherwise
!> \author  Ole Schuett
! **************************************************************************************************
   FUNCTION acc_stream_equal(this, other) RESULT(res)
      TYPE(acc_stream_type), INTENT(IN) :: this, other
      LOGICAL                                  :: res
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(other))==-1) EXIT ;  END DO ; ENDIF
      res = .TRUE.
   END FUNCTION acc_stream_equal

! **************************************************************************************************
!> \brief Checks if a streams is associated
!> \param[in] this stream
!> \retval res true if associated, false otherwise
!> \author  Ole Schuett
! **************************************************************************************************
   FUNCTION acc_stream_associated(this) RESULT(res)
      TYPE(acc_stream_type), INTENT(IN) :: this
      LOGICAL                                  :: res
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
      res = .FALSE.
   END FUNCTION acc_stream_associated

END MODULE acc_stream
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Accelerator support
!> \author  Ole Schuett
!> \date    2013-04
! **************************************************************************************************
MODULE acc_event
  USE acc_stream,                      ONLY: acc_stream_cptr,&
                                             acc_stream_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'acc_event'

   PUBLIC :: acc_event_type
   PUBLIC :: acc_event_create, acc_event_destroy
   PUBLIC :: acc_event_record, acc_event_query
   PUBLIC :: acc_stream_wait_event, acc_event_synchronize

   TYPE acc_event_type
      PRIVATE
      INTEGER :: dummy = 1
   END TYPE acc_event_type


CONTAINS

! **************************************************************************************************
!> \brief Fortran-wrapper for cudaStreamWaitEvent.
!>        Because of fortran circular dependency restriction this can not go into acc_stream.F
!> \param[in] stream stream
!> \param[in] event event
!> \author  Ole Schuett
! **************************************************************************************************
   SUBROUTINE acc_stream_wait_event(stream, event)
      TYPE(acc_stream_type), INTENT(IN) :: stream
      TYPE(acc_event_type), INTENT(IN)  :: event

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(event))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("acc/acc_event.F",112,"__ACC not compiled in.")
   END SUBROUTINE acc_stream_wait_event

! **************************************************************************************************
!> \brief Fortran-wrapper for cudaEventRecord.
!> \param[in] this event
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
   SUBROUTINE acc_event_record(this, stream)
      TYPE(acc_event_type), INTENT(IN)  :: this
      TYPE(acc_stream_type), INTENT(IN) :: stream

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("acc/acc_event.F",141,"__ACC not compiled in.")
   END SUBROUTINE acc_event_record

! **************************************************************************************************
!> \brief Fortran-wrapper for cudaEventCreate.
!> \param[in,out] this event
!> \author  Ole Schuett
! **************************************************************************************************
   SUBROUTINE acc_event_create(this)
      TYPE(acc_event_type), &
         INTENT(INOUT)                          :: this

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("acc/acc_event.F",168,"__ACC not compiled in.")
   END SUBROUTINE acc_event_create

! **************************************************************************************************
!> \brief Fortran-wrapper for cudaEventDestroy.
!> \param[in,out] this event
!> \author  Ole Schuett
! **************************************************************************************************
   SUBROUTINE acc_event_destroy(this)
      TYPE(acc_event_type), &
         INTENT(INOUT)                          :: this

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("acc/acc_event.F",191,"__ACC not compiled in.")
   END SUBROUTINE acc_event_destroy

! **************************************************************************************************
!> \brief Fortran-wrapper for cudaEventQuery.
!> \param[in] this event
!> \retval res true if event has occured, false otherwise
!> \author  Ole Schuett
! **************************************************************************************************
   FUNCTION acc_event_query(this) RESULT(res)
      TYPE(acc_event_type), INTENT(IN)         :: this
      LOGICAL                                  :: res

      res = .FALSE.
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("acc/acc_event.F",216,"__ACC not compiled in.")
   END FUNCTION acc_event_query

! **************************************************************************************************
!> \brief Fortran-wrapper for cudaEventSynchronize.
!> \param[in] this event
!> \author  Ole Schuett
! **************************************************************************************************
   SUBROUTINE acc_event_synchronize(this)
      TYPE(acc_event_type), INTENT(IN)  :: this

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("acc/acc_event.F",238,"__ACC not compiled in.")
   END SUBROUTINE acc_event_synchronize

END MODULE acc_event
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Defines the basic variable types
!> \author vw
! **************************************************************************************************
MODULE acc_kinds

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: int_4, int_8
   PUBLIC :: real_4, real_8
   PUBLIC :: int_4_size, int_8_size
   PUBLIC :: real_4_size, real_8_size

   INTEGER, PARAMETER :: real_4 = SELECTED_REAL_KIND(6, 30)
   INTEGER, PARAMETER :: real_8 = SELECTED_REAL_KIND(14, 200)
   INTEGER, PARAMETER :: real_4_size = 4
   INTEGER, PARAMETER :: real_8_size = 8

   INTEGER, PARAMETER :: int_1 = SELECTED_INT_KIND(2)
   INTEGER, PARAMETER :: int_1_size = BIT_SIZE(INT(0, int_1))/8

   INTEGER, PARAMETER :: int_2 = SELECTED_INT_KIND(4)
   INTEGER, PARAMETER :: int_2_size = BIT_SIZE(INT(0, int_2))/8

   INTEGER, PARAMETER :: int_4 = SELECTED_INT_KIND(5)
   INTEGER, PARAMETER :: int_4_size = BIT_SIZE(INT(0, int_4))/8

   INTEGER, PARAMETER :: int_8 = SELECTED_INT_KIND(10)
   INTEGER, PARAMETER :: int_8_size = BIT_SIZE(INT(0, int_8))/8

   INTEGER, PARAMETER :: sp = real_4
   INTEGER, PARAMETER :: dp = real_8

   INTEGER, PARAMETER :: default_string_length = 80

END MODULE acc_kinds

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Accelerator support
!> \author  Ole Schuett
! **************************************************************************************************
MODULE acc_devmem
  USE acc_kinds,                       ONLY: int_4,&
                                             int_4_size,&
                                             int_8,&
                                             int_8_size,&
                                             real_4,&
                                             real_4_size,&
                                             real_8,&
                                             real_8_size
  USE acc_stream,                      ONLY: acc_stream_associated,&
                                             acc_stream_cptr,&
                                             acc_stream_synchronize,&
                                             acc_stream_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'acc_devmem'

   PUBLIC :: acc_devmem_type
   PUBLIC :: acc_devmem_allocate_bytes, acc_devmem_deallocate
   PUBLIC :: acc_devmem_setzero_bytes
   PUBLIC :: acc_devmem_allocated
   PUBLIC :: acc_devmem_dev2host, acc_devmem_host2dev
   PUBLIC :: acc_devmem_size_in_bytes
   PUBLIC :: acc_devmem_ensure_size_bytes
   PUBLIC :: acc_devmem_cptr

   INTERFACE acc_devmem_dev2host
      MODULE PROCEDURE dev2host_i4_1D
      MODULE PROCEDURE dev2host_i8_1D
      MODULE PROCEDURE dev2host_r4_1D
      MODULE PROCEDURE dev2host_r8_1D
      MODULE PROCEDURE dev2host_c4_1D
      MODULE PROCEDURE dev2host_c8_1D
   END INTERFACE acc_devmem_dev2host

   INTERFACE acc_devmem_host2dev
      MODULE PROCEDURE host2dev_i4_1D
      MODULE PROCEDURE host2dev_i8_1D
      MODULE PROCEDURE host2dev_r4_1D
      MODULE PROCEDURE host2dev_r8_1D
      MODULE PROCEDURE host2dev_c4_1D
      MODULE PROCEDURE host2dev_c8_1D
      MODULE PROCEDURE host2dev_i4_2D
      MODULE PROCEDURE host2dev_i8_2D
      MODULE PROCEDURE host2dev_r4_2D
      MODULE PROCEDURE host2dev_r8_2D
      MODULE PROCEDURE host2dev_c4_2D
      MODULE PROCEDURE host2dev_c8_2D
   END INTERFACE acc_devmem_host2dev

   TYPE acc_devmem_type
      PRIVATE
      INTEGER                      :: size_in_bytes = -1
   END TYPE acc_devmem_type


CONTAINS

! **************************************************************************************************
!> \brief Ensures that given devmem has at least the requested size.
!> \param[in,out] this device memory
!> \param[in] stream on which zeroing and memcopying is performed
!> \param[in] requested_size_in_bytes requested size in bytes
!> \param[in] nocopy (optional) if after growin old content should NOT be copied over. Default: false.
!> \param[in] zero_pad (optional) if after growing the new memory should be zeroed. Default: false.
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE acc_devmem_ensure_size_bytes(this, stream, requested_size_in_bytes, nocopy, zero_pad)
      TYPE(acc_devmem_type), &
         INTENT(INOUT)                          :: this
      TYPE(acc_stream_type), INTENT(IN) :: stream
      INTEGER, INTENT(IN)                      :: requested_size_in_bytes
      LOGICAL, INTENT(IN), OPTIONAL            :: nocopy, zero_pad

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(requested_size_in_bytes))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(nocopy))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(zero_pad))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("acc/acc_devmem.F",186,"__ACC not compiled in.")
   END SUBROUTINE acc_devmem_ensure_size_bytes

! **************************************************************************************************
!> \brief Returns a logical, which indicates if the given devmem is allocated.
!> \param[in] this device memory
!> \retval res true if device memory is allocated, false otherwise
!> \author  Ole Schuett
! **************************************************************************************************
   FUNCTION acc_devmem_allocated(this) RESULT(res)
      TYPE(acc_devmem_type), INTENT(IN)                  :: this
      LOGICAL                                            :: res

      res = this%size_in_bytes >= 0
   END FUNCTION acc_devmem_allocated

! **************************************************************************************************
!> \brief Returns size of given devmem in terms of item count (not bytes!)
!> \param[in] this device memory
!> \retval res size of device memory (item count)
!> \author  Ole Schuett
! **************************************************************************************************
   FUNCTION acc_devmem_size_in_bytes(this) RESULT(res)
      TYPE(acc_devmem_type), INTENT(IN)                  :: this
      INTEGER                                            :: res

      IF (this%size_in_bytes < 0) &
         CALL cp__b("acc/acc_devmem.F",260,"acc_devmem_len: not allocated")
      res = this%size_in_bytes
   END FUNCTION acc_devmem_size_in_bytes

! **************************************************************************************************
!> \brief Returns C-pointer to data of given devmem.
!> \param[in] this device memory
!> \retval res false (accelerator support is not enabled)
!> \author  Ole Schuett
! **************************************************************************************************
   FUNCTION acc_devmem_cptr(this) RESULT(res)
      INTEGER, INTENT(IN)                                :: this
      LOGICAL                                            :: res

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
      res = .FALSE.
   END FUNCTION acc_devmem_cptr

! **************************************************************************************************
!> \brief Fortran-wrapper for cudaMemGetInfo.
!> \param[out] free free device memory
!> \param[out] avail available device memory
!> \author  Ole Schuett
! **************************************************************************************************
   SUBROUTINE acc_devmem_info(free, avail)
      INTEGER, INTENT(OUT)                     :: free, avail

      avail = -1; free = -1 ! assign intent-out arguments to silence compiler warnings
      CALL cp__b("acc/acc_devmem.F",306,"__ACC not compiled in.")
   END SUBROUTINE acc_devmem_info

! **************************************************************************************************
!> \brief Allocates a given devmem.
!> \param[in,out] this device memory
!> \param[in] size_in_bytes size in bytes
!> \author  Ole Schuett
! **************************************************************************************************
   SUBROUTINE acc_devmem_allocate_bytes(this, size_in_bytes)
      TYPE(acc_devmem_type), INTENT(INOUT)     :: this
      INTEGER, INTENT(IN)                      :: size_in_bytes

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(size_in_bytes))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("acc/acc_devmem.F",328,"__ACC not compiled in.")
   END SUBROUTINE acc_devmem_allocate_bytes

! **************************************************************************************************
!> \brief Deallocates a given devmem.
!> \param[in,out] this device memory
!> \author  Ole Schuett
! **************************************************************************************************
   SUBROUTINE acc_devmem_deallocate(this)
      TYPE(acc_devmem_type), INTENT(INOUT) :: this

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("acc/acc_devmem.F",356,"__ACC not compiled in.")
   END SUBROUTINE acc_devmem_deallocate

! **************************************************************************************************
!> \brief Sets entries in given devmem to zero, asynchronously.
!> \param[in,out] this device memory
!> \param[in] first_byte (optional) begin of region to zero, defaults to 1 if not given.
!> \param[in] last_byte (optional) end of region to zero, defaults to size if not given.
!> \param[in] stream stream on which zeroing is performed.
!> \author  Ole Schuett
! **************************************************************************************************
   SUBROUTINE acc_devmem_setzero_bytes(this, first_byte, last_byte, stream)
      TYPE(acc_devmem_type), INTENT(INOUT) :: this
      INTEGER, INTENT(IN), OPTIONAL        :: first_byte, last_byte
      TYPE(acc_stream_type), INTENT(IN)    :: stream

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(first_byte))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(last_byte))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("acc/acc_devmem.F",394,"__ACC not compiled in.")
   END SUBROUTINE acc_devmem_setzero_bytes

! **************************************************************************************************
!> \brief Helper-routine performing actuall host2dev transfers.
!> \param[in] this device memory
!> \param hostmem_cptr host memory
!> \param[in] n_bytes number of bytes
!> \param[in] stream stream used for memory transfer
!> \author  Ole Schuett
! **************************************************************************************************

! **************************************************************************************************
!> \brief Helper-routine performing actual dev2host transfers.
!> \param[in] this device memory
!> \param hostmem_cptr host memory
!> \param[in] n_bytes number of bytes
!> \param[in] stream stream used for memory transfer
!> \author  Ole Schuett
! **************************************************************************************************

! **************************************************************************************************
!> \brief Transfers 1D fortran-array from host to cuda devmem.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE host2dev_i4_1D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    INTEGER(kind=int_4), DIMENSION(:), POINTER :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",17,"__ACC not compiled in.")
 END SUBROUTINE host2dev_i4_1D


! **************************************************************************************************
!> \brief Transfers 2D fortran-array from host to cuda devmem.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE host2dev_i4_2D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    INTEGER(kind=int_4), DIMENSION(:, :), POINTER         :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",40,"__ACC not compiled in.")
 END SUBROUTINE host2dev_i4_2D


! **************************************************************************************************
!> \brief Transfers cuda devmem to 1D fortran-array.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE dev2host_i4_1D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    INTEGER(kind=int_4), DIMENSION(:), POINTER            :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",63,"__ACC not compiled in.")
 END SUBROUTINE dev2host_i4_1D

! **************************************************************************************************
!> \brief Transfers 1D fortran-array from host to cuda devmem.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE host2dev_i8_1D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    INTEGER(kind=int_8), DIMENSION(:), POINTER :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",17,"__ACC not compiled in.")
 END SUBROUTINE host2dev_i8_1D


! **************************************************************************************************
!> \brief Transfers 2D fortran-array from host to cuda devmem.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE host2dev_i8_2D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    INTEGER(kind=int_8), DIMENSION(:, :), POINTER         :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",40,"__ACC not compiled in.")
 END SUBROUTINE host2dev_i8_2D


! **************************************************************************************************
!> \brief Transfers cuda devmem to 1D fortran-array.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE dev2host_i8_1D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    INTEGER(kind=int_8), DIMENSION(:), POINTER            :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",63,"__ACC not compiled in.")
 END SUBROUTINE dev2host_i8_1D

! **************************************************************************************************
!> \brief Transfers 1D fortran-array from host to cuda devmem.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE host2dev_r4_1D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    REAL(kind=real_4), DIMENSION(:), POINTER :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",17,"__ACC not compiled in.")
 END SUBROUTINE host2dev_r4_1D


! **************************************************************************************************
!> \brief Transfers 2D fortran-array from host to cuda devmem.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE host2dev_r4_2D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    REAL(kind=real_4), DIMENSION(:, :), POINTER         :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",40,"__ACC not compiled in.")
 END SUBROUTINE host2dev_r4_2D


! **************************************************************************************************
!> \brief Transfers cuda devmem to 1D fortran-array.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE dev2host_r4_1D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    REAL(kind=real_4), DIMENSION(:), POINTER            :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",63,"__ACC not compiled in.")
 END SUBROUTINE dev2host_r4_1D

! **************************************************************************************************
!> \brief Transfers 1D fortran-array from host to cuda devmem.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE host2dev_r8_1D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    REAL(kind=real_8), DIMENSION(:), POINTER :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",17,"__ACC not compiled in.")
 END SUBROUTINE host2dev_r8_1D


! **************************************************************************************************
!> \brief Transfers 2D fortran-array from host to cuda devmem.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE host2dev_r8_2D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    REAL(kind=real_8), DIMENSION(:, :), POINTER         :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",40,"__ACC not compiled in.")
 END SUBROUTINE host2dev_r8_2D


! **************************************************************************************************
!> \brief Transfers cuda devmem to 1D fortran-array.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE dev2host_r8_1D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    REAL(kind=real_8), DIMENSION(:), POINTER            :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",63,"__ACC not compiled in.")
 END SUBROUTINE dev2host_r8_1D

! **************************************************************************************************
!> \brief Transfers 1D fortran-array from host to cuda devmem.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE host2dev_c4_1D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    COMPLEX(kind=real_4), DIMENSION(:), POINTER :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",17,"__ACC not compiled in.")
 END SUBROUTINE host2dev_c4_1D


! **************************************************************************************************
!> \brief Transfers 2D fortran-array from host to cuda devmem.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE host2dev_c4_2D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    COMPLEX(kind=real_4), DIMENSION(:, :), POINTER         :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",40,"__ACC not compiled in.")
 END SUBROUTINE host2dev_c4_2D


! **************************************************************************************************
!> \brief Transfers cuda devmem to 1D fortran-array.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE dev2host_c4_1D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    COMPLEX(kind=real_4), DIMENSION(:), POINTER            :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",63,"__ACC not compiled in.")
 END SUBROUTINE dev2host_c4_1D

! **************************************************************************************************
!> \brief Transfers 1D fortran-array from host to cuda devmem.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE host2dev_c8_1D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    COMPLEX(kind=real_8), DIMENSION(:), POINTER :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",17,"__ACC not compiled in.")
 END SUBROUTINE host2dev_c8_1D


! **************************************************************************************************
!> \brief Transfers 2D fortran-array from host to cuda devmem.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE host2dev_c8_2D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    COMPLEX(kind=real_8), DIMENSION(:, :), POINTER         :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",40,"__ACC not compiled in.")
 END SUBROUTINE host2dev_c8_2D


! **************************************************************************************************
!> \brief Transfers cuda devmem to 1D fortran-array.
!> \param[in] this device memory
!> \param hostmem host memory
!> \param[in] stream stream
!> \author  Ole Schuett
! **************************************************************************************************
 SUBROUTINE dev2host_c8_1D(this, hostmem, stream)
    TYPE(acc_devmem_type), INTENT(IN) :: this
    COMPLEX(kind=real_8), DIMENSION(:), POINTER            :: hostmem
    TYPE(acc_stream_type), INTENT(IN) :: stream

    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(this))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hostmem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_devmem.F",63,"__ACC not compiled in.")
 END SUBROUTINE dev2host_c8_1D


END MODULE acc_devmem
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Types related to DBCSR data area
!> \author  Ole Schuett
!> \date    10-2014
! **************************************************************************************************
MODULE dbcsr_data_types
   USE acc_devmem,                      ONLY: acc_devmem_type
   USE acc_event,                       ONLY: acc_event_type
   USE acc_stream,                      ONLY: acc_stream_type
   USE kinds,                           ONLY: &
        dp, int_4, int_4_size, int_8, int_8_size, real_4, real_4_size, real_8, real_8_size

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads, omp_lock_kind

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_data_types'

   PUBLIC :: dbcsr_data_obj, dbcsr_data_area_type, dbcsr_scalar_type
   PUBLIC :: dbcsr_datatype_sizeof
   PUBLIC :: dbcsr_mempool_type, &
             dbcsr_mempool_entry_type, &
             dbcsr_memtype_type, &
             dbcsr_memtype_default
   PUBLIC :: dbcsr_type_real_4, dbcsr_type_real_8, &
             dbcsr_type_complex_4, dbcsr_type_complex_8, &
             dbcsr_type_real_default, dbcsr_type_complex_default, &
             dbcsr_type_real_4_2d, dbcsr_type_real_8_2d, &
             dbcsr_type_complex_4_2d, dbcsr_type_complex_8_2d, &
             dbcsr_type_int_4, dbcsr_type_int_8

! **************************************************************************************************
!> \brief Stores a scalar in any of the supported data types.
!> \par Reasoning
!>      Easier routine interfaces
!> \var r_sp        stores real values in single precision
!> \var r_dp        stores real values in double precision
!> \var c_sp        stores complex values in single precision
!> \var c_dp        stores complex values in double precision
!> \var refcount    reference counter for current structure
!> \var data_type   which of the data types is actually used
! **************************************************************************************************
   TYPE dbcsr_scalar_type
      REAL(KIND=real_4)    :: r_sp
      REAL(KIND=real_8)    :: r_dp
      COMPLEX(KIND=real_4) :: c_sp
      COMPLEX(KIND=real_8) :: c_dp
      INTEGER              :: data_type
   END TYPE dbcsr_scalar_type

   TYPE dbcsr_data_obj
      TYPE(dbcsr_data_area_type), POINTER           :: d => Null()
   END TYPE dbcsr_data_obj

! **************************************************************************************************
!> \brief Memory related types
! **************************************************************************************************
   TYPE dbcsr_mempool_type
      TYPE(dbcsr_mempool_entry_type), POINTER :: root => Null()
      INTEGER                                 :: capacity = 1
!$    INTEGER(KIND=omp_lock_kind)          :: lock
   END TYPE dbcsr_mempool_type

   TYPE dbcsr_mempool_entry_type
      TYPE(dbcsr_data_obj)                    :: area
      TYPE(dbcsr_mempool_entry_type), POINTER :: next => Null()
   END TYPE dbcsr_mempool_entry_type

   TYPE dbcsr_memtype_type
      LOGICAL                           :: mpi = .FALSE.
      LOGICAL                           :: acc_hostalloc = .FALSE.
      LOGICAL                           :: acc_devalloc = .FALSE.
      TYPE(acc_stream_type)             :: acc_stream = acc_stream_type()
      TYPE(dbcsr_mempool_type), POINTER :: pool => Null()
      REAL(KIND=dp)                     :: oversize_factor = 1.0
   END TYPE dbcsr_memtype_type

   !providing pool=Null() explicitly to circumvent bug in ifort 12.1
   TYPE(dbcsr_memtype_type), PARAMETER :: dbcsr_memtype_default = dbcsr_memtype_type(pool=Null())

! **************************************************************************************************
!> \brief Stores actual data
!> \var r_sp        stores real values in single precision
!> \var r_dp        stores real values in double precision
!> \var c_sp        stores complex values in single precision
!> \var c_dp        stores complex values in double precision
!> \var ref_size    last data element in array actually referenced
!> \var refcount    reference counter for current structure
!> \var memory_type   type of memory where data lives
!> \var data_type   which of the data types is actually used
! **************************************************************************************************
   TYPE dbcsr_data_area_type
      INTEGER(KIND=int_4), DIMENSION(:), POINTER    :: i4 => Null()
      INTEGER(KIND=int_8), DIMENSION(:), POINTER    :: i8 => Null()
      REAL(KIND=real_4), DIMENSION(:), POINTER      :: r_sp => Null()
      REAL(KIND=real_8), DIMENSION(:), POINTER      :: r_dp => Null()
      COMPLEX(KIND=real_4), DIMENSION(:), POINTER   :: c_sp => Null()
      COMPLEX(KIND=real_8), DIMENSION(:), POINTER   :: c_dp => Null()
      REAL(KIND=real_4), DIMENSION(:, :), POINTER    :: r2_sp => Null()
      REAL(KIND=real_8), DIMENSION(:, :), POINTER    :: r2_dp => Null()
      COMPLEX(KIND=real_4), DIMENSION(:, :), POINTER :: c2_sp => Null()
      COMPLEX(KIND=real_8), DIMENSION(:, :), POINTER :: c2_dp => Null()
      INTEGER                                  :: ref_size = -1
      INTEGER                                  :: refcount = -1
      INTEGER                                  :: id = -1
      TYPE(dbcsr_memtype_type)                 :: memory_type = dbcsr_memtype_default
      INTEGER                                  :: data_type = -1
      TYPE(acc_devmem_type)                    :: acc_devmem
      TYPE(acc_event_type)                     :: acc_ready
   END TYPE dbcsr_data_area_type

!> Type definitions:
!> * bit 0: always 1
!> * bit 1: single (0: 4) vs. double (1: 1)
!> * bit 2: real (0) vs. complex (1)
!> * bit 3: dimension (0: 1, 1: 2)
!> * bit 4: floating point (0) or integer type (1)
   INTEGER, PARAMETER          :: dbcsr_type_real_4 = 1 !001
   INTEGER, PARAMETER          :: dbcsr_type_real_8 = 3 !011
   INTEGER, PARAMETER          :: dbcsr_type_complex_4 = 5 !101
   INTEGER, PARAMETER          :: dbcsr_type_complex_8 = 7 !111

   INTEGER, PARAMETER          :: dbcsr_type_real_4_2d = 9 !1001
   INTEGER, PARAMETER          :: dbcsr_type_real_8_2d = 11 !1011
   INTEGER, PARAMETER          :: dbcsr_type_complex_4_2d = 13 !1101
   INTEGER, PARAMETER          :: dbcsr_type_complex_8_2d = 15 !1111

   INTEGER, PARAMETER          :: dbcsr_type_int_4 = 17 !10001
   INTEGER, PARAMETER          :: dbcsr_type_int_8 = 19 !10011

   INTEGER, PARAMETER          :: dbcsr_type_real_default = dbcsr_type_real_8
   INTEGER, PARAMETER          :: dbcsr_type_complex_default = dbcsr_type_complex_8

CONTAINS

! **************************************************************************************************
!> \brief Helper-routine, returns size of given datatype in terms of bytes.
!> \param datatype ...
!> \retval size ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION dbcsr_datatype_sizeof(datatype) RESULT(size)
      INTEGER, INTENT(IN)                                :: datatype
      INTEGER                                            :: size

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_datatype_sizeof', &
         routineP = moduleN//':'//routineN

      SELECT CASE (datatype)
      CASE (dbcsr_type_int_4)
         size = int_4_size
      CASE (dbcsr_type_int_8)
         size = int_8_size
      CASE (dbcsr_type_real_4)
         size = real_4_size
      CASE (dbcsr_type_real_8)
         size = real_8_size
      CASE (dbcsr_type_complex_4)
         size = (2*real_4_size)
      CASE (dbcsr_type_complex_8)
         size = (2*real_8_size)
      CASE default
         CALL cp__b("dbcsr/data/dbcsr_data_types.F",171,"Invalid data type")
      END SELECT
   END FUNCTION dbcsr_datatype_sizeof

END MODULE dbcsr_data_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Array objects with reference counting.
!> \author  Urban Borstnik
!> \date    2009-05-12
!> \version 1.0
!>
!> <b>Modification history:</b>
!> - Created
! **************************************************************************************************
MODULE array_types

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: array_i1d_obj
   PUBLIC :: array_new, &
             array_hold, &
             array_release, &
             array_nullify, &
             array_exists
   PUBLIC :: array_data, &
             array_size, &
             array_equality
   PUBLIC :: array_get

   INTERFACE array_new
      MODULE PROCEDURE array_new_i1d, array_new_i1d_lb
   END INTERFACE
   INTERFACE array_hold
      MODULE PROCEDURE array_hold_i1d
   END INTERFACE
   INTERFACE array_release
      MODULE PROCEDURE array_release_i1d
   END INTERFACE
   INTERFACE array_nullify
      MODULE PROCEDURE array_nullify_i1d
   END INTERFACE
   INTERFACE array_exists
      MODULE PROCEDURE array_exists_i1d
   END INTERFACE
   INTERFACE array_data
      MODULE PROCEDURE array_data_i1d
   END INTERFACE
   INTERFACE array_size
      MODULE PROCEDURE array_size_i1d
   END INTERFACE

   INTERFACE array_equality
      MODULE PROCEDURE array_equality_i1
      MODULE PROCEDURE array_equality_i1d
   END INTERFACE

   INTERFACE array_get
      MODULE PROCEDURE array_get_i1d
      MODULE PROCEDURE array_get_i1
   END INTERFACE

   TYPE array_i1d_type
      INTEGER, DIMENSION(:), POINTER     :: DATA
      INTEGER                            :: refcount
   END TYPE array_i1d_type
   TYPE array_i1d_obj
      TYPE(array_i1d_type), POINTER      :: low
   END TYPE array_i1d_obj

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param DATA ...
!> \param gift ...
! **************************************************************************************************
   SUBROUTINE array_new_i1d(array, DATA, gift)
      TYPE(array_i1d_obj), INTENT(OUT)                   :: array
      INTEGER, DIMENSION(:), POINTER                     :: DATA
      LOGICAL, INTENT(IN), OPTIONAL                      :: gift

      INTEGER                                            :: lb, ub
      LOGICAL                                            :: g

      ALLOCATE (array%low)
      array%low%refcount = 1
      g = .FALSE.
      IF (PRESENT(gift)) g = gift
      IF (g) THEN
         array%low%data => DATA
         NULLIFY (DATA)
      ELSE
         lb = LBOUND(DATA, 1)
         ub = UBOUND(DATA, 1)
         ALLOCATE (array%low%data(lb:ub))
         array%low%data(:) = DATA(:)
      ENDIF
   END SUBROUTINE array_new_i1d
! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param DATA ...
!> \param lb ...
! **************************************************************************************************
   SUBROUTINE array_new_i1d_lb(array, DATA, lb)
      TYPE(array_i1d_obj), INTENT(OUT)                   :: array
      INTEGER, DIMENSION(:), INTENT(IN)                  :: DATA
      INTEGER, INTENT(IN)                                :: lb

      INTEGER                                            :: ub

      ALLOCATE (array%low)
      array%low%refcount = 1
      ub = lb+SIZE(DATA)-1
      ALLOCATE (array%low%data(lb:ub))
      array%low%data(:) = DATA(:)
   END SUBROUTINE array_new_i1d_lb
! **************************************************************************************************
!> \brief ...
!> \param array ...
! **************************************************************************************************
   SUBROUTINE array_hold_i1d(array)
      TYPE(array_i1d_obj), INTENT(INOUT)                 :: array

!$OMP ATOMIC
      array%low%refcount = array%low%refcount+1
   END SUBROUTINE array_hold_i1d
! **************************************************************************************************
!> \brief ...
!> \param array ...
! **************************************************************************************************
   SUBROUTINE array_release_i1d(array)
      TYPE(array_i1d_obj), INTENT(INOUT)                 :: array

      IF (ASSOCIATED(array%low)) THEN
         array%low%refcount = array%low%refcount-1
         IF (array%low%refcount .EQ. 0) THEN
            DEALLOCATE (array%low%data)
            DEALLOCATE (array%low)
         ENDIF
      ENDIF
   END SUBROUTINE array_release_i1d
! **************************************************************************************************
!> \brief ...
!> \param array ...
! **************************************************************************************************
   PURE SUBROUTINE array_nullify_i1d(array)
      TYPE(array_i1d_obj), INTENT(INOUT)                 :: array

      NULLIFY (array%low)
   END SUBROUTINE array_nullify_i1d
! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \retval array_exists ...
! **************************************************************************************************
   PURE FUNCTION array_exists_i1d(array) RESULT(array_exists)
      TYPE(array_i1d_obj), INTENT(IN)                    :: array
      LOGICAL                                            :: array_exists

      array_exists = ASSOCIATED(array%low)
      IF (array_exists) array_exists = array%low%refcount .GT. 0
   END FUNCTION array_exists_i1d
! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \retval DATA ...
! **************************************************************************************************
   FUNCTION array_data_i1d(array) RESULT(DATA)
      TYPE(array_i1d_obj), INTENT(IN)                    :: array
      INTEGER, DIMENSION(:), POINTER                     :: DATA

      IF (ASSOCIATED(array%low)) THEN
         DATA => array%low%data
      ELSE
         NULLIFY (DATA)
      ENDIF
   END FUNCTION array_data_i1d
! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \retval the_size ...
! **************************************************************************************************
   PURE FUNCTION array_size_i1d(array) RESULT(the_size)
      TYPE(array_i1d_obj), INTENT(IN)                    :: array
      INTEGER                                            :: the_size

      IF (ASSOCIATED(array%low)) THEN
         the_size = SIZE(array%low%data)
      ELSE
         the_size = 0
      ENDIF
   END FUNCTION array_size_i1d
! **************************************************************************************************
!> \brief ...
!> \param array1 ...
!> \param array2 ...
!> \retval are_equal ...
! **************************************************************************************************
   PURE FUNCTION array_equality_i1(array1, array2) RESULT(are_equal)
      INTEGER, DIMENSION(:), POINTER                     :: array1, array2
      LOGICAL                                            :: are_equal

      are_equal = .FALSE.
      IF (ASSOCIATED(array1) .AND. ASSOCIATED(array2)) THEN
         IF (SIZE(array1) .NE. SIZE(array2)) RETURN
         are_equal = ALL(array1 .EQ. array2)
      ENDIF
   END FUNCTION array_equality_i1
! **************************************************************************************************
!> \brief ...
!> \param array1 ...
!> \param array2 ...
!> \retval are_equal ...
! **************************************************************************************************
   PURE FUNCTION array_equality_i1d(array1, array2) RESULT(are_equal)
      TYPE(array_i1d_obj), INTENT(IN)                    :: array1, array2
      LOGICAL                                            :: are_equal

      are_equal = .FALSE.
      IF (ASSOCIATED(array1%low) .AND. ASSOCIATED(array2%low)) THEN
         IF (SIZE(array1%low%data) .NE. SIZE(array2%low%data)) RETURN
         are_equal = ALL(array1%low%data .EQ. array2%low%data)
      ENDIF
   END FUNCTION array_equality_i1d
! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param index1 ...
!> \retval value ...
! **************************************************************************************************
   PURE FUNCTION array_get_i1d(array, index1) RESULT(value)
      TYPE(array_i1d_obj), INTENT(IN)                    :: array
      INTEGER, INTENT(IN)                                :: index1
      INTEGER                                            :: value

      value = array%low%data(index1)
   END FUNCTION array_get_i1d
! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param index1 ...
!> \retval value ...
! **************************************************************************************************
   PURE FUNCTION array_get_i1(array, index1) RESULT(value)
      INTEGER, DIMENSION(:), INTENT(IN), POINTER         :: array
      INTEGER, INTENT(IN)                                :: index1
      INTEGER                                            :: value

      value = array(index1)
   END FUNCTION array_get_i1
END MODULE array_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   B-tree implementation template
!> \author  Urban Borstnik
!> \date    2000-05-12
!> \version 1.0
!> <b>Modification history:</b>
!> - Created
! **************************************************************************************************

MODULE btree_i8_k_sp2d_v

   IMPLICIT NONE
   PRIVATE

   !API
   PUBLIC :: btree, btree_new, btree_add, btree_find, &
             btree_delete, btree_get_entries

   INTEGER, PARAMETER :: keyt = SELECTED_INT_KIND(10)
   INTEGER, PARAMETER :: valt = SELECTED_INT_KIND(5); 
   INTEGER, PARAMETER :: sp = KIND(0.0)

   TYPE sp2d
      REAL(KIND=sp), DIMENSION(:, :), POINTER :: p
      LOGICAL :: tr
   END TYPE sp2d

   PUBLIC :: sp2d

   TYPE btree_node
      INTEGER id
      INTEGER :: filled
      INTEGER(KIND=keyt), DIMENSION(:), POINTER :: keys
      TYPE(sp2d), DIMENSION(:), POINTER :: values
      TYPE(btree_node_p), DIMENSION(:), POINTER :: subtrees
      TYPE(btree_node), POINTER :: parent
   END TYPE btree_node

   TYPE btree_node_p
      TYPE(btree_node), POINTER :: node
   END TYPE btree_node_p

   TYPE btree_node_structure
      INTEGER :: min_fill, max_fill
      INTEGER :: n
      INTEGER :: lastid
      INTEGER :: refcount
      TYPE(btree_node), POINTER :: root
   END TYPE btree_node_structure

   TYPE btree
      TYPE(btree_node_structure) :: b
   END TYPE btree

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE btree_new(tree, order)
      TYPE(btree), INTENT(OUT)                           :: tree
      INTEGER, INTENT(IN), OPTIONAL                      :: order

      INTEGER                                            :: maxs, mins

      !
      IF (PRESENT(order)) THEN
         maxs = order-1
      ELSE
         maxs = 15
      ENDIF
      mins = ISHFT(maxs, -1)
      IF (mins*2 .GT. maxs) maxs = 2*maxs
      IF (mins .LT. 1) mins = 1
      IF (maxs .LT. 3) maxs = 3
      tree%b%min_fill = mins
      tree%b%max_fill = maxs
      tree%b%refcount = 1
      tree%b%n = 0
      NULLIFY (tree%b%root)
      tree%b%lastid = 0
   END SUBROUTINE btree_new

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \retval num_entries ...
! **************************************************************************************************
   FUNCTION btree_get_entries(tree) RESULT(num_entries)
      TYPE(btree), INTENT(INOUT)                         :: tree
      INTEGER                                            :: num_entries

      num_entries = tree%b%n
   END FUNCTION btree_get_entries

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param keys ...
!> \param values ...
! **************************************************************************************************
   SUBROUTINE btree_delete(tree, keys, values)
      TYPE(btree), INTENT(INOUT)                         :: tree
      INTEGER(KIND=keyt), DIMENSION(:), INTENT(OUT), &
         OPTIONAL                                        :: keys
      TYPE(sp2d), DIMENSION(:), INTENT(OUT), OPTIONAL    :: values

      INTEGER                                            :: pos

      !
      IF (ASSOCIATED(tree%b%root)) THEN
         pos = 0
         IF (PRESENT(keys) .AND. PRESENT(values)) THEN
            pos = 1
            CALL btree_delete_node(tree%b%root, pos, keys, values)
         ELSE
            CALL btree_delete_node(tree%b%root)
         ENDIF
      ENDIF
      NULLIFY (tree%b%root)
   END SUBROUTINE btree_delete

! **************************************************************************************************
!> \brief ...
!> \param node ...
!> \param pos ...
!> \param keys ...
!> \param values ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE btree_delete_node(node, pos, keys, values)
      TYPE(btree_node), POINTER                          :: node
      INTEGER, INTENT(INOUT), OPTIONAL                   :: pos
      INTEGER(KIND=keyt), DIMENSION(:), INTENT(INOUT), &
         OPTIONAL                                        :: keys
      TYPE(sp2d), DIMENSION(:), INTENT(INOUT), OPTIONAL  :: values

      INTEGER                                            :: i

!

      !
      IF (node%filled .GT. 0 .AND. ASSOCIATED(node%subtrees(1)%node)) THEN
         DO i = 1, node%filled+1
            IF (PRESENT(pos)) THEN
               CALL btree_delete_node(node%subtrees(i)%node, pos, keys, values)
            ELSE
               CALL btree_delete_node(node%subtrees(i)%node)
            ENDIF
            IF (PRESENT(pos) .AND. i .LE. node%filled) THEN
               keys(pos) = node%keys(i)
               values(pos) = node%values(i)
               pos = pos+1
            ENDIF
         ENDDO
      ELSEIF (PRESENT(pos) .AND. node%filled .GT. 0) THEN
         keys(pos:pos+node%filled-1) = node%keys(1:node%filled)
         values(pos:pos+node%filled-1) = node%values(1:node%filled)
         pos = pos+node%filled
      ENDIF
      CALL btree_free_node(node)
   END SUBROUTINE btree_delete_node

   ! Find the key
   ! IF node still has space, insert & update the node
   ! else
   ! 1. select median
   ! 2. split keys into two nodes (one is new)
   ! 3. insert separation key put into parent, and repeat upwards
! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param value ...
!> \param exists ...
!> \param existing_value ...
!> \param replace ...
! **************************************************************************************************
   SUBROUTINE btree_add(tree, key, value, exists, existing_value, replace)
      TYPE(btree), INTENT(INOUT)                         :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(sp2d), INTENT(IN)                             :: value
      LOGICAL, INTENT(OUT), OPTIONAL                     :: exists
      TYPE(sp2d), INTENT(OUT), OPTIONAL                  :: existing_value
      LOGICAL, INTENT(IN), OPTIONAL                      :: replace

      INTEGER                                            :: ge_pos, position
      TYPE(btree_node), POINTER                          :: node

!

      !
      IF (PRESENT(exists)) THEN
         CALL btree_find_full(tree, key, node, position, ge_pos, short=.TRUE.)
         IF (position .GT. 0) THEN
            exists = .TRUE.
            existing_value = node%values(position)
            IF (PRESENT(replace)) THEN
               IF (replace) THEN
                  node%values(position) = value
               ENDIF
            ENDIF
            RETURN
         ELSE
            exists = .FALSE.
         ENDIF
      ELSE
         CALL btree_find_leaf(tree, key, node, ge_pos)
      ENDIF
      CALL btree_add_into(tree, node, key, value, before=ge_pos)
      IF (PRESENT(exists)) existing_value = value
      tree%b%n = tree%b%n+1
   END SUBROUTINE btree_add

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param node ...
!> \param key ...
!> \param value ...
!> \param before ...
!> \param subtree ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE btree_add_into(tree, node, key, value, before, subtree)
      TYPE(btree), INTENT(INOUT)                         :: tree
      TYPE(btree_node), POINTER                          :: node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(sp2d), INTENT(IN)                             :: value
      INTEGER, INTENT(IN), OPTIONAL                      :: before
      TYPE(btree_node), OPTIONAL, POINTER                :: subtree

      INTEGER                                            :: ge_pos, split_pos
      INTEGER(KIND=keyt)                                 :: upgrade_key
      LOGICAL                                            :: leaf
      TYPE(btree_node), POINTER                          :: new_node
      TYPE(sp2d)                                         :: upgrade_value

!

      !
      ! Root is special
      IF (.NOT. ASSOCIATED(node)) THEN
         CALL btree_new_root(tree, key, value)
         IF (PRESENT(subtree)) THEN
            tree%b%root%subtrees(2)%node => subtree
            subtree%parent => tree%b%root
         ENDIF
         RETURN
      ENDIF
      ! Where the insertion takes place.
      IF (PRESENT(before)) THEN
         ge_pos = before
      ELSE
         CALL btree_node_find_gt_pos(node%keys, key, ge_pos, node%filled)
      ENDIF
      ! Addition is easy if the node has enough space.
      leaf = .NOT. ASSOCIATED(node%subtrees(1)%node)
      IF (node%filled .LT. tree%b%max_fill) THEN
         IF (PRESENT(subtree)) THEN
            CALL btree_simple_insertion(node, key, value, ge_pos, subtree)
         ELSE
            CALL btree_simple_insertion(node, key, value, ge_pos)
         ENDIF
         RETURN
      ELSE
         split_pos = ISHFT(tree%b%max_fill+1, -1)
         ! I assert that split_pos <= SIZE(node%keys)
         CALL btree_new_node(tree, new_node)
         ! The key to be added falls in the left node
         node%filled = split_pos-1
         IF (ge_pos .LE. split_pos) THEN
            IF (ge_pos .EQ. split_pos) THEN
               upgrade_key = key
               upgrade_value = value
            ELSE
               upgrade_key = node%keys(split_pos-1)
               upgrade_value = node%values(split_pos-1)
            ENDIF
            IF (PRESENT(subtree)) THEN
               CALL btree_left_insertion(tree, node, new_node, key, value, &
                                         ge_pos, split_pos, subtree)
               !CALL btree_adopt_subtrees (new_node)
            ELSE
               CALL btree_left_insertion(tree, node, new_node, key, value, &
                                         ge_pos, split_pos)
            ENDIF
            !
         ELSE
            upgrade_key = node%keys(split_pos)
            upgrade_value = node%values(split_pos)
            IF (PRESENT(subtree)) THEN
               CALL btree_right_insertion(tree, node, new_node, key, value, &
                                          ge_pos, split_pos, subtree)
               !CALL btree_adopt_subtrees (new_node)
            ELSE
               CALL btree_right_insertion(tree, node, new_node, key, value, &
                                          ge_pos, split_pos)
            ENDIF
            !
         ENDIF
         !
         new_node%parent => node%parent
         !
         IF (.NOT. leaf) THEN
            CALL btree_adopt_subtrees(new_node)
         ENDIF
         !
         CALL btree_add_into(tree, node%parent, upgrade_key, upgrade_value, &
                             subtree=new_node)
         !
      ENDIF
   END SUBROUTINE btree_add_into

! **************************************************************************************************
!> \brief ...
!> \param node ...
!> \param key ...
!> \param value ...
!> \param before ...
!> \param subtree ...
! **************************************************************************************************
   SUBROUTINE btree_simple_insertion(node, key, value, before, subtree)
      TYPE(btree_node), INTENT(INOUT)                    :: node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(sp2d), INTENT(IN)                             :: value
      INTEGER, INTENT(IN)                                :: before
      TYPE(btree_node), OPTIONAL, POINTER                :: subtree

      !
      ! Shift keys
      node%keys(before+1:node%filled+1) = node%keys(before:node%filled)
      node%keys(before) = key
      ! Shift values
      node%values(before+1:node%filled+1) = node%values(before:node%filled)
      node%values(before) = value
      ! Shift subtree pointers, but only if node is not a leaf ; assume
      ! leaf <=> present(subtree)
      IF (PRESENT(subtree)) THEN
         node%subtrees(before+2:node%filled+2) = &
            node%subtrees(before+1:node%filled+1)
         node%subtrees(before+1)%node => subtree
      ENDIF
      node%filled = node%filled+1
   END SUBROUTINE btree_simple_insertion

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param node ...
!> \param new_node ...
!> \param key ...
!> \param value ...
!> \param before ...
!> \param split_pos ...
!> \param subtree ...
! **************************************************************************************************
   SUBROUTINE btree_left_insertion(tree, node, new_node, key, value, before, split_pos, subtree)
      TYPE(btree), INTENT(IN)                            :: tree
      TYPE(btree_node), INTENT(INOUT)                    :: node, new_node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(sp2d), INTENT(IN)                             :: value
      INTEGER, INTENT(IN)                                :: before, split_pos
      TYPE(btree_node), OPTIONAL, POINTER                :: subtree

      !
      new_node%filled = (tree%b%max_fill)-(split_pos-1)
      new_node%keys(1:new_node%filled) = &
         node%keys(split_pos:tree%b%max_fill)
      new_node%values(1:new_node%filled) = &
         node%values(split_pos:tree%b%max_fill)
      !IF (ASSOCIATED (node%subtrees(1)%node)) THEN
      IF (PRESENT(subtree)) THEN
         IF (before .EQ. split_pos) THEN
            new_node%subtrees(2:new_node%filled+1) = &
               node%subtrees(split_pos+1:tree%b%max_fill+1)
            new_node%subtrees(1)%node => subtree
         ELSE
            new_node%subtrees(1:new_node%filled+1) = &
               node%subtrees(split_pos:tree%b%max_fill+1)
         ENDIF
      ENDIF
      ! Fill node%{keys,values}(1:node%filled), where node%filled
      ! is split_pos-1, but do insert the new value at ge_pos. The
      ! key/value at split_pos is to be inserted into the
      ! parent.
      ! The new tree is added to the right of the new insertion.
      node%keys(before+1:node%filled) = node%keys(before:node%filled-1)
      node%keys(before) = key
      node%values(before+1:node%filled) = node%values(before:node%filled-1)
      node%values(before) = value
      IF (PRESENT(subtree)) THEN
         node%subtrees(before+2:node%filled+1) = &
            node%subtrees(before+1:node%filled)
         node%subtrees(before+1)%node => subtree
      ELSE
         NULLIFY (node%subtrees(before+1)%node)
      ENDIF
   END SUBROUTINE btree_left_insertion

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param node ...
!> \param new_node ...
!> \param key ...
!> \param value ...
!> \param before ...
!> \param split_pos ...
!> \param subtree ...
! **************************************************************************************************
   SUBROUTINE btree_right_insertion(tree, node, new_node, key, value, before, split_pos, subtree)
      TYPE(btree), INTENT(IN)                            :: tree
      TYPE(btree_node), INTENT(INOUT)                    :: node, new_node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(sp2d), INTENT(IN)                             :: value
      INTEGER, INTENT(IN)                                :: before, split_pos
      TYPE(btree_node), OPTIONAL, POINTER                :: subtree

      !
      new_node%filled = (tree%b%max_fill+1)-split_pos
      new_node%keys(1:before-split_pos-1) = &
         node%keys(split_pos+1:before-1)
      new_node%keys(before-split_pos) = key
      new_node%keys(before-split_pos+1:new_node%filled) = &
         node%keys(before:tree%b%max_fill)
      new_node%values(1:before-split_pos-1) = &
         node%values(split_pos+1:before-1)
      new_node%values(before-split_pos) = value
      new_node%values(before-split_pos+1:new_node%filled) = &
         node%values(before:tree%b%max_fill)
      IF (PRESENT(subtree)) THEN
         new_node%subtrees(1:before-split_pos) = &
            node%subtrees(split_pos+1:before)
         new_node%subtrees(before-split_pos+1)%node => subtree
         new_node%subtrees(before-split_pos+2:new_node%filled+1) = &
            node%subtrees(before+1:tree%b%max_fill+1)
      ENDIF
   END SUBROUTINE btree_right_insertion

   ! node is a non-leaf node
! **************************************************************************************************
!> \brief ...
!> \param node ...
! **************************************************************************************************
   SUBROUTINE btree_adopt_subtrees(node)
      TYPE(btree_node), POINTER                          :: node

      INTEGER                                            :: i

      !
      ! Assume that node is not a leaf!
      DO i = 1, node%filled+1
         !IF (ASSOCIATED (node%subtrees(i)%node)) THEN
         !IF (.NOT. ASSOCIATED (node%subtrees(i)%node%parent,&
         ! node)) THEN
         node%subtrees(i)%node%parent => node
         !ENDIF
         !ENDIF
      ENDDO
   END SUBROUTINE btree_adopt_subtrees

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param value ...
! **************************************************************************************************
   SUBROUTINE btree_new_root(tree, key, value)
      TYPE(btree), INTENT(INOUT)                         :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(sp2d), INTENT(IN)                             :: value

      TYPE(btree_node), POINTER                          :: new_root, old_root

      !
      CALL btree_new_node(tree, new_root)
      new_root%filled = 1
      new_root%keys(1) = key
      new_root%values(1) = value
      IF (ASSOCIATED(tree%b%root)) THEN
         old_root => tree%b%root
         old_root%parent => new_root
         new_root%subtrees(1)%node => old_root
         old_root%parent => new_root
      ENDIF
      tree%b%root => new_root
   END SUBROUTINE btree_new_root

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param node ...
! **************************************************************************************************
   SUBROUTINE btree_new_node(tree, node)
      TYPE(btree), INTENT(INOUT)                         :: tree
      TYPE(btree_node), POINTER                          :: node

      INTEGER                                            :: i

      !
      ALLOCATE (node)
      ALLOCATE (node%keys(tree%b%max_fill))
      ALLOCATE (node%values(tree%b%max_fill))
      ALLOCATE (node%subtrees(tree%b%max_fill+1))
      DO i = 1, tree%b%max_fill+1
         NULLIFY (node%subtrees(i)%node)
      ENDDO
      node%filled = 0
      NULLIFY (node%parent)
      tree%b%lastid = tree%b%lastid+1
      node%id = tree%b%lastid
   END SUBROUTINE btree_new_node

! **************************************************************************************************
!> \brief ...
!> \param node ...
! **************************************************************************************************
   SUBROUTINE btree_free_node(node)
      TYPE(btree_node), POINTER                          :: node

      !
      DEALLOCATE (node%keys)
      DEALLOCATE (node%values)
      DEALLOCATE (node%subtrees)
      DEALLOCATE (node)
   END SUBROUTINE btree_free_node

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param value ...
!> \param exists ...
! **************************************************************************************************
   SUBROUTINE btree_find(tree, key, value, exists)
      TYPE(btree), INTENT(IN)                            :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(sp2d), INTENT(OUT)                            :: value
      LOGICAL, INTENT(OUT), OPTIONAL                     :: exists

      INTEGER                                            :: position
      TYPE(btree_node), POINTER                          :: node

!

      !
      CALL btree_find_full(tree, key, node, position, short=.TRUE.)
      IF (PRESENT(exists)) THEN
         exists = position .GT. 0
      ENDIF
      IF (position .GT. 0) THEN
         value = node%values(position)
      ENDIF
   END SUBROUTINE btree_find

! **************************************************************************************************
!> \brief ...
!> \param keys ...
!> \param key ...
!> \param position ...
!> \param filled ...
! **************************************************************************************************
   SUBROUTINE btree_node_find_ge_pos(keys, key, position, filled)
      INTEGER(KIND=keyt), DIMENSION(:)                   :: keys
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER, INTENT(OUT)                               :: position
      INTEGER, INTENT(IN)                                :: filled

      INTEGER                                            :: left, right

      !
      IF (keys(1) .GE. key) THEN
         position = 1
         RETURN
      ENDIF
      IF (keys(filled) .LT. key) THEN
         position = filled+1
         RETURN
      ENDIF
      left = 2
      right = filled
      position = MAX(ISHFT(left+right, -1), left)
      DO WHILE (left .LE. right)
         IF (keys(position) .GE. key .AND. keys(position-1) .LT. key) THEN
            RETURN
         ENDIF
         IF (keys(position) .GE. key) right = MIN(position, right-1)
         IF (keys(position) .LT. key) left = MAX(position, left+1)
         position = MAX(ISHFT(left+right, -1), left)
      ENDDO
   END SUBROUTINE btree_node_find_ge_pos

! **************************************************************************************************
!> \brief ...
!> \param keys ...
!> \param key ...
!> \param position ...
!> \param filled ...
! **************************************************************************************************
   SUBROUTINE btree_node_find_gt_pos(keys, key, position, filled)
      INTEGER(KIND=keyt), DIMENSION(:)                   :: keys
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER, INTENT(OUT)                               :: position
      INTEGER, INTENT(IN)                                :: filled

      INTEGER                                            :: left, right

      !
      IF (keys(1) .GT. key) THEN
         position = 1
         RETURN
      ENDIF
      IF (keys(filled) .LE. key) THEN
         position = filled+1
         RETURN
      ENDIF
      left = 2
      right = filled
      position = MAX(ISHFT(left+right, -1), left)
      DO WHILE (left .LE. right)
         IF (keys(position) .GT. key .AND. keys(position-1) .LE. key) THEN
            RETURN
         ENDIF
         IF (keys(position) .GT. key) right = MIN(position, right-1)
         IF (keys(position) .LE. key) left = MAX(position, left+1)
         position = MAX(ISHFT(left+right, -1), left)
      ENDDO
   END SUBROUTINE btree_node_find_gt_pos
! **************************************************************************************************
!> \brief ...
!> \param keys ...
!> \param key ...
!> \param position ...
!> \param filled ...
!> \param first ...
! **************************************************************************************************
   SUBROUTINE btree_node_find_gte_pos(keys, key, position, filled, first)
      INTEGER(KIND=keyt), DIMENSION(:)                   :: keys
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER, INTENT(OUT)                               :: position
      INTEGER, INTENT(IN)                                :: filled
      INTEGER, INTENT(IN), OPTIONAL                      :: first

      INTEGER                                            :: left, one, right

      !
      one = 1
      IF (PRESENT(FIRST)) one = first
      IF (one .LE. filled) THEN
         IF (keys(one) .GT. key) THEN
            position = one
            RETURN
         ENDIF
      ENDIF
      IF (keys(filled) .LE. key) THEN
         position = filled+1
         RETURN
      ENDIF
      left = one+1
      right = filled
      position = MAX(ISHFT(left+right, -1), left)
      DO WHILE (left .LE. right)
         IF (keys(position) .GT. key .AND. keys(position-1) .LE. key) THEN
            RETURN
         ENDIF
         IF (keys(position) .GT. key) right = MIN(position, right-1)
         IF (keys(position) .LE. key) left = MAX(position, left+1)
         position = MAX(ISHFT(left+right, -1), left)
      ENDDO
   END SUBROUTINE btree_node_find_gte_pos

   ! node is unassociated and position=0 if not found
   ! Precondition: The key is tree or its subtree.
! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param node ...
!> \param position ...
!> \param ge_position ...
!> \param short ...
! **************************************************************************************************
   SUBROUTINE btree_find_full(tree, key, node, position, ge_position, short)
      TYPE(btree), INTENT(IN)                            :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(btree_node), POINTER                          :: node
      INTEGER, INTENT(OUT)                               :: position
      INTEGER, INTENT(OUT), OPTIONAL                     :: ge_position
      LOGICAL, INTENT(IN), OPTIONAL                      :: short

      INTEGER                                            :: gti
      LOGICAL                                            :: stop_short

! Used mark searches

      !
      stop_short = .FALSE.
      IF (PRESENT(short)) stop_short = short
      NULLIFY (node)
      position = 0
      IF (PRESENT(ge_position)) ge_position = 0
      !IF (tree%b%n .EQ. 0) RETURN
      IF (.NOT. ASSOCIATED(tree%b%root)) RETURN
      gti = 1
      ! Try to find the key in the given node. If it's found, then
      ! return the node.
      node => tree%b%root
      descent: DO WHILE (.TRUE.)
         ! Try to find the first element equal to or greater than the
         ! one we're searching for.
         CALL btree_node_find_ge_pos(node%keys, key, position, node%filled)
         ! One of three things is now true about position: it's now
         ! greater than the number of keys (if all keys are smaller), or
         ! it points to the key that is equal to or greater than the one
         ! we are searching for. If it is found and we are just
         ! searching for one equal element (i.e., user search), we can
         ! return.
         IF (stop_short .AND. position .LE. node%filled) THEN
            IF (node%keys(position) .EQ. key) THEN
               IF (PRESENT(ge_position)) ge_position = position
               RETURN
            ENDIF
         ENDIF
         ! If the key is not found, then either return the GE position
         ! if we're in a leaf (case 2 here), otherwise descend into the
         ! subtrees.
         !CALL btree_node_find_gt_pos (node%keys, key, gti, node%filled, position)
         CALL btree_node_find_gte_pos(node%keys, key, gti, node%filled, position)
         IF (ASSOCIATED(node%subtrees(1)%node)) THEN
            node => node%subtrees(gti)%node
         ELSE
            IF (PRESENT(ge_position)) ge_position = gti
            position = 0
            RETURN
         ENDIF
      END DO descent
   END SUBROUTINE btree_find_full

   ! node is unassociated and position=0 if not found
   ! Precondition: The key is tree or its subtree.
! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param node ...
!> \param gti ...
! **************************************************************************************************
   SUBROUTINE btree_find_leaf(tree, key, node, gti)
      TYPE(btree), INTENT(IN)                            :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(btree_node), POINTER                          :: node
      INTEGER, INTENT(OUT)                               :: gti

      !
      NULLIFY (node)
      !IF (tree%b%n .EQ. 0) RETURN
      IF (.NOT. ASSOCIATED(tree%b%root)) RETURN
      gti = 1
      ! Try to find the key in the given node. If it's found, then
      ! return the node.
      node => tree%b%root
      descent: DO WHILE (.TRUE.)
         ! Try to find the first element equal to or greater than the
         ! one we're searching for.
         !CALL btree_node_find_ge_pos (node%keys, key, position, node%filled)
         ! One of three things is now true about position: it's now
         ! greater than the number of keys (if all keys are smaller), or
         ! it points to the key that is equal to or greater than the one
         ! we are searching for. If it is found and we are just
         ! searching for one equal element (i.e., user search), we can
         ! return.
         !
         ! If the key is not found, then either return the GE position
         ! if we're in a leaf (case 2 here), otherwise descend into the
         ! subtrees.
         CALL btree_node_find_gt_pos(node%keys, key, gti, node%filled)
         !CALL btree_node_find_gt2_pos (node%keys, key, i, node%filled)
         !IF (i .NE. gti) WRITE(*,*)'XXXX difference',i,gti
         IF (ASSOCIATED(node%subtrees(1)%node)) THEN
            node => node%subtrees(gti)%node
         ELSE
            RETURN
         ENDIF
      END DO descent
   END SUBROUTINE btree_find_leaf

END MODULE btree_i8_k_sp2d_v
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   B-tree implementation template
!> \author  Urban Borstnik
!> \date    2000-05-12
!> \version 1.0
!> <b>Modification history:</b>
!> - Created
! **************************************************************************************************

MODULE btree_i8_k_dp2d_v

   IMPLICIT NONE
   PRIVATE

   !API
   PUBLIC :: btree, btree_new, btree_add, btree_find, &
             btree_delete, btree_get_entries

   INTEGER, PARAMETER :: keyt = SELECTED_INT_KIND(10)
   INTEGER, PARAMETER :: valt = SELECTED_INT_KIND(5); 
   INTEGER, PARAMETER :: dp = KIND(0.0d0)

   TYPE dp2d
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: p
      LOGICAL :: tr
   END TYPE dp2d

   PUBLIC :: dp2d

   TYPE btree_node
      INTEGER id
      INTEGER :: filled
      INTEGER(KIND=keyt), DIMENSION(:), POINTER :: keys
      TYPE(dp2d), DIMENSION(:), POINTER :: values
      TYPE(btree_node_p), DIMENSION(:), POINTER :: subtrees
      TYPE(btree_node), POINTER :: parent
   END TYPE btree_node

   TYPE btree_node_p
      TYPE(btree_node), POINTER :: node
   END TYPE btree_node_p

   TYPE btree_node_structure
      INTEGER :: min_fill, max_fill
      INTEGER :: n
      INTEGER :: lastid
      INTEGER :: refcount
      TYPE(btree_node), POINTER :: root
   END TYPE btree_node_structure

   TYPE btree
      TYPE(btree_node_structure) :: b
   END TYPE btree

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE btree_new(tree, order)
      TYPE(btree), INTENT(OUT)                           :: tree
      INTEGER, INTENT(IN), OPTIONAL                      :: order

      INTEGER                                            :: maxs, mins

      !
      IF (PRESENT(order)) THEN
         maxs = order-1
      ELSE
         maxs = 15
      ENDIF
      mins = ISHFT(maxs, -1)
      IF (mins*2 .GT. maxs) maxs = 2*maxs
      IF (mins .LT. 1) mins = 1
      IF (maxs .LT. 3) maxs = 3
      tree%b%min_fill = mins
      tree%b%max_fill = maxs
      tree%b%refcount = 1
      tree%b%n = 0
      NULLIFY (tree%b%root)
      tree%b%lastid = 0
   END SUBROUTINE btree_new

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \retval num_entries ...
! **************************************************************************************************
   FUNCTION btree_get_entries(tree) RESULT(num_entries)
      TYPE(btree), INTENT(INOUT)                         :: tree
      INTEGER                                            :: num_entries

      num_entries = tree%b%n
   END FUNCTION btree_get_entries

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param keys ...
!> \param values ...
! **************************************************************************************************
   SUBROUTINE btree_delete(tree, keys, values)
      TYPE(btree), INTENT(INOUT)                         :: tree
      INTEGER(KIND=keyt), DIMENSION(:), INTENT(OUT), &
         OPTIONAL                                        :: keys
      TYPE(dp2d), DIMENSION(:), INTENT(OUT), OPTIONAL    :: values

      INTEGER                                            :: pos

      !
      IF (ASSOCIATED(tree%b%root)) THEN
         pos = 0
         IF (PRESENT(keys) .AND. PRESENT(values)) THEN
            pos = 1
            CALL btree_delete_node(tree%b%root, pos, keys, values)
         ELSE
            CALL btree_delete_node(tree%b%root)
         ENDIF
      ENDIF
      NULLIFY (tree%b%root)
   END SUBROUTINE btree_delete

! **************************************************************************************************
!> \brief ...
!> \param node ...
!> \param pos ...
!> \param keys ...
!> \param values ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE btree_delete_node(node, pos, keys, values)
      TYPE(btree_node), POINTER                          :: node
      INTEGER, INTENT(INOUT), OPTIONAL                   :: pos
      INTEGER(KIND=keyt), DIMENSION(:), INTENT(INOUT), &
         OPTIONAL                                        :: keys
      TYPE(dp2d), DIMENSION(:), INTENT(INOUT), OPTIONAL  :: values

      INTEGER                                            :: i

!

      !
      IF (node%filled .GT. 0 .AND. ASSOCIATED(node%subtrees(1)%node)) THEN
         DO i = 1, node%filled+1
            IF (PRESENT(pos)) THEN
               CALL btree_delete_node(node%subtrees(i)%node, pos, keys, values)
            ELSE
               CALL btree_delete_node(node%subtrees(i)%node)
            ENDIF
            IF (PRESENT(pos) .AND. i .LE. node%filled) THEN
               keys(pos) = node%keys(i)
               values(pos) = node%values(i)
               pos = pos+1
            ENDIF
         ENDDO
      ELSEIF (PRESENT(pos) .AND. node%filled .GT. 0) THEN
         keys(pos:pos+node%filled-1) = node%keys(1:node%filled)
         values(pos:pos+node%filled-1) = node%values(1:node%filled)
         pos = pos+node%filled
      ENDIF
      CALL btree_free_node(node)
   END SUBROUTINE btree_delete_node

   ! Find the key
   ! IF node still has space, insert & update the node
   ! else
   ! 1. select median
   ! 2. split keys into two nodes (one is new)
   ! 3. insert separation key put into parent, and repeat upwards
! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param value ...
!> \param exists ...
!> \param existing_value ...
!> \param replace ...
! **************************************************************************************************
   SUBROUTINE btree_add(tree, key, value, exists, existing_value, replace)
      TYPE(btree), INTENT(INOUT)                         :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(dp2d), INTENT(IN)                             :: value
      LOGICAL, INTENT(OUT), OPTIONAL                     :: exists
      TYPE(dp2d), INTENT(OUT), OPTIONAL                  :: existing_value
      LOGICAL, INTENT(IN), OPTIONAL                      :: replace

      INTEGER                                            :: ge_pos, position
      TYPE(btree_node), POINTER                          :: node

!

      !
      IF (PRESENT(exists)) THEN
         CALL btree_find_full(tree, key, node, position, ge_pos, short=.TRUE.)
         IF (position .GT. 0) THEN
            exists = .TRUE.
            existing_value = node%values(position)
            IF (PRESENT(replace)) THEN
               IF (replace) THEN
                  node%values(position) = value
               ENDIF
            ENDIF
            RETURN
         ELSE
            exists = .FALSE.
         ENDIF
      ELSE
         CALL btree_find_leaf(tree, key, node, ge_pos)
      ENDIF
      CALL btree_add_into(tree, node, key, value, before=ge_pos)
      IF (PRESENT(exists)) existing_value = value
      tree%b%n = tree%b%n+1
   END SUBROUTINE btree_add

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param node ...
!> \param key ...
!> \param value ...
!> \param before ...
!> \param subtree ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE btree_add_into(tree, node, key, value, before, subtree)
      TYPE(btree), INTENT(INOUT)                         :: tree
      TYPE(btree_node), POINTER                          :: node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(dp2d), INTENT(IN)                             :: value
      INTEGER, INTENT(IN), OPTIONAL                      :: before
      TYPE(btree_node), OPTIONAL, POINTER                :: subtree

      INTEGER                                            :: ge_pos, split_pos
      INTEGER(KIND=keyt)                                 :: upgrade_key
      LOGICAL                                            :: leaf
      TYPE(btree_node), POINTER                          :: new_node
      TYPE(dp2d)                                         :: upgrade_value

!

      !
      ! Root is special
      IF (.NOT. ASSOCIATED(node)) THEN
         CALL btree_new_root(tree, key, value)
         IF (PRESENT(subtree)) THEN
            tree%b%root%subtrees(2)%node => subtree
            subtree%parent => tree%b%root
         ENDIF
         RETURN
      ENDIF
      ! Where the insertion takes place.
      IF (PRESENT(before)) THEN
         ge_pos = before
      ELSE
         CALL btree_node_find_gt_pos(node%keys, key, ge_pos, node%filled)
      ENDIF
      ! Addition is easy if the node has enough space.
      leaf = .NOT. ASSOCIATED(node%subtrees(1)%node)
      IF (node%filled .LT. tree%b%max_fill) THEN
         IF (PRESENT(subtree)) THEN
            CALL btree_simple_insertion(node, key, value, ge_pos, subtree)
         ELSE
            CALL btree_simple_insertion(node, key, value, ge_pos)
         ENDIF
         RETURN
      ELSE
         split_pos = ISHFT(tree%b%max_fill+1, -1)
         ! I assert that split_pos <= SIZE(node%keys)
         CALL btree_new_node(tree, new_node)
         ! The key to be added falls in the left node
         node%filled = split_pos-1
         IF (ge_pos .LE. split_pos) THEN
            IF (ge_pos .EQ. split_pos) THEN
               upgrade_key = key
               upgrade_value = value
            ELSE
               upgrade_key = node%keys(split_pos-1)
               upgrade_value = node%values(split_pos-1)
            ENDIF
            IF (PRESENT(subtree)) THEN
               CALL btree_left_insertion(tree, node, new_node, key, value, &
                                         ge_pos, split_pos, subtree)
               !CALL btree_adopt_subtrees (new_node)
            ELSE
               CALL btree_left_insertion(tree, node, new_node, key, value, &
                                         ge_pos, split_pos)
            ENDIF
            !
         ELSE
            upgrade_key = node%keys(split_pos)
            upgrade_value = node%values(split_pos)
            IF (PRESENT(subtree)) THEN
               CALL btree_right_insertion(tree, node, new_node, key, value, &
                                          ge_pos, split_pos, subtree)
               !CALL btree_adopt_subtrees (new_node)
            ELSE
               CALL btree_right_insertion(tree, node, new_node, key, value, &
                                          ge_pos, split_pos)
            ENDIF
            !
         ENDIF
         !
         new_node%parent => node%parent
         !
         IF (.NOT. leaf) THEN
            CALL btree_adopt_subtrees(new_node)
         ENDIF
         !
         CALL btree_add_into(tree, node%parent, upgrade_key, upgrade_value, &
                             subtree=new_node)
         !
      ENDIF
   END SUBROUTINE btree_add_into

! **************************************************************************************************
!> \brief ...
!> \param node ...
!> \param key ...
!> \param value ...
!> \param before ...
!> \param subtree ...
! **************************************************************************************************
   SUBROUTINE btree_simple_insertion(node, key, value, before, subtree)
      TYPE(btree_node), INTENT(INOUT)                    :: node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(dp2d), INTENT(IN)                             :: value
      INTEGER, INTENT(IN)                                :: before
      TYPE(btree_node), OPTIONAL, POINTER                :: subtree

      !
      ! Shift keys
      node%keys(before+1:node%filled+1) = node%keys(before:node%filled)
      node%keys(before) = key
      ! Shift values
      node%values(before+1:node%filled+1) = node%values(before:node%filled)
      node%values(before) = value
      ! Shift subtree pointers, but only if node is not a leaf ; assume
      ! leaf <=> present(subtree)
      IF (PRESENT(subtree)) THEN
         node%subtrees(before+2:node%filled+2) = &
            node%subtrees(before+1:node%filled+1)
         node%subtrees(before+1)%node => subtree
      ENDIF
      node%filled = node%filled+1
   END SUBROUTINE btree_simple_insertion

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param node ...
!> \param new_node ...
!> \param key ...
!> \param value ...
!> \param before ...
!> \param split_pos ...
!> \param subtree ...
! **************************************************************************************************
   SUBROUTINE btree_left_insertion(tree, node, new_node, key, value, before, split_pos, subtree)
      TYPE(btree), INTENT(IN)                            :: tree
      TYPE(btree_node), INTENT(INOUT)                    :: node, new_node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(dp2d), INTENT(IN)                             :: value
      INTEGER, INTENT(IN)                                :: before, split_pos
      TYPE(btree_node), OPTIONAL, POINTER                :: subtree

      !
      new_node%filled = (tree%b%max_fill)-(split_pos-1)
      new_node%keys(1:new_node%filled) = &
         node%keys(split_pos:tree%b%max_fill)
      new_node%values(1:new_node%filled) = &
         node%values(split_pos:tree%b%max_fill)
      !IF (ASSOCIATED (node%subtrees(1)%node)) THEN
      IF (PRESENT(subtree)) THEN
         IF (before .EQ. split_pos) THEN
            new_node%subtrees(2:new_node%filled+1) = &
               node%subtrees(split_pos+1:tree%b%max_fill+1)
            new_node%subtrees(1)%node => subtree
         ELSE
            new_node%subtrees(1:new_node%filled+1) = &
               node%subtrees(split_pos:tree%b%max_fill+1)
         ENDIF
      ENDIF
      ! Fill node%{keys,values}(1:node%filled), where node%filled
      ! is split_pos-1, but do insert the new value at ge_pos. The
      ! key/value at split_pos is to be inserted into the
      ! parent.
      ! The new tree is added to the right of the new insertion.
      node%keys(before+1:node%filled) = node%keys(before:node%filled-1)
      node%keys(before) = key
      node%values(before+1:node%filled) = node%values(before:node%filled-1)
      node%values(before) = value
      IF (PRESENT(subtree)) THEN
         node%subtrees(before+2:node%filled+1) = &
            node%subtrees(before+1:node%filled)
         node%subtrees(before+1)%node => subtree
      ELSE
         NULLIFY (node%subtrees(before+1)%node)
      ENDIF
   END SUBROUTINE btree_left_insertion

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param node ...
!> \param new_node ...
!> \param key ...
!> \param value ...
!> \param before ...
!> \param split_pos ...
!> \param subtree ...
! **************************************************************************************************
   SUBROUTINE btree_right_insertion(tree, node, new_node, key, value, before, split_pos, subtree)
      TYPE(btree), INTENT(IN)                            :: tree
      TYPE(btree_node), INTENT(INOUT)                    :: node, new_node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(dp2d), INTENT(IN)                             :: value
      INTEGER, INTENT(IN)                                :: before, split_pos
      TYPE(btree_node), OPTIONAL, POINTER                :: subtree

      !
      new_node%filled = (tree%b%max_fill+1)-split_pos
      new_node%keys(1:before-split_pos-1) = &
         node%keys(split_pos+1:before-1)
      new_node%keys(before-split_pos) = key
      new_node%keys(before-split_pos+1:new_node%filled) = &
         node%keys(before:tree%b%max_fill)
      new_node%values(1:before-split_pos-1) = &
         node%values(split_pos+1:before-1)
      new_node%values(before-split_pos) = value
      new_node%values(before-split_pos+1:new_node%filled) = &
         node%values(before:tree%b%max_fill)
      IF (PRESENT(subtree)) THEN
         new_node%subtrees(1:before-split_pos) = &
            node%subtrees(split_pos+1:before)
         new_node%subtrees(before-split_pos+1)%node => subtree
         new_node%subtrees(before-split_pos+2:new_node%filled+1) = &
            node%subtrees(before+1:tree%b%max_fill+1)
      ENDIF
   END SUBROUTINE btree_right_insertion

   ! node is a non-leaf node
! **************************************************************************************************
!> \brief ...
!> \param node ...
! **************************************************************************************************
   SUBROUTINE btree_adopt_subtrees(node)
      TYPE(btree_node), POINTER                          :: node

      INTEGER                                            :: i

      !
      ! Assume that node is not a leaf!
      DO i = 1, node%filled+1
         !IF (ASSOCIATED (node%subtrees(i)%node)) THEN
         !IF (.NOT. ASSOCIATED (node%subtrees(i)%node%parent,&
         ! node)) THEN
         node%subtrees(i)%node%parent => node
         !ENDIF
         !ENDIF
      ENDDO
   END SUBROUTINE btree_adopt_subtrees

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param value ...
! **************************************************************************************************
   SUBROUTINE btree_new_root(tree, key, value)
      TYPE(btree), INTENT(INOUT)                         :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(dp2d), INTENT(IN)                             :: value

      TYPE(btree_node), POINTER                          :: new_root, old_root

      !
      CALL btree_new_node(tree, new_root)
      new_root%filled = 1
      new_root%keys(1) = key
      new_root%values(1) = value
      IF (ASSOCIATED(tree%b%root)) THEN
         old_root => tree%b%root
         old_root%parent => new_root
         new_root%subtrees(1)%node => old_root
         old_root%parent => new_root
      ENDIF
      tree%b%root => new_root
   END SUBROUTINE btree_new_root

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param node ...
! **************************************************************************************************
   SUBROUTINE btree_new_node(tree, node)
      TYPE(btree), INTENT(INOUT)                         :: tree
      TYPE(btree_node), POINTER                          :: node

      INTEGER                                            :: i

      !
      ALLOCATE (node)
      ALLOCATE (node%keys(tree%b%max_fill))
      ALLOCATE (node%values(tree%b%max_fill))
      ALLOCATE (node%subtrees(tree%b%max_fill+1))
      DO i = 1, tree%b%max_fill+1
         NULLIFY (node%subtrees(i)%node)
      ENDDO
      node%filled = 0
      NULLIFY (node%parent)
      tree%b%lastid = tree%b%lastid+1
      node%id = tree%b%lastid
   END SUBROUTINE btree_new_node

! **************************************************************************************************
!> \brief ...
!> \param node ...
! **************************************************************************************************
   SUBROUTINE btree_free_node(node)
      TYPE(btree_node), POINTER                          :: node

      !
      DEALLOCATE (node%keys)
      DEALLOCATE (node%values)
      DEALLOCATE (node%subtrees)
      DEALLOCATE (node)
   END SUBROUTINE btree_free_node

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param value ...
!> \param exists ...
! **************************************************************************************************
   SUBROUTINE btree_find(tree, key, value, exists)
      TYPE(btree), INTENT(IN)                            :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(dp2d), INTENT(OUT)                            :: value
      LOGICAL, INTENT(OUT), OPTIONAL                     :: exists

      INTEGER                                            :: position
      TYPE(btree_node), POINTER                          :: node

!

      !
      CALL btree_find_full(tree, key, node, position, short=.TRUE.)
      IF (PRESENT(exists)) THEN
         exists = position .GT. 0
      ENDIF
      IF (position .GT. 0) THEN
         value = node%values(position)
      ENDIF
   END SUBROUTINE btree_find

! **************************************************************************************************
!> \brief ...
!> \param keys ...
!> \param key ...
!> \param position ...
!> \param filled ...
! **************************************************************************************************
   SUBROUTINE btree_node_find_ge_pos(keys, key, position, filled)
      INTEGER(KIND=keyt), DIMENSION(:)                   :: keys
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER, INTENT(OUT)                               :: position
      INTEGER, INTENT(IN)                                :: filled

      INTEGER                                            :: left, right

      !
      IF (keys(1) .GE. key) THEN
         position = 1
         RETURN
      ENDIF
      IF (keys(filled) .LT. key) THEN
         position = filled+1
         RETURN
      ENDIF
      left = 2
      right = filled
      position = MAX(ISHFT(left+right, -1), left)
      DO WHILE (left .LE. right)
         IF (keys(position) .GE. key .AND. keys(position-1) .LT. key) THEN
            RETURN
         ENDIF
         IF (keys(position) .GE. key) right = MIN(position, right-1)
         IF (keys(position) .LT. key) left = MAX(position, left+1)
         position = MAX(ISHFT(left+right, -1), left)
      ENDDO
   END SUBROUTINE btree_node_find_ge_pos

! **************************************************************************************************
!> \brief ...
!> \param keys ...
!> \param key ...
!> \param position ...
!> \param filled ...
! **************************************************************************************************
   SUBROUTINE btree_node_find_gt_pos(keys, key, position, filled)
      INTEGER(KIND=keyt), DIMENSION(:)                   :: keys
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER, INTENT(OUT)                               :: position
      INTEGER, INTENT(IN)                                :: filled

      INTEGER                                            :: left, right

      !
      IF (keys(1) .GT. key) THEN
         position = 1
         RETURN
      ENDIF
      IF (keys(filled) .LE. key) THEN
         position = filled+1
         RETURN
      ENDIF
      left = 2
      right = filled
      position = MAX(ISHFT(left+right, -1), left)
      DO WHILE (left .LE. right)
         IF (keys(position) .GT. key .AND. keys(position-1) .LE. key) THEN
            RETURN
         ENDIF
         IF (keys(position) .GT. key) right = MIN(position, right-1)
         IF (keys(position) .LE. key) left = MAX(position, left+1)
         position = MAX(ISHFT(left+right, -1), left)
      ENDDO
   END SUBROUTINE btree_node_find_gt_pos
! **************************************************************************************************
!> \brief ...
!> \param keys ...
!> \param key ...
!> \param position ...
!> \param filled ...
!> \param first ...
! **************************************************************************************************
   SUBROUTINE btree_node_find_gte_pos(keys, key, position, filled, first)
      INTEGER(KIND=keyt), DIMENSION(:)                   :: keys
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER, INTENT(OUT)                               :: position
      INTEGER, INTENT(IN)                                :: filled
      INTEGER, INTENT(IN), OPTIONAL                      :: first

      INTEGER                                            :: left, one, right

      !
      one = 1
      IF (PRESENT(FIRST)) one = first
      IF (one .LE. filled) THEN
         IF (keys(one) .GT. key) THEN
            position = one
            RETURN
         ENDIF
      ENDIF
      IF (keys(filled) .LE. key) THEN
         position = filled+1
         RETURN
      ENDIF
      left = one+1
      right = filled
      position = MAX(ISHFT(left+right, -1), left)
      DO WHILE (left .LE. right)
         IF (keys(position) .GT. key .AND. keys(position-1) .LE. key) THEN
            RETURN
         ENDIF
         IF (keys(position) .GT. key) right = MIN(position, right-1)
         IF (keys(position) .LE. key) left = MAX(position, left+1)
         position = MAX(ISHFT(left+right, -1), left)
      ENDDO
   END SUBROUTINE btree_node_find_gte_pos

   ! node is unassociated and position=0 if not found
   ! Precondition: The key is tree or its subtree.
! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param node ...
!> \param position ...
!> \param ge_position ...
!> \param short ...
! **************************************************************************************************
   SUBROUTINE btree_find_full(tree, key, node, position, ge_position, short)
      TYPE(btree), INTENT(IN)                            :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(btree_node), POINTER                          :: node
      INTEGER, INTENT(OUT)                               :: position
      INTEGER, INTENT(OUT), OPTIONAL                     :: ge_position
      LOGICAL, INTENT(IN), OPTIONAL                      :: short

      INTEGER                                            :: gti
      LOGICAL                                            :: stop_short

! Used mark searches

      !
      stop_short = .FALSE.
      IF (PRESENT(short)) stop_short = short
      NULLIFY (node)
      position = 0
      IF (PRESENT(ge_position)) ge_position = 0
      !IF (tree%b%n .EQ. 0) RETURN
      IF (.NOT. ASSOCIATED(tree%b%root)) RETURN
      gti = 1
      ! Try to find the key in the given node. If it's found, then
      ! return the node.
      node => tree%b%root
      descent: DO WHILE (.TRUE.)
         ! Try to find the first element equal to or greater than the
         ! one we're searching for.
         CALL btree_node_find_ge_pos(node%keys, key, position, node%filled)
         ! One of three things is now true about position: it's now
         ! greater than the number of keys (if all keys are smaller), or
         ! it points to the key that is equal to or greater than the one
         ! we are searching for. If it is found and we are just
         ! searching for one equal element (i.e., user search), we can
         ! return.
         IF (stop_short .AND. position .LE. node%filled) THEN
            IF (node%keys(position) .EQ. key) THEN
               IF (PRESENT(ge_position)) ge_position = position
               RETURN
            ENDIF
         ENDIF
         ! If the key is not found, then either return the GE position
         ! if we're in a leaf (case 2 here), otherwise descend into the
         ! subtrees.
         !CALL btree_node_find_gt_pos (node%keys, key, gti, node%filled, position)
         CALL btree_node_find_gte_pos(node%keys, key, gti, node%filled, position)
         IF (ASSOCIATED(node%subtrees(1)%node)) THEN
            node => node%subtrees(gti)%node
         ELSE
            IF (PRESENT(ge_position)) ge_position = gti
            position = 0
            RETURN
         ENDIF
      END DO descent
   END SUBROUTINE btree_find_full

   ! node is unassociated and position=0 if not found
   ! Precondition: The key is tree or its subtree.
! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param node ...
!> \param gti ...
! **************************************************************************************************
   SUBROUTINE btree_find_leaf(tree, key, node, gti)
      TYPE(btree), INTENT(IN)                            :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(btree_node), POINTER                          :: node
      INTEGER, INTENT(OUT)                               :: gti

      !
      NULLIFY (node)
      !IF (tree%b%n .EQ. 0) RETURN
      IF (.NOT. ASSOCIATED(tree%b%root)) RETURN
      gti = 1
      ! Try to find the key in the given node. If it's found, then
      ! return the node.
      node => tree%b%root
      descent: DO WHILE (.TRUE.)
         ! Try to find the first element equal to or greater than the
         ! one we're searching for.
         !CALL btree_node_find_ge_pos (node%keys, key, position, node%filled)
         ! One of three things is now true about position: it's now
         ! greater than the number of keys (if all keys are smaller), or
         ! it points to the key that is equal to or greater than the one
         ! we are searching for. If it is found and we are just
         ! searching for one equal element (i.e., user search), we can
         ! return.
         !
         ! If the key is not found, then either return the GE position
         ! if we're in a leaf (case 2 here), otherwise descend into the
         ! subtrees.
         CALL btree_node_find_gt_pos(node%keys, key, gti, node%filled)
         !CALL btree_node_find_gt2_pos (node%keys, key, i, node%filled)
         !IF (i .NE. gti) WRITE(*,*)'XXXX difference',i,gti
         IF (ASSOCIATED(node%subtrees(1)%node)) THEN
            node => node%subtrees(gti)%node
         ELSE
            RETURN
         ENDIF
      END DO descent
   END SUBROUTINE btree_find_leaf

END MODULE btree_i8_k_dp2d_v
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   B-tree implementation template
!> \author  Urban Borstnik
!> \date    2000-05-12
!> \version 1.0
!> <b>Modification history:</b>
!> - Created
! **************************************************************************************************

MODULE btree_i8_k_cp2d_v

   IMPLICIT NONE
   PRIVATE

   !API
   PUBLIC :: btree, btree_new, btree_add, btree_find, &
             btree_delete, btree_get_entries

   INTEGER, PARAMETER :: keyt = SELECTED_INT_KIND(10)
   INTEGER, PARAMETER :: valt = SELECTED_INT_KIND(5); 
   INTEGER, PARAMETER :: sp = KIND(0.0)

   TYPE cp2d
      COMPLEX(KIND=sp), DIMENSION(:, :), POINTER :: p
      LOGICAL :: tr
   END TYPE cp2d

   PUBLIC :: cp2d

   TYPE btree_node
      INTEGER id
      INTEGER :: filled
      INTEGER(KIND=keyt), DIMENSION(:), POINTER :: keys
      TYPE(cp2d), DIMENSION(:), POINTER :: values
      TYPE(btree_node_p), DIMENSION(:), POINTER :: subtrees
      TYPE(btree_node), POINTER :: parent
   END TYPE btree_node

   TYPE btree_node_p
      TYPE(btree_node), POINTER :: node
   END TYPE btree_node_p

   TYPE btree_node_structure
      INTEGER :: min_fill, max_fill
      INTEGER :: n
      INTEGER :: lastid
      INTEGER :: refcount
      TYPE(btree_node), POINTER :: root
   END TYPE btree_node_structure

   TYPE btree
      TYPE(btree_node_structure) :: b
   END TYPE btree

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE btree_new(tree, order)
      TYPE(btree), INTENT(OUT)                           :: tree
      INTEGER, INTENT(IN), OPTIONAL                      :: order

      INTEGER                                            :: maxs, mins

      !
      IF (PRESENT(order)) THEN
         maxs = order-1
      ELSE
         maxs = 15
      ENDIF
      mins = ISHFT(maxs, -1)
      IF (mins*2 .GT. maxs) maxs = 2*maxs
      IF (mins .LT. 1) mins = 1
      IF (maxs .LT. 3) maxs = 3
      tree%b%min_fill = mins
      tree%b%max_fill = maxs
      tree%b%refcount = 1
      tree%b%n = 0
      NULLIFY (tree%b%root)
      tree%b%lastid = 0
   END SUBROUTINE btree_new

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \retval num_entries ...
! **************************************************************************************************
   FUNCTION btree_get_entries(tree) RESULT(num_entries)
      TYPE(btree), INTENT(INOUT)                         :: tree
      INTEGER                                            :: num_entries

      num_entries = tree%b%n
   END FUNCTION btree_get_entries

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param keys ...
!> \param values ...
! **************************************************************************************************
   SUBROUTINE btree_delete(tree, keys, values)
      TYPE(btree), INTENT(INOUT)                         :: tree
      INTEGER(KIND=keyt), DIMENSION(:), INTENT(OUT), &
         OPTIONAL                                        :: keys
      TYPE(cp2d), DIMENSION(:), INTENT(OUT), OPTIONAL    :: values

      INTEGER                                            :: pos

      !
      IF (ASSOCIATED(tree%b%root)) THEN
         pos = 0
         IF (PRESENT(keys) .AND. PRESENT(values)) THEN
            pos = 1
            CALL btree_delete_node(tree%b%root, pos, keys, values)
         ELSE
            CALL btree_delete_node(tree%b%root)
         ENDIF
      ENDIF
      NULLIFY (tree%b%root)
   END SUBROUTINE btree_delete

! **************************************************************************************************
!> \brief ...
!> \param node ...
!> \param pos ...
!> \param keys ...
!> \param values ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE btree_delete_node(node, pos, keys, values)
      TYPE(btree_node), POINTER                          :: node
      INTEGER, INTENT(INOUT), OPTIONAL                   :: pos
      INTEGER(KIND=keyt), DIMENSION(:), INTENT(INOUT), &
         OPTIONAL                                        :: keys
      TYPE(cp2d), DIMENSION(:), INTENT(INOUT), OPTIONAL  :: values

      INTEGER                                            :: i

!

      !
      IF (node%filled .GT. 0 .AND. ASSOCIATED(node%subtrees(1)%node)) THEN
         DO i = 1, node%filled+1
            IF (PRESENT(pos)) THEN
               CALL btree_delete_node(node%subtrees(i)%node, pos, keys, values)
            ELSE
               CALL btree_delete_node(node%subtrees(i)%node)
            ENDIF
            IF (PRESENT(pos) .AND. i .LE. node%filled) THEN
               keys(pos) = node%keys(i)
               values(pos) = node%values(i)
               pos = pos+1
            ENDIF
         ENDDO
      ELSEIF (PRESENT(pos) .AND. node%filled .GT. 0) THEN
         keys(pos:pos+node%filled-1) = node%keys(1:node%filled)
         values(pos:pos+node%filled-1) = node%values(1:node%filled)
         pos = pos+node%filled
      ENDIF
      CALL btree_free_node(node)
   END SUBROUTINE btree_delete_node

   ! Find the key
   ! IF node still has space, insert & update the node
   ! else
   ! 1. select median
   ! 2. split keys into two nodes (one is new)
   ! 3. insert separation key put into parent, and repeat upwards
! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param value ...
!> \param exists ...
!> \param existing_value ...
!> \param replace ...
! **************************************************************************************************
   SUBROUTINE btree_add(tree, key, value, exists, existing_value, replace)
      TYPE(btree), INTENT(INOUT)                         :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(cp2d), INTENT(IN)                             :: value
      LOGICAL, INTENT(OUT), OPTIONAL                     :: exists
      TYPE(cp2d), INTENT(OUT), OPTIONAL                  :: existing_value
      LOGICAL, INTENT(IN), OPTIONAL                      :: replace

      INTEGER                                            :: ge_pos, position
      TYPE(btree_node), POINTER                          :: node

!

      !
      IF (PRESENT(exists)) THEN
         CALL btree_find_full(tree, key, node, position, ge_pos, short=.TRUE.)
         IF (position .GT. 0) THEN
            exists = .TRUE.
            existing_value = node%values(position)
            IF (PRESENT(replace)) THEN
               IF (replace) THEN
                  node%values(position) = value
               ENDIF
            ENDIF
            RETURN
         ELSE
            exists = .FALSE.
         ENDIF
      ELSE
         CALL btree_find_leaf(tree, key, node, ge_pos)
      ENDIF
      CALL btree_add_into(tree, node, key, value, before=ge_pos)
      IF (PRESENT(exists)) existing_value = value
      tree%b%n = tree%b%n+1
   END SUBROUTINE btree_add

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param node ...
!> \param key ...
!> \param value ...
!> \param before ...
!> \param subtree ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE btree_add_into(tree, node, key, value, before, subtree)
      TYPE(btree), INTENT(INOUT)                         :: tree
      TYPE(btree_node), POINTER                          :: node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(cp2d), INTENT(IN)                             :: value
      INTEGER, INTENT(IN), OPTIONAL                      :: before
      TYPE(btree_node), OPTIONAL, POINTER                :: subtree

      INTEGER                                            :: ge_pos, split_pos
      INTEGER(KIND=keyt)                                 :: upgrade_key
      LOGICAL                                            :: leaf
      TYPE(btree_node), POINTER                          :: new_node
      TYPE(cp2d)                                         :: upgrade_value

!

      !
      ! Root is special
      IF (.NOT. ASSOCIATED(node)) THEN
         CALL btree_new_root(tree, key, value)
         IF (PRESENT(subtree)) THEN
            tree%b%root%subtrees(2)%node => subtree
            subtree%parent => tree%b%root
         ENDIF
         RETURN
      ENDIF
      ! Where the insertion takes place.
      IF (PRESENT(before)) THEN
         ge_pos = before
      ELSE
         CALL btree_node_find_gt_pos(node%keys, key, ge_pos, node%filled)
      ENDIF
      ! Addition is easy if the node has enough space.
      leaf = .NOT. ASSOCIATED(node%subtrees(1)%node)
      IF (node%filled .LT. tree%b%max_fill) THEN
         IF (PRESENT(subtree)) THEN
            CALL btree_simple_insertion(node, key, value, ge_pos, subtree)
         ELSE
            CALL btree_simple_insertion(node, key, value, ge_pos)
         ENDIF
         RETURN
      ELSE
         split_pos = ISHFT(tree%b%max_fill+1, -1)
         ! I assert that split_pos <= SIZE(node%keys)
         CALL btree_new_node(tree, new_node)
         ! The key to be added falls in the left node
         node%filled = split_pos-1
         IF (ge_pos .LE. split_pos) THEN
            IF (ge_pos .EQ. split_pos) THEN
               upgrade_key = key
               upgrade_value = value
            ELSE
               upgrade_key = node%keys(split_pos-1)
               upgrade_value = node%values(split_pos-1)
            ENDIF
            IF (PRESENT(subtree)) THEN
               CALL btree_left_insertion(tree, node, new_node, key, value, &
                                         ge_pos, split_pos, subtree)
               !CALL btree_adopt_subtrees (new_node)
            ELSE
               CALL btree_left_insertion(tree, node, new_node, key, value, &
                                         ge_pos, split_pos)
            ENDIF
            !
         ELSE
            upgrade_key = node%keys(split_pos)
            upgrade_value = node%values(split_pos)
            IF (PRESENT(subtree)) THEN
               CALL btree_right_insertion(tree, node, new_node, key, value, &
                                          ge_pos, split_pos, subtree)
               !CALL btree_adopt_subtrees (new_node)
            ELSE
               CALL btree_right_insertion(tree, node, new_node, key, value, &
                                          ge_pos, split_pos)
            ENDIF
            !
         ENDIF
         !
         new_node%parent => node%parent
         !
         IF (.NOT. leaf) THEN
            CALL btree_adopt_subtrees(new_node)
         ENDIF
         !
         CALL btree_add_into(tree, node%parent, upgrade_key, upgrade_value, &
                             subtree=new_node)
         !
      ENDIF
   END SUBROUTINE btree_add_into

! **************************************************************************************************
!> \brief ...
!> \param node ...
!> \param key ...
!> \param value ...
!> \param before ...
!> \param subtree ...
! **************************************************************************************************
   SUBROUTINE btree_simple_insertion(node, key, value, before, subtree)
      TYPE(btree_node), INTENT(INOUT)                    :: node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(cp2d), INTENT(IN)                             :: value
      INTEGER, INTENT(IN)                                :: before
      TYPE(btree_node), OPTIONAL, POINTER                :: subtree

      !
      ! Shift keys
      node%keys(before+1:node%filled+1) = node%keys(before:node%filled)
      node%keys(before) = key
      ! Shift values
      node%values(before+1:node%filled+1) = node%values(before:node%filled)
      node%values(before) = value
      ! Shift subtree pointers, but only if node is not a leaf ; assume
      ! leaf <=> present(subtree)
      IF (PRESENT(subtree)) THEN
         node%subtrees(before+2:node%filled+2) = &
            node%subtrees(before+1:node%filled+1)
         node%subtrees(before+1)%node => subtree
      ENDIF
      node%filled = node%filled+1
   END SUBROUTINE btree_simple_insertion

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param node ...
!> \param new_node ...
!> \param key ...
!> \param value ...
!> \param before ...
!> \param split_pos ...
!> \param subtree ...
! **************************************************************************************************
   SUBROUTINE btree_left_insertion(tree, node, new_node, key, value, before, split_pos, subtree)
      TYPE(btree), INTENT(IN)                            :: tree
      TYPE(btree_node), INTENT(INOUT)                    :: node, new_node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(cp2d), INTENT(IN)                             :: value
      INTEGER, INTENT(IN)                                :: before, split_pos
      TYPE(btree_node), OPTIONAL, POINTER                :: subtree

      !
      new_node%filled = (tree%b%max_fill)-(split_pos-1)
      new_node%keys(1:new_node%filled) = &
         node%keys(split_pos:tree%b%max_fill)
      new_node%values(1:new_node%filled) = &
         node%values(split_pos:tree%b%max_fill)
      !IF (ASSOCIATED (node%subtrees(1)%node)) THEN
      IF (PRESENT(subtree)) THEN
         IF (before .EQ. split_pos) THEN
            new_node%subtrees(2:new_node%filled+1) = &
               node%subtrees(split_pos+1:tree%b%max_fill+1)
            new_node%subtrees(1)%node => subtree
         ELSE
            new_node%subtrees(1:new_node%filled+1) = &
               node%subtrees(split_pos:tree%b%max_fill+1)
         ENDIF
      ENDIF
      ! Fill node%{keys,values}(1:node%filled), where node%filled
      ! is split_pos-1, but do insert the new value at ge_pos. The
      ! key/value at split_pos is to be inserted into the
      ! parent.
      ! The new tree is added to the right of the new insertion.
      node%keys(before+1:node%filled) = node%keys(before:node%filled-1)
      node%keys(before) = key
      node%values(before+1:node%filled) = node%values(before:node%filled-1)
      node%values(before) = value
      IF (PRESENT(subtree)) THEN
         node%subtrees(before+2:node%filled+1) = &
            node%subtrees(before+1:node%filled)
         node%subtrees(before+1)%node => subtree
      ELSE
         NULLIFY (node%subtrees(before+1)%node)
      ENDIF
   END SUBROUTINE btree_left_insertion

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param node ...
!> \param new_node ...
!> \param key ...
!> \param value ...
!> \param before ...
!> \param split_pos ...
!> \param subtree ...
! **************************************************************************************************
   SUBROUTINE btree_right_insertion(tree, node, new_node, key, value, before, split_pos, subtree)
      TYPE(btree), INTENT(IN)                            :: tree
      TYPE(btree_node), INTENT(INOUT)                    :: node, new_node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(cp2d), INTENT(IN)                             :: value
      INTEGER, INTENT(IN)                                :: before, split_pos
      TYPE(btree_node), OPTIONAL, POINTER                :: subtree

      !
      new_node%filled = (tree%b%max_fill+1)-split_pos
      new_node%keys(1:before-split_pos-1) = &
         node%keys(split_pos+1:before-1)
      new_node%keys(before-split_pos) = key
      new_node%keys(before-split_pos+1:new_node%filled) = &
         node%keys(before:tree%b%max_fill)
      new_node%values(1:before-split_pos-1) = &
         node%values(split_pos+1:before-1)
      new_node%values(before-split_pos) = value
      new_node%values(before-split_pos+1:new_node%filled) = &
         node%values(before:tree%b%max_fill)
      IF (PRESENT(subtree)) THEN
         new_node%subtrees(1:before-split_pos) = &
            node%subtrees(split_pos+1:before)
         new_node%subtrees(before-split_pos+1)%node => subtree
         new_node%subtrees(before-split_pos+2:new_node%filled+1) = &
            node%subtrees(before+1:tree%b%max_fill+1)
      ENDIF
   END SUBROUTINE btree_right_insertion

   ! node is a non-leaf node
! **************************************************************************************************
!> \brief ...
!> \param node ...
! **************************************************************************************************
   SUBROUTINE btree_adopt_subtrees(node)
      TYPE(btree_node), POINTER                          :: node

      INTEGER                                            :: i

      !
      ! Assume that node is not a leaf!
      DO i = 1, node%filled+1
         !IF (ASSOCIATED (node%subtrees(i)%node)) THEN
         !IF (.NOT. ASSOCIATED (node%subtrees(i)%node%parent,&
         ! node)) THEN
         node%subtrees(i)%node%parent => node
         !ENDIF
         !ENDIF
      ENDDO
   END SUBROUTINE btree_adopt_subtrees

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param value ...
! **************************************************************************************************
   SUBROUTINE btree_new_root(tree, key, value)
      TYPE(btree), INTENT(INOUT)                         :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(cp2d), INTENT(IN)                             :: value

      TYPE(btree_node), POINTER                          :: new_root, old_root

      !
      CALL btree_new_node(tree, new_root)
      new_root%filled = 1
      new_root%keys(1) = key
      new_root%values(1) = value
      IF (ASSOCIATED(tree%b%root)) THEN
         old_root => tree%b%root
         old_root%parent => new_root
         new_root%subtrees(1)%node => old_root
         old_root%parent => new_root
      ENDIF
      tree%b%root => new_root
   END SUBROUTINE btree_new_root

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param node ...
! **************************************************************************************************
   SUBROUTINE btree_new_node(tree, node)
      TYPE(btree), INTENT(INOUT)                         :: tree
      TYPE(btree_node), POINTER                          :: node

      INTEGER                                            :: i

      !
      ALLOCATE (node)
      ALLOCATE (node%keys(tree%b%max_fill))
      ALLOCATE (node%values(tree%b%max_fill))
      ALLOCATE (node%subtrees(tree%b%max_fill+1))
      DO i = 1, tree%b%max_fill+1
         NULLIFY (node%subtrees(i)%node)
      ENDDO
      node%filled = 0
      NULLIFY (node%parent)
      tree%b%lastid = tree%b%lastid+1
      node%id = tree%b%lastid
   END SUBROUTINE btree_new_node

! **************************************************************************************************
!> \brief ...
!> \param node ...
! **************************************************************************************************
   SUBROUTINE btree_free_node(node)
      TYPE(btree_node), POINTER                          :: node

      !
      DEALLOCATE (node%keys)
      DEALLOCATE (node%values)
      DEALLOCATE (node%subtrees)
      DEALLOCATE (node)
   END SUBROUTINE btree_free_node

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param value ...
!> \param exists ...
! **************************************************************************************************
   SUBROUTINE btree_find(tree, key, value, exists)
      TYPE(btree), INTENT(IN)                            :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(cp2d), INTENT(OUT)                            :: value
      LOGICAL, INTENT(OUT), OPTIONAL                     :: exists

      INTEGER                                            :: position
      TYPE(btree_node), POINTER                          :: node

!

      !
      CALL btree_find_full(tree, key, node, position, short=.TRUE.)
      IF (PRESENT(exists)) THEN
         exists = position .GT. 0
      ENDIF
      IF (position .GT. 0) THEN
         value = node%values(position)
      ENDIF
   END SUBROUTINE btree_find

! **************************************************************************************************
!> \brief ...
!> \param keys ...
!> \param key ...
!> \param position ...
!> \param filled ...
! **************************************************************************************************
   SUBROUTINE btree_node_find_ge_pos(keys, key, position, filled)
      INTEGER(KIND=keyt), DIMENSION(:)                   :: keys
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER, INTENT(OUT)                               :: position
      INTEGER, INTENT(IN)                                :: filled

      INTEGER                                            :: left, right

      !
      IF (keys(1) .GE. key) THEN
         position = 1
         RETURN
      ENDIF
      IF (keys(filled) .LT. key) THEN
         position = filled+1
         RETURN
      ENDIF
      left = 2
      right = filled
      position = MAX(ISHFT(left+right, -1), left)
      DO WHILE (left .LE. right)
         IF (keys(position) .GE. key .AND. keys(position-1) .LT. key) THEN
            RETURN
         ENDIF
         IF (keys(position) .GE. key) right = MIN(position, right-1)
         IF (keys(position) .LT. key) left = MAX(position, left+1)
         position = MAX(ISHFT(left+right, -1), left)
      ENDDO
   END SUBROUTINE btree_node_find_ge_pos

! **************************************************************************************************
!> \brief ...
!> \param keys ...
!> \param key ...
!> \param position ...
!> \param filled ...
! **************************************************************************************************
   SUBROUTINE btree_node_find_gt_pos(keys, key, position, filled)
      INTEGER(KIND=keyt), DIMENSION(:)                   :: keys
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER, INTENT(OUT)                               :: position
      INTEGER, INTENT(IN)                                :: filled

      INTEGER                                            :: left, right

      !
      IF (keys(1) .GT. key) THEN
         position = 1
         RETURN
      ENDIF
      IF (keys(filled) .LE. key) THEN
         position = filled+1
         RETURN
      ENDIF
      left = 2
      right = filled
      position = MAX(ISHFT(left+right, -1), left)
      DO WHILE (left .LE. right)
         IF (keys(position) .GT. key .AND. keys(position-1) .LE. key) THEN
            RETURN
         ENDIF
         IF (keys(position) .GT. key) right = MIN(position, right-1)
         IF (keys(position) .LE. key) left = MAX(position, left+1)
         position = MAX(ISHFT(left+right, -1), left)
      ENDDO
   END SUBROUTINE btree_node_find_gt_pos
! **************************************************************************************************
!> \brief ...
!> \param keys ...
!> \param key ...
!> \param position ...
!> \param filled ...
!> \param first ...
! **************************************************************************************************
   SUBROUTINE btree_node_find_gte_pos(keys, key, position, filled, first)
      INTEGER(KIND=keyt), DIMENSION(:)                   :: keys
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER, INTENT(OUT)                               :: position
      INTEGER, INTENT(IN)                                :: filled
      INTEGER, INTENT(IN), OPTIONAL                      :: first

      INTEGER                                            :: left, one, right

      !
      one = 1
      IF (PRESENT(FIRST)) one = first
      IF (one .LE. filled) THEN
         IF (keys(one) .GT. key) THEN
            position = one
            RETURN
         ENDIF
      ENDIF
      IF (keys(filled) .LE. key) THEN
         position = filled+1
         RETURN
      ENDIF
      left = one+1
      right = filled
      position = MAX(ISHFT(left+right, -1), left)
      DO WHILE (left .LE. right)
         IF (keys(position) .GT. key .AND. keys(position-1) .LE. key) THEN
            RETURN
         ENDIF
         IF (keys(position) .GT. key) right = MIN(position, right-1)
         IF (keys(position) .LE. key) left = MAX(position, left+1)
         position = MAX(ISHFT(left+right, -1), left)
      ENDDO
   END SUBROUTINE btree_node_find_gte_pos

   ! node is unassociated and position=0 if not found
   ! Precondition: The key is tree or its subtree.
! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param node ...
!> \param position ...
!> \param ge_position ...
!> \param short ...
! **************************************************************************************************
   SUBROUTINE btree_find_full(tree, key, node, position, ge_position, short)
      TYPE(btree), INTENT(IN)                            :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(btree_node), POINTER                          :: node
      INTEGER, INTENT(OUT)                               :: position
      INTEGER, INTENT(OUT), OPTIONAL                     :: ge_position
      LOGICAL, INTENT(IN), OPTIONAL                      :: short

      INTEGER                                            :: gti
      LOGICAL                                            :: stop_short

! Used mark searches

      !
      stop_short = .FALSE.
      IF (PRESENT(short)) stop_short = short
      NULLIFY (node)
      position = 0
      IF (PRESENT(ge_position)) ge_position = 0
      !IF (tree%b%n .EQ. 0) RETURN
      IF (.NOT. ASSOCIATED(tree%b%root)) RETURN
      gti = 1
      ! Try to find the key in the given node. If it's found, then
      ! return the node.
      node => tree%b%root
      descent: DO WHILE (.TRUE.)
         ! Try to find the first element equal to or greater than the
         ! one we're searching for.
         CALL btree_node_find_ge_pos(node%keys, key, position, node%filled)
         ! One of three things is now true about position: it's now
         ! greater than the number of keys (if all keys are smaller), or
         ! it points to the key that is equal to or greater than the one
         ! we are searching for. If it is found and we are just
         ! searching for one equal element (i.e., user search), we can
         ! return.
         IF (stop_short .AND. position .LE. node%filled) THEN
            IF (node%keys(position) .EQ. key) THEN
               IF (PRESENT(ge_position)) ge_position = position
               RETURN
            ENDIF
         ENDIF
         ! If the key is not found, then either return the GE position
         ! if we're in a leaf (case 2 here), otherwise descend into the
         ! subtrees.
         !CALL btree_node_find_gt_pos (node%keys, key, gti, node%filled, position)
         CALL btree_node_find_gte_pos(node%keys, key, gti, node%filled, position)
         IF (ASSOCIATED(node%subtrees(1)%node)) THEN
            node => node%subtrees(gti)%node
         ELSE
            IF (PRESENT(ge_position)) ge_position = gti
            position = 0
            RETURN
         ENDIF
      END DO descent
   END SUBROUTINE btree_find_full

   ! node is unassociated and position=0 if not found
   ! Precondition: The key is tree or its subtree.
! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param node ...
!> \param gti ...
! **************************************************************************************************
   SUBROUTINE btree_find_leaf(tree, key, node, gti)
      TYPE(btree), INTENT(IN)                            :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(btree_node), POINTER                          :: node
      INTEGER, INTENT(OUT)                               :: gti

      !
      NULLIFY (node)
      !IF (tree%b%n .EQ. 0) RETURN
      IF (.NOT. ASSOCIATED(tree%b%root)) RETURN
      gti = 1
      ! Try to find the key in the given node. If it's found, then
      ! return the node.
      node => tree%b%root
      descent: DO WHILE (.TRUE.)
         ! Try to find the first element equal to or greater than the
         ! one we're searching for.
         !CALL btree_node_find_ge_pos (node%keys, key, position, node%filled)
         ! One of three things is now true about position: it's now
         ! greater than the number of keys (if all keys are smaller), or
         ! it points to the key that is equal to or greater than the one
         ! we are searching for. If it is found and we are just
         ! searching for one equal element (i.e., user search), we can
         ! return.
         !
         ! If the key is not found, then either return the GE position
         ! if we're in a leaf (case 2 here), otherwise descend into the
         ! subtrees.
         CALL btree_node_find_gt_pos(node%keys, key, gti, node%filled)
         !CALL btree_node_find_gt2_pos (node%keys, key, i, node%filled)
         !IF (i .NE. gti) WRITE(*,*)'XXXX difference',i,gti
         IF (ASSOCIATED(node%subtrees(1)%node)) THEN
            node => node%subtrees(gti)%node
         ELSE
            RETURN
         ENDIF
      END DO descent
   END SUBROUTINE btree_find_leaf

END MODULE btree_i8_k_cp2d_v
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   B-tree implementation template
!> \author  Urban Borstnik
!> \date    2000-05-12
!> \version 1.0
!> <b>Modification history:</b>
!> - Created
! **************************************************************************************************

MODULE btree_i8_k_zp2d_v

   IMPLICIT NONE
   PRIVATE

   !API
   PUBLIC :: btree, btree_new, btree_add, btree_find, &
             btree_delete, btree_get_entries

   INTEGER, PARAMETER :: keyt = SELECTED_INT_KIND(10)
   INTEGER, PARAMETER :: valt = SELECTED_INT_KIND(5); 
   INTEGER, PARAMETER :: dp = KIND(0.0d0)

   TYPE zp2d
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER :: p
      LOGICAL :: tr
   END TYPE zp2d

   PUBLIC :: zp2d

   TYPE btree_node
      INTEGER id
      INTEGER :: filled
      INTEGER(KIND=keyt), DIMENSION(:), POINTER :: keys
      TYPE(zp2d), DIMENSION(:), POINTER :: values
      TYPE(btree_node_p), DIMENSION(:), POINTER :: subtrees
      TYPE(btree_node), POINTER :: parent
   END TYPE btree_node

   TYPE btree_node_p
      TYPE(btree_node), POINTER :: node
   END TYPE btree_node_p

   TYPE btree_node_structure
      INTEGER :: min_fill, max_fill
      INTEGER :: n
      INTEGER :: lastid
      INTEGER :: refcount
      TYPE(btree_node), POINTER :: root
   END TYPE btree_node_structure

   TYPE btree
      TYPE(btree_node_structure) :: b
   END TYPE btree

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE btree_new(tree, order)
      TYPE(btree), INTENT(OUT)                           :: tree
      INTEGER, INTENT(IN), OPTIONAL                      :: order

      INTEGER                                            :: maxs, mins

      !
      IF (PRESENT(order)) THEN
         maxs = order-1
      ELSE
         maxs = 15
      ENDIF
      mins = ISHFT(maxs, -1)
      IF (mins*2 .GT. maxs) maxs = 2*maxs
      IF (mins .LT. 1) mins = 1
      IF (maxs .LT. 3) maxs = 3
      tree%b%min_fill = mins
      tree%b%max_fill = maxs
      tree%b%refcount = 1
      tree%b%n = 0
      NULLIFY (tree%b%root)
      tree%b%lastid = 0
   END SUBROUTINE btree_new

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \retval num_entries ...
! **************************************************************************************************
   FUNCTION btree_get_entries(tree) RESULT(num_entries)
      TYPE(btree), INTENT(INOUT)                         :: tree
      INTEGER                                            :: num_entries

      num_entries = tree%b%n
   END FUNCTION btree_get_entries

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param keys ...
!> \param values ...
! **************************************************************************************************
   SUBROUTINE btree_delete(tree, keys, values)
      TYPE(btree), INTENT(INOUT)                         :: tree
      INTEGER(KIND=keyt), DIMENSION(:), INTENT(OUT), &
         OPTIONAL                                        :: keys
      TYPE(zp2d), DIMENSION(:), INTENT(OUT), OPTIONAL    :: values

      INTEGER                                            :: pos

      !
      IF (ASSOCIATED(tree%b%root)) THEN
         pos = 0
         IF (PRESENT(keys) .AND. PRESENT(values)) THEN
            pos = 1
            CALL btree_delete_node(tree%b%root, pos, keys, values)
         ELSE
            CALL btree_delete_node(tree%b%root)
         ENDIF
      ENDIF
      NULLIFY (tree%b%root)
   END SUBROUTINE btree_delete

! **************************************************************************************************
!> \brief ...
!> \param node ...
!> \param pos ...
!> \param keys ...
!> \param values ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE btree_delete_node(node, pos, keys, values)
      TYPE(btree_node), POINTER                          :: node
      INTEGER, INTENT(INOUT), OPTIONAL                   :: pos
      INTEGER(KIND=keyt), DIMENSION(:), INTENT(INOUT), &
         OPTIONAL                                        :: keys
      TYPE(zp2d), DIMENSION(:), INTENT(INOUT), OPTIONAL  :: values

      INTEGER                                            :: i

!

      !
      IF (node%filled .GT. 0 .AND. ASSOCIATED(node%subtrees(1)%node)) THEN
         DO i = 1, node%filled+1
            IF (PRESENT(pos)) THEN
               CALL btree_delete_node(node%subtrees(i)%node, pos, keys, values)
            ELSE
               CALL btree_delete_node(node%subtrees(i)%node)
            ENDIF
            IF (PRESENT(pos) .AND. i .LE. node%filled) THEN
               keys(pos) = node%keys(i)
               values(pos) = node%values(i)
               pos = pos+1
            ENDIF
         ENDDO
      ELSEIF (PRESENT(pos) .AND. node%filled .GT. 0) THEN
         keys(pos:pos+node%filled-1) = node%keys(1:node%filled)
         values(pos:pos+node%filled-1) = node%values(1:node%filled)
         pos = pos+node%filled
      ENDIF
      CALL btree_free_node(node)
   END SUBROUTINE btree_delete_node

   ! Find the key
   ! IF node still has space, insert & update the node
   ! else
   ! 1. select median
   ! 2. split keys into two nodes (one is new)
   ! 3. insert separation key put into parent, and repeat upwards
! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param value ...
!> \param exists ...
!> \param existing_value ...
!> \param replace ...
! **************************************************************************************************
   SUBROUTINE btree_add(tree, key, value, exists, existing_value, replace)
      TYPE(btree), INTENT(INOUT)                         :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(zp2d), INTENT(IN)                             :: value
      LOGICAL, INTENT(OUT), OPTIONAL                     :: exists
      TYPE(zp2d), INTENT(OUT), OPTIONAL                  :: existing_value
      LOGICAL, INTENT(IN), OPTIONAL                      :: replace

      INTEGER                                            :: ge_pos, position
      TYPE(btree_node), POINTER                          :: node

!

      !
      IF (PRESENT(exists)) THEN
         CALL btree_find_full(tree, key, node, position, ge_pos, short=.TRUE.)
         IF (position .GT. 0) THEN
            exists = .TRUE.
            existing_value = node%values(position)
            IF (PRESENT(replace)) THEN
               IF (replace) THEN
                  node%values(position) = value
               ENDIF
            ENDIF
            RETURN
         ELSE
            exists = .FALSE.
         ENDIF
      ELSE
         CALL btree_find_leaf(tree, key, node, ge_pos)
      ENDIF
      CALL btree_add_into(tree, node, key, value, before=ge_pos)
      IF (PRESENT(exists)) existing_value = value
      tree%b%n = tree%b%n+1
   END SUBROUTINE btree_add

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param node ...
!> \param key ...
!> \param value ...
!> \param before ...
!> \param subtree ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE btree_add_into(tree, node, key, value, before, subtree)
      TYPE(btree), INTENT(INOUT)                         :: tree
      TYPE(btree_node), POINTER                          :: node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(zp2d), INTENT(IN)                             :: value
      INTEGER, INTENT(IN), OPTIONAL                      :: before
      TYPE(btree_node), OPTIONAL, POINTER                :: subtree

      INTEGER                                            :: ge_pos, split_pos
      INTEGER(KIND=keyt)                                 :: upgrade_key
      LOGICAL                                            :: leaf
      TYPE(btree_node), POINTER                          :: new_node
      TYPE(zp2d)                                         :: upgrade_value

!

      !
      ! Root is special
      IF (.NOT. ASSOCIATED(node)) THEN
         CALL btree_new_root(tree, key, value)
         IF (PRESENT(subtree)) THEN
            tree%b%root%subtrees(2)%node => subtree
            subtree%parent => tree%b%root
         ENDIF
         RETURN
      ENDIF
      ! Where the insertion takes place.
      IF (PRESENT(before)) THEN
         ge_pos = before
      ELSE
         CALL btree_node_find_gt_pos(node%keys, key, ge_pos, node%filled)
      ENDIF
      ! Addition is easy if the node has enough space.
      leaf = .NOT. ASSOCIATED(node%subtrees(1)%node)
      IF (node%filled .LT. tree%b%max_fill) THEN
         IF (PRESENT(subtree)) THEN
            CALL btree_simple_insertion(node, key, value, ge_pos, subtree)
         ELSE
            CALL btree_simple_insertion(node, key, value, ge_pos)
         ENDIF
         RETURN
      ELSE
         split_pos = ISHFT(tree%b%max_fill+1, -1)
         ! I assert that split_pos <= SIZE(node%keys)
         CALL btree_new_node(tree, new_node)
         ! The key to be added falls in the left node
         node%filled = split_pos-1
         IF (ge_pos .LE. split_pos) THEN
            IF (ge_pos .EQ. split_pos) THEN
               upgrade_key = key
               upgrade_value = value
            ELSE
               upgrade_key = node%keys(split_pos-1)
               upgrade_value = node%values(split_pos-1)
            ENDIF
            IF (PRESENT(subtree)) THEN
               CALL btree_left_insertion(tree, node, new_node, key, value, &
                                         ge_pos, split_pos, subtree)
               !CALL btree_adopt_subtrees (new_node)
            ELSE
               CALL btree_left_insertion(tree, node, new_node, key, value, &
                                         ge_pos, split_pos)
            ENDIF
            !
         ELSE
            upgrade_key = node%keys(split_pos)
            upgrade_value = node%values(split_pos)
            IF (PRESENT(subtree)) THEN
               CALL btree_right_insertion(tree, node, new_node, key, value, &
                                          ge_pos, split_pos, subtree)
               !CALL btree_adopt_subtrees (new_node)
            ELSE
               CALL btree_right_insertion(tree, node, new_node, key, value, &
                                          ge_pos, split_pos)
            ENDIF
            !
         ENDIF
         !
         new_node%parent => node%parent
         !
         IF (.NOT. leaf) THEN
            CALL btree_adopt_subtrees(new_node)
         ENDIF
         !
         CALL btree_add_into(tree, node%parent, upgrade_key, upgrade_value, &
                             subtree=new_node)
         !
      ENDIF
   END SUBROUTINE btree_add_into

! **************************************************************************************************
!> \brief ...
!> \param node ...
!> \param key ...
!> \param value ...
!> \param before ...
!> \param subtree ...
! **************************************************************************************************
   SUBROUTINE btree_simple_insertion(node, key, value, before, subtree)
      TYPE(btree_node), INTENT(INOUT)                    :: node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(zp2d), INTENT(IN)                             :: value
      INTEGER, INTENT(IN)                                :: before
      TYPE(btree_node), OPTIONAL, POINTER                :: subtree

      !
      ! Shift keys
      node%keys(before+1:node%filled+1) = node%keys(before:node%filled)
      node%keys(before) = key
      ! Shift values
      node%values(before+1:node%filled+1) = node%values(before:node%filled)
      node%values(before) = value
      ! Shift subtree pointers, but only if node is not a leaf ; assume
      ! leaf <=> present(subtree)
      IF (PRESENT(subtree)) THEN
         node%subtrees(before+2:node%filled+2) = &
            node%subtrees(before+1:node%filled+1)
         node%subtrees(before+1)%node => subtree
      ENDIF
      node%filled = node%filled+1
   END SUBROUTINE btree_simple_insertion

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param node ...
!> \param new_node ...
!> \param key ...
!> \param value ...
!> \param before ...
!> \param split_pos ...
!> \param subtree ...
! **************************************************************************************************
   SUBROUTINE btree_left_insertion(tree, node, new_node, key, value, before, split_pos, subtree)
      TYPE(btree), INTENT(IN)                            :: tree
      TYPE(btree_node), INTENT(INOUT)                    :: node, new_node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(zp2d), INTENT(IN)                             :: value
      INTEGER, INTENT(IN)                                :: before, split_pos
      TYPE(btree_node), OPTIONAL, POINTER                :: subtree

      !
      new_node%filled = (tree%b%max_fill)-(split_pos-1)
      new_node%keys(1:new_node%filled) = &
         node%keys(split_pos:tree%b%max_fill)
      new_node%values(1:new_node%filled) = &
         node%values(split_pos:tree%b%max_fill)
      !IF (ASSOCIATED (node%subtrees(1)%node)) THEN
      IF (PRESENT(subtree)) THEN
         IF (before .EQ. split_pos) THEN
            new_node%subtrees(2:new_node%filled+1) = &
               node%subtrees(split_pos+1:tree%b%max_fill+1)
            new_node%subtrees(1)%node => subtree
         ELSE
            new_node%subtrees(1:new_node%filled+1) = &
               node%subtrees(split_pos:tree%b%max_fill+1)
         ENDIF
      ENDIF
      ! Fill node%{keys,values}(1:node%filled), where node%filled
      ! is split_pos-1, but do insert the new value at ge_pos. The
      ! key/value at split_pos is to be inserted into the
      ! parent.
      ! The new tree is added to the right of the new insertion.
      node%keys(before+1:node%filled) = node%keys(before:node%filled-1)
      node%keys(before) = key
      node%values(before+1:node%filled) = node%values(before:node%filled-1)
      node%values(before) = value
      IF (PRESENT(subtree)) THEN
         node%subtrees(before+2:node%filled+1) = &
            node%subtrees(before+1:node%filled)
         node%subtrees(before+1)%node => subtree
      ELSE
         NULLIFY (node%subtrees(before+1)%node)
      ENDIF
   END SUBROUTINE btree_left_insertion

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param node ...
!> \param new_node ...
!> \param key ...
!> \param value ...
!> \param before ...
!> \param split_pos ...
!> \param subtree ...
! **************************************************************************************************
   SUBROUTINE btree_right_insertion(tree, node, new_node, key, value, before, split_pos, subtree)
      TYPE(btree), INTENT(IN)                            :: tree
      TYPE(btree_node), INTENT(INOUT)                    :: node, new_node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(zp2d), INTENT(IN)                             :: value
      INTEGER, INTENT(IN)                                :: before, split_pos
      TYPE(btree_node), OPTIONAL, POINTER                :: subtree

      !
      new_node%filled = (tree%b%max_fill+1)-split_pos
      new_node%keys(1:before-split_pos-1) = &
         node%keys(split_pos+1:before-1)
      new_node%keys(before-split_pos) = key
      new_node%keys(before-split_pos+1:new_node%filled) = &
         node%keys(before:tree%b%max_fill)
      new_node%values(1:before-split_pos-1) = &
         node%values(split_pos+1:before-1)
      new_node%values(before-split_pos) = value
      new_node%values(before-split_pos+1:new_node%filled) = &
         node%values(before:tree%b%max_fill)
      IF (PRESENT(subtree)) THEN
         new_node%subtrees(1:before-split_pos) = &
            node%subtrees(split_pos+1:before)
         new_node%subtrees(before-split_pos+1)%node => subtree
         new_node%subtrees(before-split_pos+2:new_node%filled+1) = &
            node%subtrees(before+1:tree%b%max_fill+1)
      ENDIF
   END SUBROUTINE btree_right_insertion

   ! node is a non-leaf node
! **************************************************************************************************
!> \brief ...
!> \param node ...
! **************************************************************************************************
   SUBROUTINE btree_adopt_subtrees(node)
      TYPE(btree_node), POINTER                          :: node

      INTEGER                                            :: i

      !
      ! Assume that node is not a leaf!
      DO i = 1, node%filled+1
         !IF (ASSOCIATED (node%subtrees(i)%node)) THEN
         !IF (.NOT. ASSOCIATED (node%subtrees(i)%node%parent,&
         ! node)) THEN
         node%subtrees(i)%node%parent => node
         !ENDIF
         !ENDIF
      ENDDO
   END SUBROUTINE btree_adopt_subtrees

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param value ...
! **************************************************************************************************
   SUBROUTINE btree_new_root(tree, key, value)
      TYPE(btree), INTENT(INOUT)                         :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(zp2d), INTENT(IN)                             :: value

      TYPE(btree_node), POINTER                          :: new_root, old_root

      !
      CALL btree_new_node(tree, new_root)
      new_root%filled = 1
      new_root%keys(1) = key
      new_root%values(1) = value
      IF (ASSOCIATED(tree%b%root)) THEN
         old_root => tree%b%root
         old_root%parent => new_root
         new_root%subtrees(1)%node => old_root
         old_root%parent => new_root
      ENDIF
      tree%b%root => new_root
   END SUBROUTINE btree_new_root

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param node ...
! **************************************************************************************************
   SUBROUTINE btree_new_node(tree, node)
      TYPE(btree), INTENT(INOUT)                         :: tree
      TYPE(btree_node), POINTER                          :: node

      INTEGER                                            :: i

      !
      ALLOCATE (node)
      ALLOCATE (node%keys(tree%b%max_fill))
      ALLOCATE (node%values(tree%b%max_fill))
      ALLOCATE (node%subtrees(tree%b%max_fill+1))
      DO i = 1, tree%b%max_fill+1
         NULLIFY (node%subtrees(i)%node)
      ENDDO
      node%filled = 0
      NULLIFY (node%parent)
      tree%b%lastid = tree%b%lastid+1
      node%id = tree%b%lastid
   END SUBROUTINE btree_new_node

! **************************************************************************************************
!> \brief ...
!> \param node ...
! **************************************************************************************************
   SUBROUTINE btree_free_node(node)
      TYPE(btree_node), POINTER                          :: node

      !
      DEALLOCATE (node%keys)
      DEALLOCATE (node%values)
      DEALLOCATE (node%subtrees)
      DEALLOCATE (node)
   END SUBROUTINE btree_free_node

! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param value ...
!> \param exists ...
! **************************************************************************************************
   SUBROUTINE btree_find(tree, key, value, exists)
      TYPE(btree), INTENT(IN)                            :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(zp2d), INTENT(OUT)                            :: value
      LOGICAL, INTENT(OUT), OPTIONAL                     :: exists

      INTEGER                                            :: position
      TYPE(btree_node), POINTER                          :: node

!

      !
      CALL btree_find_full(tree, key, node, position, short=.TRUE.)
      IF (PRESENT(exists)) THEN
         exists = position .GT. 0
      ENDIF
      IF (position .GT. 0) THEN
         value = node%values(position)
      ENDIF
   END SUBROUTINE btree_find

! **************************************************************************************************
!> \brief ...
!> \param keys ...
!> \param key ...
!> \param position ...
!> \param filled ...
! **************************************************************************************************
   SUBROUTINE btree_node_find_ge_pos(keys, key, position, filled)
      INTEGER(KIND=keyt), DIMENSION(:)                   :: keys
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER, INTENT(OUT)                               :: position
      INTEGER, INTENT(IN)                                :: filled

      INTEGER                                            :: left, right

      !
      IF (keys(1) .GE. key) THEN
         position = 1
         RETURN
      ENDIF
      IF (keys(filled) .LT. key) THEN
         position = filled+1
         RETURN
      ENDIF
      left = 2
      right = filled
      position = MAX(ISHFT(left+right, -1), left)
      DO WHILE (left .LE. right)
         IF (keys(position) .GE. key .AND. keys(position-1) .LT. key) THEN
            RETURN
         ENDIF
         IF (keys(position) .GE. key) right = MIN(position, right-1)
         IF (keys(position) .LT. key) left = MAX(position, left+1)
         position = MAX(ISHFT(left+right, -1), left)
      ENDDO
   END SUBROUTINE btree_node_find_ge_pos

! **************************************************************************************************
!> \brief ...
!> \param keys ...
!> \param key ...
!> \param position ...
!> \param filled ...
! **************************************************************************************************
   SUBROUTINE btree_node_find_gt_pos(keys, key, position, filled)
      INTEGER(KIND=keyt), DIMENSION(:)                   :: keys
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER, INTENT(OUT)                               :: position
      INTEGER, INTENT(IN)                                :: filled

      INTEGER                                            :: left, right

      !
      IF (keys(1) .GT. key) THEN
         position = 1
         RETURN
      ENDIF
      IF (keys(filled) .LE. key) THEN
         position = filled+1
         RETURN
      ENDIF
      left = 2
      right = filled
      position = MAX(ISHFT(left+right, -1), left)
      DO WHILE (left .LE. right)
         IF (keys(position) .GT. key .AND. keys(position-1) .LE. key) THEN
            RETURN
         ENDIF
         IF (keys(position) .GT. key) right = MIN(position, right-1)
         IF (keys(position) .LE. key) left = MAX(position, left+1)
         position = MAX(ISHFT(left+right, -1), left)
      ENDDO
   END SUBROUTINE btree_node_find_gt_pos
! **************************************************************************************************
!> \brief ...
!> \param keys ...
!> \param key ...
!> \param position ...
!> \param filled ...
!> \param first ...
! **************************************************************************************************
   SUBROUTINE btree_node_find_gte_pos(keys, key, position, filled, first)
      INTEGER(KIND=keyt), DIMENSION(:)                   :: keys
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER, INTENT(OUT)                               :: position
      INTEGER, INTENT(IN)                                :: filled
      INTEGER, INTENT(IN), OPTIONAL                      :: first

      INTEGER                                            :: left, one, right

      !
      one = 1
      IF (PRESENT(FIRST)) one = first
      IF (one .LE. filled) THEN
         IF (keys(one) .GT. key) THEN
            position = one
            RETURN
         ENDIF
      ENDIF
      IF (keys(filled) .LE. key) THEN
         position = filled+1
         RETURN
      ENDIF
      left = one+1
      right = filled
      position = MAX(ISHFT(left+right, -1), left)
      DO WHILE (left .LE. right)
         IF (keys(position) .GT. key .AND. keys(position-1) .LE. key) THEN
            RETURN
         ENDIF
         IF (keys(position) .GT. key) right = MIN(position, right-1)
         IF (keys(position) .LE. key) left = MAX(position, left+1)
         position = MAX(ISHFT(left+right, -1), left)
      ENDDO
   END SUBROUTINE btree_node_find_gte_pos

   ! node is unassociated and position=0 if not found
   ! Precondition: The key is tree or its subtree.
! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param node ...
!> \param position ...
!> \param ge_position ...
!> \param short ...
! **************************************************************************************************
   SUBROUTINE btree_find_full(tree, key, node, position, ge_position, short)
      TYPE(btree), INTENT(IN)                            :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(btree_node), POINTER                          :: node
      INTEGER, INTENT(OUT)                               :: position
      INTEGER, INTENT(OUT), OPTIONAL                     :: ge_position
      LOGICAL, INTENT(IN), OPTIONAL                      :: short

      INTEGER                                            :: gti
      LOGICAL                                            :: stop_short

! Used mark searches

      !
      stop_short = .FALSE.
      IF (PRESENT(short)) stop_short = short
      NULLIFY (node)
      position = 0
      IF (PRESENT(ge_position)) ge_position = 0
      !IF (tree%b%n .EQ. 0) RETURN
      IF (.NOT. ASSOCIATED(tree%b%root)) RETURN
      gti = 1
      ! Try to find the key in the given node. If it's found, then
      ! return the node.
      node => tree%b%root
      descent: DO WHILE (.TRUE.)
         ! Try to find the first element equal to or greater than the
         ! one we're searching for.
         CALL btree_node_find_ge_pos(node%keys, key, position, node%filled)
         ! One of three things is now true about position: it's now
         ! greater than the number of keys (if all keys are smaller), or
         ! it points to the key that is equal to or greater than the one
         ! we are searching for. If it is found and we are just
         ! searching for one equal element (i.e., user search), we can
         ! return.
         IF (stop_short .AND. position .LE. node%filled) THEN
            IF (node%keys(position) .EQ. key) THEN
               IF (PRESENT(ge_position)) ge_position = position
               RETURN
            ENDIF
         ENDIF
         ! If the key is not found, then either return the GE position
         ! if we're in a leaf (case 2 here), otherwise descend into the
         ! subtrees.
         !CALL btree_node_find_gt_pos (node%keys, key, gti, node%filled, position)
         CALL btree_node_find_gte_pos(node%keys, key, gti, node%filled, position)
         IF (ASSOCIATED(node%subtrees(1)%node)) THEN
            node => node%subtrees(gti)%node
         ELSE
            IF (PRESENT(ge_position)) ge_position = gti
            position = 0
            RETURN
         ENDIF
      END DO descent
   END SUBROUTINE btree_find_full

   ! node is unassociated and position=0 if not found
   ! Precondition: The key is tree or its subtree.
! **************************************************************************************************
!> \brief ...
!> \param tree ...
!> \param key ...
!> \param node ...
!> \param gti ...
! **************************************************************************************************
   SUBROUTINE btree_find_leaf(tree, key, node, gti)
      TYPE(btree), INTENT(IN)                            :: tree
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      TYPE(btree_node), POINTER                          :: node
      INTEGER, INTENT(OUT)                               :: gti

      !
      NULLIFY (node)
      !IF (tree%b%n .EQ. 0) RETURN
      IF (.NOT. ASSOCIATED(tree%b%root)) RETURN
      gti = 1
      ! Try to find the key in the given node. If it's found, then
      ! return the node.
      node => tree%b%root
      descent: DO WHILE (.TRUE.)
         ! Try to find the first element equal to or greater than the
         ! one we're searching for.
         !CALL btree_node_find_ge_pos (node%keys, key, position, node%filled)
         ! One of three things is now true about position: it's now
         ! greater than the number of keys (if all keys are smaller), or
         ! it points to the key that is equal to or greater than the one
         ! we are searching for. If it is found and we are just
         ! searching for one equal element (i.e., user search), we can
         ! return.
         !
         ! If the key is not found, then either return the GE position
         ! if we're in a leaf (case 2 here), otherwise descend into the
         ! subtrees.
         CALL btree_node_find_gt_pos(node%keys, key, gti, node%filled)
         !CALL btree_node_find_gt2_pos (node%keys, key, i, node%filled)
         !IF (i .NE. gti) WRITE(*,*)'XXXX difference',i,gti
         IF (ASSOCIATED(node%subtrees(1)%node)) THEN
            node => node%subtrees(gti)%node
         ELSE
            RETURN
         ENDIF
      END DO descent
   END SUBROUTINE btree_find_leaf

END MODULE btree_i8_k_zp2d_v
